#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import ConfigParser
import StringIO
import re

from collections import OrderedDict

from duoauthproxy.lib import protect


class ConfigError(Exception):
    pass


class ConfigProvider:
    """
    Provides access to an Auth Proxy configuration
    """
    def __init__(self, ini_string, source=None):
        """
        Create a ConfigProvider based on the provided configuration string (in INI style).

        Run validations on the configuration:
          No duplicates allowed
          At least one section
        and raise a ConfigError if there's a problem

        Args:
            ini_string (str): An INI style string containing the configuration
            source (str): Where the INI string came from, for logging/error reporting purposes
        """
        duplicates = find_duplicate_sections(ini_string)
        if duplicates:
            raise ConfigError("Configuration loaded from {0} has duplicate sections: {1}".format(source, duplicates))

        self.parser = ConfigParser.ConfigParser()

        config_stringio = StringIO.StringIO(ini_string)
        try:
            self.parser.readfp(config_stringio)
        except ConfigParser.MissingSectionHeaderError:
            raise ConfigError('Config loaded from {0} has no sections.'.format(source))

        self._apply_main_overrides()

    def list_sections(self):
        """
        Return a list of the sections present in the config

        Returns:
            A list of the section names (strings) that are in the config
        """
        return self.parser.sections()

    def get_section_config(self, section_name):
        """
        Get the configuration details for a specified config section.

        Args:
            section_name (str): The section name whose config we want

        Returns:
            A base.ConfigDict for the requested section.  Will be blank if no such section exists.
        """
        if not section_name or section_name not in self.list_sections():
            return ConfigDict({})

        return ConfigDict(self.parser.items(section_name, raw=True))

    def get_main_section_config(self):
        """
        Convenience method for getting the 'main' section config, so the caller doesn't need to know
        the casing of the config

        Returns:
            The section config for the main section
        """
        main_section_name = find_main_section_name(self.list_sections())
        return self.get_section_config(main_section_name)

    def get_all_sections(self):
        """
        Convenience method to get all the section configs

        Returns:
            OrderedDict: map of section name -> ConfigDict for that section, includes all sections present in the config

        """
        all_sections = OrderedDict()

        for section_name in self.list_sections():
            all_sections[section_name] = self.get_section_config(section_name)

        return all_sections

    def _apply_main_overrides(self):
        """ Apply special logic to the main section.

        The main section config may contain a 'server' and 'client' key.  If so, the specified client
        should override any client declaration made in the specified server section.
        The main section can also specify flags that should be passed down into the server sections
        and that will happen here as well.
        """
        main_config = self.get_main_section_config()
        if 'server' in main_config and 'client' in main_config:
            main_server = main_config.get_str('server')
            main_client = main_config.get_str('client')
            try:
                self.parser.set(main_server, 'client', main_client)
            except ConfigParser.NoSectionError:
                # There is no section for the specified server in [main], so ignore it
                pass

        for section in self.list_sections():
            if main_config.get_bool('debug', False):
                self.parser.set(section, 'debug', value='True')

            http_ca_certs_file = main_config.get_str('http_ca_certs_file', '')
            if http_ca_certs_file:
                self.parser.set(section, 'http_ca_certs_file', value=http_ca_certs_file)

            interface = main_config.get_str('interface', '')
            if interface:
                if get_module_type(section) == 'server':
                    # Only apply main interface if one isn't specified already
                    if not self.parser.has_option(section, 'interface'):
                            self.parser.set(section, 'interface', value=interface)

    @staticmethod
    def from_ini_file(config_file, source):
        """
        Factory method for creating a ConfigProvider from a provided INI file.

        Will read the file contents, clean it up*, and return the created ConfigProvider.

        *Normalize line endings

        Args:
            config_file (file): the INI file to read
            source (str): Where the INI file came from, for logging/error reporting purposes

        Returns:
            A ConfigProvider built with the contents of the INI file
        """
        file_contents = config_file.read()
        file_contents = normalize_line_endings(file_contents)
        return ConfigProvider(file_contents, source)


def find_duplicate_sections(ini_string):
    """
    Find (case insensitively) any duplicate sections in the provided INI-like string

    Args:
        ini_string (str): the config ini string to check

    Returns:
        A (possibly empty) list of any duplicated sections
    """
    sections_found = set()
    duplicates = set()

    for match in map(str.lower, re.findall('^\[(\w+)\]', ini_string, re.MULTILINE)):
        if match in sections_found:
            duplicates.add(match)
        else:
            sections_found.add(match)

    return list(duplicates)


def find_main_section_name(section_names):
    """
    Check the list of section_names (strings) for anything that matches 'main' in any casing.

    Args:
        section_names (list of str): The seciton name list to search

    Returns:
        The 'main' section name in original casing, or None.
    """
    for name in section_names:
        if name.lower() == 'main':
            return name

    return None


def normalize_line_endings(config_string):
    """
    Transform the provided string to prevent if having both UNIX and Windows
    style line endings.  If there is a mix, replace the UNIX ones with Windows.

    Args:
        config_string (str): The string to transform

    Returns:
        The transformed string
    """
    if (re.search('(?<!\r)\n', config_string) is not None and
            re.search(r'\r\n', config_string) is not None):
        # The config string has both Windows-style(\r\n) and Unix-style(\n) line
        # endings. Since there are Windows-style line endings, assume this is
        # a Windows config file and replace all the Unix-style line endings
        # with their Windows equivalent
        config_string = re.sub('(?<!\r)\n', r'\r\n', config_string)
    return config_string


def get_module_type(section_name):
    """Return module type for a given section name
    Args:
        section_name (str): Name of section
    Returns:
        str: One of 'server' 'client' or ''
    """
    if 'server' in section_name.lower():
        return 'server'
    elif 'client' in section_name.lower():
        return 'client'
    else:
        return ''


class ConfigDict(dict):
    def get_str(self, key, default=None):
        assert(default is None or type(default) is str)

        value = self.get(key)
        if value is None:
            if default is None:
                raise ConfigError('Missing required configuration item: \'%s\'' % key)
            else:
                return default
        else:
            assert(type(value) is str)
            return value

    def get_protected_str(self, key, unsecured_key, default=None):
        assert(default is None or type(default) is str)

        if not protect.PROTECT_ENABLED:
            return self.get_str(unsecured_key, default)

        value = self.get(key)
        if value:
            assert(type(value) is str)
            return protect.unprotect(value)
        elif unsecured_key:
            return self.get_str(unsecured_key, default)
        return default

    def get_int(self, key, default=None):
        assert(default is None or type(default) is int)

        value = self.get(key)
        if value is None:
            if default is None:
                raise ConfigError('Missing required configuration item: \'%s\'' % key)
            else:
                return default
        else:
            assert(type(value)) is str
            try:
                return int(value)
            except ValueError:
                raise ConfigError('Invalid value for configuration item: \'%s\' must be an integer'
                                  % key)

    def get_bool(self, key, default=None):
        assert(default is None or type(default) is bool)

        value = self.get(key)
        if value is None:
            if default is None:
                raise ConfigError('Missing required configuration item: \'%s\'' % key)
            else:
                return default
        else:
            # accept (case-insensitive) 'true', 'false', zero, or nonzero. all other strings will be rejected
            assert(type(value)) is str
            if value.lower() == 'true':
                return True
            elif value.lower() == 'false':
                return False
            else:
                try:
                    return bool(int(value))
                except ValueError:
                    raise ConfigError('Invalid value for configuration item: \'%s\' must be a boolean value'
                                      % key)

    def get_enum(self, key, values, default=None, transform=lambda x: x):
        assert(default is None or default in values)

        value = self.get(key)
        if value is None:
            if default is None:
                raise ConfigError('Missing required configuration item: \'%s\'' % key)
            else:
                return default
        else:
            assert(type(value)) is str
            value = transform(value)

            if value not in values:
                raise ConfigError('Invalid value for configuration item: \'%s\' must be one of %s'
                                  % (key, ', '.join('\'%s\'' % x for x in values)))
            return value
