#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
from twisted.internet import defer

from duoauthproxy.lib import util
from ..connectivity_toolbox import STANDARD_TOOLBOX
from duoauthproxy.lib.config_provider import ConfigError
from duoauthproxy.modules.ad_client import _ADServiceClientFactory
from ..connectivity_results import (
    AdClientResult,
    MissingConfigKeyProblem,
    ConfigErrorProblem,
    ConfigCheckResult,
    UnmetPrerequisiteSkippedTestResult,
    LdapConnectResult,
    LdapHostResult
)


@defer.inlineCallbacks
def check_ad_client(config, toolbox=STANDARD_TOOLBOX):
    """Checks a radius client section config by testing
      1) The config is valid
      2) We can connect via TCP to the host
      3) We can make LDAP connection to the host
      4) We can bind as the service account user
      5. A search with a given filter and security group dn returns something
    Args:
        config: A ConfigDict for an ad client module
        toolbox: a ConnectivityTestToolbox for doing individual connectivity tests
    Returns:
        AdClientResult: The result of the testing
    """
    host_results = []

    config_result = validate_ad_client_config(config)
    if not config_result.is_successful():
        # Config is bad, there's no point in testing the hosts
        defer.returnValue(AdClientResult(config_result, host_results))

    factory_kwargs = util.parse_ad_client(config)
    hosts = util.get_host_list(config)
    port = util.get_ldap_port(config, factory_kwargs['transport_type'])

    for host in hosts:
        factory = _ADServiceClientFactory(**factory_kwargs)
        host_result = yield _test_one_host(host, port, factory, toolbox)
        host_results.append(host_result)

    defer.returnValue(AdClientResult(config_result, host_results))


def validate_ad_client_config(config):
    """Validate ad client config by ensuring that
    necessary fields are provided
    Args:
        config: ConfigDict
    Returns:
        ConfigCheckResult with any problems
    """

    problems = []

    try:
        util.get_host_list(config)
    except ConfigError:
        problems.append(MissingConfigKeyProblem('host'))
    try:
        util.parse_ad_client(config)
    except ConfigError as e:
        problems.append(ConfigErrorProblem(e))

    return ConfigCheckResult(problems)


@defer.inlineCallbacks
def _test_one_host(host, port, factory, toolbox):

    tcp_connect_result = yield toolbox.test_connect_tcp(host, port)
    transport_type = factory.transport_type

    if not tcp_connect_result.is_successful():
        defer.returnValue(LdapHostResult(
            host,
            port,
            transport_type,
            tcp_connect_result,
            UnmetPrerequisiteSkippedTestResult('ldap connection', 'tcp connection'),
            UnmetPrerequisiteSkippedTestResult('bind', 'tcp connection'),
            UnmetPrerequisiteSkippedTestResult('search', 'tcp connection')
        ))

    try:
        connected_client = yield _connect_host(factory, host, port)
        ldap_connect_result = LdapConnectResult(True, host, port)
    except Exception as e:
        ldap_connect_result = LdapConnectResult(False, host, port, exception=e)
        defer.returnValue(LdapHostResult(
            host,
            port,
            transport_type,
            tcp_connect_result,
            ldap_connect_result,
            UnmetPrerequisiteSkippedTestResult('bind', 'ldap connection'),
            UnmetPrerequisiteSkippedTestResult('search', 'ldap connection')
        ))

    bind_result = yield _test_bind(connected_client, toolbox)
    if not bind_result.is_successful():
        defer.returnValue(LdapHostResult(
            host,
            port,
            transport_type,
            tcp_connect_result,
            ldap_connect_result,
            bind_result,
            UnmetPrerequisiteSkippedTestResult('search', 'bind')
        ))

    search_result = yield _test_search(connected_client, toolbox)

    connected_client.perform_unbind()

    defer.returnValue(LdapHostResult(
        host,
        port,
        transport_type,
        tcp_connect_result,
        ldap_connect_result,
        bind_result,
        search_result
    ))


@defer.inlineCallbacks
def _connect_host(factory, host, port):
    """Helper function that connects to the host/port
    Args:
        factory (_ADServiceClientFactory)
        host (str)
        port (int)
    Returns:
        _ADServiceClientProtocol if connection succeeded.
    Raises:
        Yield on factory.deferred raises exception if failed to connect to host
    """
    factory.connect_ldap(host, port)
    client = yield factory.deferred
    defer.returnValue(client)


@defer.inlineCallbacks
def _test_bind(client, toolbox):
    result = yield toolbox.test_bind_service_account(client, keep_bound=True)
    defer.returnValue(result)


@defer.inlineCallbacks
def _test_search(client, toolbox):
    filter_object = client.user_filter_object()
    search_result = yield toolbox.test_ldap_search_has_results(
        client,
        client.factory.search_dn,
        filter_object)
    defer.returnValue(search_result)
