#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

from twisted.internet import defer

import base_duo_check
from duoauthproxy.lib import const
from duoauthproxy.lib import util
from duoauthproxy.modules.ssl_server import ChainingOpenSSLContextFactory
from ..connectivity_toolbox import STANDARD_TOOLBOX
from ..connectivity_results import (
    LdapServerResult,
    ConfigCheckResult,
    MissingConfigKeyProblem,
    UnmetPrerequisiteSkippedTestResult,
    NOT_APPLICABLE_TEST_RESULT
)

"""Module for testing the connectivity of a user's ldap_server_auto section."""


def _call_duo(ikey, skey, api_host, ca_certs_path, toolbox):
    """Makes the /ping and /validate calls to Duo API.
    Args:
        ikey, skey, api_host (str): User's Duo creds for this server section
        ca_certs_path (str): Path to user's CA certs file
        toolbox (ConnectivityTestToolbox): to do API call tests with
    Returns:
        results (tuple) containing ping, time drift and validate call results
    """
    ping_result = base_duo_check.perform_duo_ping(toolbox, api_host, '', '', ca_certs_path)
    time_drift_result = base_duo_check.perform_time_drift(toolbox, ping_result, api_host, '', '', ca_certs_path)
    check_result = base_duo_check.perform_credentials_check(toolbox, api_host, skey, ikey, ca_certs_path)

    return (ping_result, time_drift_result, check_result)


@defer.inlineCallbacks
def _listen_ssl(port, interface, toolbox, ssl_creds):
    """Listens ephemerally to interface:port over SSL.
    Args:
        port (int): port to listen on
        interface (str): interface to listen on
        toolbox (ConnectivityTestToolbox): to do listening tests with
        ssl_creds (dict): contains SSL key/cert paths and ciphers
    Returns:
        dict: [result] True if successful, False otherwise
    """
    key_path = ssl_creds['ssl_key_path']
    cert_path = ssl_creds['ssl_cert_path']
    ciphers = ssl_creds['cipher_list']

    ssl_context_factory = ChainingOpenSSLContextFactory(
        privatekey_filename=key_path,
        certificate_filename=cert_path,
        cipher_list=ciphers,
    )
    result = yield toolbox.test_listen_ssl(port, ssl_context_factory, interface)
    defer.returnValue(result)


@defer.inlineCallbacks
def check_ldap_server_auto(config, toolbox=STANDARD_TOOLBOX):
    """Check an ldap_server_auto section config by checking:
      1) The config is valid
      2) We can /ping and /check Duo with the provided api host and credentials
      3) SSL creds are fine, if they are specified
      4) We can listen on the specified port + interface (plain TCP or SSL)
    Args:
        config (ConfigDict): config for this [ldap_server_auto] section
        toolbox (ConnectivityTestToolbox): for doing individual connectivity tests
    Returns:
        nested result (dict) of testing
    """
    config_result = validate_ldap_server_auto_config(config)

    ikey = config.get_str('ikey', '')
    skey = config.get_protected_str('skey_protected', 'skey', '')
    api_host = config.get_str('api_host', '')

    ca_certs = config.get_str('http_ca_certs_file', const.DEFAULT_HTTP_CERTS_FILE)
    ca_certs_path = util.resolve_file_path(ca_certs)

    ping_result, time_drift_result, validate_result = _call_duo(ikey, skey, api_host, ca_certs_path, toolbox)

    ssl_key_path = config.get_str('ssl_key_path', '')
    ssl_cert_path = config.get_str('ssl_cert_path', '')
    cipher_list = config.get_str('cipher_list', "DEFAULT")

    if ssl_key_path and ssl_cert_path:
        ssl_key_path = util.resolve_file_path(ssl_key_path)
        ssl_cert_path = util.resolve_file_path(ssl_cert_path)
        ssl_result = toolbox.test_ssl_credentials(
            ssl_key_path,
            ssl_cert_path,
            cipher_list
        )
        port = config.get_int('ssl_port', const.DEFAULT_LDAPS_PORT)
    else:
        ssl_result = NOT_APPLICABLE_TEST_RESULT
        port = config.get_int('port', const.DEFAULT_LDAP_PORT)

    interface = config.get_str('interface', '')

    ssl_creds = {
        'ssl_key_path': ssl_key_path,
        'ssl_cert_path': ssl_cert_path,
        'cipher_list': cipher_list,
    }

    if not (ssl_key_path and ssl_cert_path):
        listen_result = yield toolbox.test_listen_tcp(port, interface)
    elif ssl_result.is_successful():
        listen_result = yield _listen_ssl(port, interface, toolbox, ssl_creds)
    else:
        listen_result = UnmetPrerequisiteSkippedTestResult('listen', 'ssl configuration')

    result = LdapServerResult(config_result, ping_result, time_drift_result, validate_result, ssl_result, listen_result)
    defer.returnValue(result)


def validate_ldap_server_auto_config(config):
    """Validate ldap_server_auto config by ensuring the bare minimum keys are present
    (ikey, skey OR skey_protected, api_host) as well as all proper combos for optional keys.
    Args:
        config (ConfigDict): section config
    Returns:
        result (dict) of testing
    """
    problems = []

    for required_key in ['ikey', 'api_host']:
        if required_key not in config:
            problems.append(MissingConfigKeyProblem(required_key))

    if not ('skey' in config or 'skey_protected' in config):
        problems.append(MissingConfigKeyProblem('skey / skey_protected'))

    if ('ssl_key_path' in config and 'ssl_cert_path' not in config):
        problems.append(MissingConfigKeyProblem('ssl_cert_path'))

    if ('ssl_key_path' not in config and 'ssl_cert_path' in config):
        problems.append(MissingConfigKeyProblem('ssl_key_path'))

    return ConfigCheckResult(problems)
