#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

from collections import OrderedDict

from twisted.internet import defer

from connectivity_toolbox import STANDARD_TOOLBOX
from connectivity_test_resolver import STANDARD_RESOLVER
from duoauthproxy.lib.log import _sanitize_config


NO_SECTIONS_MESSAGE = "The Auth Proxy configuration has no sections - testing will be skipped."


@defer.inlineCallbacks
def test_config(full_configuration, logger, toolbox=STANDARD_TOOLBOX, section_test_resolver=STANDARD_RESOLVER):
    """
    Test the provided Auth Proxy configuration for any connectivity issues

    Args:
        full_configuration (ConfigProvider): the Auth Proxy configuration to test
        logger (Logger): the twisted Logger to output results into
        toolbox (ConnectivityTestToolbox): the test toolbox
        section_test_resolver (SectionTesterResolver): Provides the testing function for a section name

    Returns:
        OrderedDict: The assembled results of the testing keyed by section name

    """
    try:
        sections = full_configuration.get_all_sections()

        if len(sections) == 0:
            logger.warn(NO_SECTIONS_MESSAGE)
            defer.returnValue({})

        all_results = yield _test_all_sections(sections, toolbox, section_test_resolver)

        output_results(all_results, sections, logger)
        defer.returnValue(all_results)

    except Exception as error:
        logger.error("There was a problem running the connectivity tool: {error}", error=error)


@defer.inlineCallbacks
def _test_all_sections(sections, toolbox, section_test_resolver):
    """
    Run the connectivity tests on all provided sections

    Args:
        sections (OrderedDict): The map of section name -> ConfigDict for the section, for all sections present
        toolbox (ConnectivityTestToolbox): the test toolbox
        section_test_resolver (SectionTesterResolver): Provides the testing function for a section name

    Returns:
        OrderedDict: the results of testing, mapped from section_name -> ReportableResult appropriate to that section

    """
    all_results = OrderedDict()

    for section_name, section_config in sections.iteritems():
        section_tester = section_test_resolver.find_tester(section_name)
        section_result = yield _test_one_section(section_config, toolbox, section_tester)
        all_results[section_name] = section_result

    defer.returnValue(all_results)


@defer.inlineCallbacks
def _test_one_section(section_config, toolbox, section_tester):
    """
    Test one configuration section

    Args:
        section_config (ConfigDict): The section's config
        toolbox (ConnectivityTestToolbox): The toolbox for the individual tests
        section_tester (func): The function to call to test the section

    Returns:
        ReportableResult: The result of the section test

    """
    result = yield section_tester(section_config, toolbox=toolbox)
    defer.returnValue(result)


def output_results(all_results, sections, logger):
    """
    Output the results into a Logger

    Args:
        all_results (OrderedDict): map of section name -> ReportableResult subclass appropriate to the section
        logger (Logger): the twisted Logger to output into
    """
    for section_name, section_result in all_results.iteritems():
        logger.info("Testing section \'{section_name}\' with configuration:", section_name=section_name)
        logger.info("{cfg}", cfg=_sanitize_config(
            sections[section_name], lambda x: x.startswith('radius_secret') or
            x.startswith('secret') or
            x in ('skey', 'skey_protected',
                  'service_account_password', 'service_account_password_protected')))
        section_result.to_log_output(logger)
        logger.info("-----------------------------")
