import struct

from duoauthproxy.lib.config_provider import ConfigError
from . import log
from ..modules import ssl_server

from OpenSSL import SSL

from twisted.internet import defer

EAP_SESSION_START = 0
EAP_SESSION_REQUEST_PEAP = 1
EAP_SESSION_PEAP_START = 2
EAP_SESSION_PEAP_REQUEST = 3
EAP_SESSION_PEAP = 4
EAP_SESSION_PEAP_MORE = 5
EAP_SESSION_REQUEST_GTC = 6
EAP_SESSION_PEAP_GTC = 7
EAP_SESSION_ACCEPT = 8
EAP_SESSION_PEAP_ID = 9
EAP_SESSION_PEAP_ACCEPT = 10
EAP_SESSION_SUCCESS = 11
EAP_SESSION_DENY = 12
EAP_SESSION_PEAP_DENY = 13
EAP_SESSION_FAILURE = 14

EAP_FAILURE = 0
EAP_SUCCESS = 1
EAP_PASSCODE = 2

EAP_CODE_REQUEST = 1
EAP_CODE_RESPONSE = 2
EAP_CODE_SUCCESS = 3
EAP_CODE_FAILURE = 4

EAP_TYPE_IDENTITY = 1
EAP_TYPE_NOTIFICATION = 2
EAP_TYPE_NAK = 3
EAP_TYPE_GTC = 6
EAP_TYPE_PEAP = 25
EAP_TYPE_NONE = 0
EAP_TYPE_TLV = 33

MS_RESULT_HEADER = '\x80\x03'


class BreakEAP(Exception):
    # Exception thrown to break out of an EAP layer
    # Thrown on success/failure from both sessions
    def __init__(self, code, keys=None, packet=None):
        self.mppe_keys = keys
        self.packet = packet
        self.code = code


class EAPPacketException(Exception):
    pass


class EAPPacket(object):

    def __init__(self, code, ID, eap_type, data, length=0, start=False, more=False):

        self.ID = ID
        self.code = code
        self.type = eap_type
        self.data = data

        if eap_type == EAP_TYPE_IDENTITY:
            self.add_id(data)
        elif eap_type == EAP_TYPE_NOTIFICATION:
            self.add_not(data)
        elif eap_type == EAP_TYPE_NAK:
            self.add_nak(data)
        elif eap_type == EAP_TYPE_GTC:
            self.add_gtc(data)
        elif eap_type == EAP_TYPE_PEAP:
            self.add_peap(data, length, start, more)
        elif eap_type == EAP_TYPE_NONE:
            pass  # no type, eg eap-success packet
        elif eap_type == EAP_TYPE_TLV:
            # MS accept tlv
            self.add_tlv(data)
        else:
            raise EAPPacketException('Unsupported eap type %s' % (eap_type))

    def add_id(self, data):
        self.data = data

    def add_not(self, data):
        self.data = data

    def add_nak(self, data):
        try:
            self.data = chr(data)
        except (ValueError, TypeError):
            raise EAPPacketException('Invalid Nak type %s' % (data))

    def add_gtc(self, data):
        self.data = data

    def add_peap(self, data, length=0, start=False, more=False):
        ''' start should be true if first fragment
            more should be true if not last fragment
            length must be the EAP-TLS Length'''

        flags = 0
        if start:
            flags = flags | 32  # TLS-start
            if length:
                flags = flags | 128  # Length Included

        if more:
            flags = flags | 64  # More Fragments

        self.data = chr(flags)

        if length:
            self.data += struct.pack('>I', length)
        self.data += data

    def add_tlv(self, result):
        # MS Result TLV packet
        # M bit flag = 1, TLV_Type = 3, Length=2
        result = (1 if result == True else 2)
        self.data = struct.pack('BBBBBB', 128, 3, 0, 2, 0, result)

    def render(self):
        try:
            msg = chr(self.code)
            msg += chr(self.ID)
            if self.type:
                msg += struct.pack('>H', 5 + len(self.data))
                msg += chr(self.type)
            else:
                msg += struct.pack('>H', 4)
            msg += str(self.data)
            return msg
        except ValueError:
            raise EAPPacketException('Invalid EAP code, type or ID')


class EAPSession(object):
    def __init__(self, identity=None, pkey=None, certs=None,
                 inner=False, gtc_message='Enter your password',
                 parent=None, cipher_list=''):
        self.inner = inner
        self.gtc_message = gtc_message
        self.state = 0
        self.id = 0
        self.pkey = pkey
        self.certs = certs
        self.identity = identity
        self.password = None
        self.prompt = ''
        self.enrolling = False
        self.next_state = ''
        self.handshake = False
        self.parent = parent
        self.cipher_list = cipher_list

    def gtc_received(self, session, passcode, prompt):
        ''' Called when the response to a gtc prompt is received '''
        raise NotImplementedError()

    def errback(self, session, reason):
        ''' Called when the eap session fails for some reason. '''
        raise NotImplementedError()

    def success(self, session):
        ''' Called when authentication is successful '''
        raise NotImplementedError()

    def inner_errback(self, session, reason):
        return Exception(reason)

    def inner_success(self, session):
        return True

    def start_tls(self, pkey_file, certs_file, cipher_list):
        try:
            ssl_context_factory = ssl_server.ChainingOpenSSLContextFactory(
                privatekey_filename=pkey_file,
                certificate_filename=certs_file,
                cipher_list=cipher_list
            )
        except Exception as e:
            if e.message[0][1] == 'fopen':
                if e.message[2][1] == 'SSL_CTX_use_certificate_file':
                    raise ConfigError(
                        "Could not open certificate file at %s" % (certs_file))
                if e.message[2][1] == 'SSL_CTX_use_PrivateKey_file':
                    raise ConfigError(
                        "Could not open private key file at %s" % (pkey_file))
            if e.message[0][1] == 'SSL_CTX_set_cipher_list':
                raise ConfigError(
                    "Invalid OpenSSL cipher list specified %s" % (cipher_list))
            raise e

        ssl_context = ssl_context_factory.getContext()
        self.ssl_con = SSL.Connection(ssl_context, None)
        self.handshake = False

        # Create inner EAP session and let it use the implemented callbacks
        self.innerEAP = EAPSession(identity=0, inner=True, gtc_message=self.gtc_message, parent=self)
        self.innerEAP.gtc_received = self.gtc_received
        self.innerEAP.errback = self.inner_errback
        self.innerEAP.success = self.inner_success

    def get_outgoing(self):
        # Check to see if there are outgoing messages from the SSL connection
        try:
            return self.ssl_con.bio_read(4096)
        except SSL.WantReadError:
            # No outgoing messages ready
            return None

    @defer.inlineCallbacks
    def add_incoming(self, msg):
        # Add a message to the EAP-TLS connection
        self.ssl_con.bio_write(msg)
        if not self.handshake:
            try:
                self.ssl_con.set_accept_state()
                self.ssl_con.do_handshake()
            except SSL.WantReadError:
                self.handshake = True
        try:
            # If there is a decrypted message ready, push it to inner eap
            decrypted_msg = self.ssl_con.recv(9999)

            # Inner eap lacks code, id, len
            # add them on so the packet is parsed the same
            header = chr(2) + msg[1] + struct.pack('>H', len(decrypted_msg) + 4)

            if ord(decrypted_msg[0]) == 2:  
                # For some reason the MS-Auth-TLV has a header already
                header = ''
            elif ord(decrypted_msg[0]) == 6:
                # GTC response
                gtc = yield self.gtc_received(self, decrypted_msg[1:], self.gtc_message)
                defer.returnValue(gtc)

            add_response = yield self.innerEAP.add_message(header + decrypted_msg)
            if add_response:
                error = yield self.errback(self, add_response.message)
                defer.returnValue(error)

        except SSL.WantReadError:
            # No decrypted message ready yet
            pass
        except SSL.Error:
            # Possibly retransmission or some other unknown error
            ssl_error = 'SSL Error'
            log.auth_standard(msg=ssl_error,
                              username=self.current_request.username,
                              auth_stage=log.AUTH_UNKNOWN,
                              status=log.AUTH_ERROR,
                              client_ip=self.current_request.client_ip,
                              server_section=self.server.server_section_name)
            error = yield self.errback(self, ssl_error)
            defer.returnValue(error)

        defer.returnValue(False)

    @defer.inlineCallbacks
    def next_message(self):
        ''' Get the next outgoing message from the EAP session.
        Returns a string containing the message if one exists.
        On failure or successful authentication, returns
        errback or success '''

        if self.state == EAP_SESSION_START:
            # Request Identity
            packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_IDENTITY, chr(0))

        elif self.state == EAP_SESSION_REQUEST_PEAP:
            # Send a Nak PEAP request
            packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_NAK, chr(EAP_TYPE_PEAP))

        elif self.state == EAP_SESSION_PEAP_REQUEST:
            # Start a TLS session and send TLS-start packet
            self.start_tls(self.pkey, self.certs, self.cipher_list)
            packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_PEAP, '', start=True)

        elif self.state == EAP_SESSION_PEAP:
            # Currently in a PEAP session

            # Check to see if the TLS connection has outgoing messages
            outgoing_msg = self.get_outgoing()
            if outgoing_msg:
                packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_PEAP, outgoing_msg)

            else:
                # The tls connection has nothing to send
                # Get a message from inner EAP and feed it to tls
                # Connection then retry to find something to send
                inner_msg = yield self.innerEAP.next_message()

                # Upon error or success, next_message will yield
                #  an exception or a boolean respectively
                if isinstance(inner_msg, basestring):
                    self.ssl_con.send(inner_msg)
                    msg = yield self.next_message()
                    defer.returnValue(msg)
                elif isinstance(inner_msg, Exception):
                    log.auth_standard(msg=inner_msg.message,
                                      username=self.current_request.username,
                                      auth_stage=log.AUTH_UNKNOWN,
                                      status=log.AUTH_ERROR,
                                      client_ip=self.current_request.client_ip,
                                      server_section=self.server.server_section_name)
                    error = yield self.errback(self, inner_msg.message)
                    defer.returnValue(error)
                elif inner_msg == True:
                    success = yield self.success(self)
                    defer.returnValue(success)
                else:
                    peap_error_message = "Error in PEAP session"
                    log.auth_standard(msg=peap_error_message,
                                      username=self.current_request.username,
                                      auth_stage=log.AUTH_UNKNOWN,
                                      status=log.AUTH_ERROR,
                                      client_ip=self.current_request.client_ip,
                                      server_section=self.server.server_section_name)
                    error = self.errback(self, peap_error_message)
                    defer.returnValue(error)

        elif self.state == EAP_SESSION_REQUEST_GTC:
            # Send a password (GTC) prompt
            packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_GTC, self.gtc_message)

        elif self.state == EAP_SESSION_PEAP_GTC:
            # Have inner EAP send GTC prompt
            self.innerEAP.state = EAP_SESSION_REQUEST_GTC
            self.state = EAP_SESSION_PEAP
            msg = yield self.next_message()
            defer.returnValue(msg)

        elif self.state == EAP_SESSION_PEAP_ID:
            # Have inner EAP request ID
            self.innerEAP.state = EAP_SESSION_START
            self.state = EAP_SESSION_PEAP
            msg = yield self.next_message()
            defer.returnValue(msg)

        elif self.state == EAP_SESSION_PEAP_ACCEPT:
            # Save the TLS key/randoms for generating MPPE keys
            self.mk = self.ssl_con.master_key()
            self.cr = self.ssl_con.client_random()
            self.sr = self.ssl_con.server_random()

            # Have inner EAP send accept message
            self.innerEAP.state = EAP_SESSION_ACCEPT
            self.state = EAP_SESSION_PEAP
            msg = yield self.next_message()
            defer.returnValue(msg)

        elif self.state == EAP_SESSION_PEAP_DENY:
            # Have inner EAP send reject message
            self.innerEAP.state = EAP_SESSION_DENY
            self.state = EAP_SESSION_PEAP
            msg = yield self.next_message()
            defer.returnValue(msg)

        elif self.state == EAP_SESSION_ACCEPT:
            # Send accept packet
            packet = EAPPacket(EAP_CODE_REQUEST, self.id, EAP_TYPE_TLV, True)

        elif self.state == EAP_SESSION_DENY:
            error = yield self.errback(self, 'Authentication denied.')
            defer.returnValue(error)

        elif self.state == EAP_SESSION_FAILURE:
            error = yield self.errback(self, 'EAP session failed.')
            defer.returnValue(error)

        elif self.state == EAP_SESSION_SUCCESS:
            success = yield self.success(self)
            defer.returnValue(success)

        if self.inner:
            defer.returnValue(packet.render()[4:])  # tunneled EAP skips code, id, length
        defer.returnValue(packet.render())

    @defer.inlineCallbacks
    def add_message(self, msg):
        ''' Add a message to the EAP session. Returns False
        on success, returns errback on error, or gtc_received
        on receiving a token card response. '''

        code, identifier, length, eap_type = struct.unpack('>BBHB', msg[:5])
        if not self.inner and self.id != 0 and identifier != self.id:
            # Silently discard response if ID does not match
            log.msg("Response ID %s does not match request ID %s" % (identifier, self.id))
            defer.returnValue(False)
        if len(msg) != length:
            message = 'Malformed EAP packet: length ({0}) does not match actual length ({1})'.format(length, len(msg))
            log.auth_standard(msg=message,
                              username=self.current_request.username,
                              auth_stage=log.AUTH_UNKNOWN,
                              status=log.AUTH_ERROR,
                              client_ip=self.current_request.client_ip,
                              server_section=self.server.server_section_name)
            error = yield self.errback(self, message)
            defer.returnValue(error)

        if code == EAP_CODE_REQUEST:
            # EAP request, we are talking to RADIUS, this should not happen
            message = "EAP request received."
            log.auth_standard(msg=message,
                              username=self.current_request.username,
                              auth_stage=log.AUTH_UNKNOWN,
                              status=log.AUTH_ERROR,
                              client_ip=self.current_request.client_ip,
                              server_section=self.server.server_section_name)
            error = yield self.errback(self, message)
            defer.returnValue(error)
        elif code == EAP_CODE_RESPONSE:
            # EAP response, we are talking to NetMotion
            self.id += 1
            if eap_type == EAP_TYPE_IDENTITY:  # Identity
                self.state = EAP_SESSION_REQUEST_GTC
                identity = msg[5:]
            elif eap_type == EAP_TYPE_NAK:  # Legacy Nak
                desired_type = ord(msg[5])
                if desired_type == EAP_TYPE_PEAP:  # PEAP request
                    self.state = EAP_SESSION_PEAP_REQUEST
                else:
                    message = "Unknown Nak request %s." % desired_type
                    log.auth_standard(msg=message,
                                      username=self.current_request.username,
                                      auth_stage=log.AUTH_UNKNOWN,
                                      status=log.AUTH_ERROR,
                                      client_ip=self.current_request.client_ip,
                                      server_section=self.server.server_section_name)
                    error = yield self.errback(self, message)
                    defer.returnValue(error)
            elif eap_type == EAP_TYPE_TLV:
                if msg[5:7] == MS_RESULT_HEADER and len(msg) == 11:  # MS-Result-TLV
                    if ord(msg[-1]) == 1 and self.state == EAP_SESSION_ACCEPT:  # Success
                        self.state = EAP_SESSION_SUCCESS
                    else:  # Failure
                        self.state = EAP_SESSION_FAILURE
                else:
                    # No support for cryptobinding or SoH TLV
                    # At least exit with some relevant info
                    message = "Unsupported TLV: {0}.".format(msg)
                    log.auth_standard(msg=message,
                                      username=self.current_request.username,
                                      auth_stage=log.AUTH_UNKNOWN,
                                      status=log.AUTH_ERROR,
                                      client_ip=self.current_request.client_ip,
                                      server_section=self.server.server_section_name)
                    error = yield self.errback(self, message)
                    defer.returnValue(error)

            elif eap_type == EAP_TYPE_PEAP:  # PEAP
                tls_flags = ord(msg[5])
                if tls_flags & 0x80:  # Length included
                    self.tls_len = struct.unpack('>I', msg[6:10])
                    tls_msg = msg[10:]
                else:
                    tls_msg = msg[6:]

                if self.state == EAP_SESSION_PEAP_MORE:
                    self.tls_msg += tls_msg
                else:
                    self.tls_msg = tls_msg

                if tls_flags & 0x40:  # More fragments
                    self.state = EAP_SESSION_PEAP_MORE
                else:
                    self.state = EAP_SESSION_PEAP
                    if self.tls_msg:
                        if hasattr(self, 'ssl_con'):
                            add_response = yield self.add_incoming(tls_msg)
                            if add_response:
                                # Errback or gtc callback occured, pass it through
                                defer.returnValue(add_response)
                        else:
                            # We reset the EAP session but the peer is trying to continue
                            self.state = EAP_SESSION_PEAP_REQUEST
                    else:
                        # Empty PEAP message, probably handshake over
                        self.state = EAP_SESSION_PEAP_ID
        defer.returnValue(False)
