import pprint
import json
from datetime import datetime

from twisted.python import log
from twisted.logger import PredicateResult
from twisted.logger import Logger


AUTH_ALLOW = 'Allow'
AUTH_REJECT = 'Reject'
AUTH_ERROR = 'Error'
AUTH_PRIMARY = 'Primary authentication'
AUTH_SECONDARY = 'Secondary authentication'
AUTH_UNKNOWN = 'Unknown'
AUTH_ENROLL_MSG = 'User is not enrolled. Enrollment link sent.'

MSG_TYPE_AUTH = "AUTH"
MSG_TYPE_KEY = "msg_type"
AUTH_VALUES_KEY = "auth_values"
CONN_LOGGER = 'connectivity'


class BufferingLogger(object):
    """ Simple class that will intercept and buffer calls to twisted logging
    facilities until told it is ready - at which point, it will replay them
    all """

    def __init__(self, logger=log):
        self._calls = []
        self._ready = False
        self._log = logger

    def ready(self):
        for call in self._calls:
            call[0](*call[1], **call[2])
        self._calls = []
        self._ready = True

    def msg(self, *args, **kwargs):
        if not self._ready:
            self._calls.append((self._log.msg, args, kwargs))
        else:
            self._log.msg(*args, **kwargs)

    def config(self, config, scrub=lambda x: False):
        result = _sanitize_config(config, scrub)

        if not self._ready:
            self._calls.append((self._log.msg, [result], {}))
        else:
            self._log.msg(result)

    def err(self, *args, **kwargs):
        if not self._ready:
            self._calls.append((self._log.err, args, kwargs))
        else:
            self._log.err(*args, **kwargs)

    def auth(self, **kwargs):
        ts = datetime.utcnow().isoformat() + 'Z'
        arguments = {'timestamp': ts}
        arguments.update(kwargs)
        msg_type = MSG_TYPE_AUTH
        if not self._ready:
            self._calls.append((self._log.msg, [''], {MSG_TYPE_KEY: msg_type, AUTH_VALUES_KEY: arguments}))
        else:
            self._log.msg('', msg_type=msg_type, auth_values=arguments)

    def auth_standard(self, msg, username, auth_stage, status, client_ip, server_section):
        self.auth(
            msg=msg,
            username=username,
            auth_stage=auth_stage,
            status=status,
            client_ip=client_ip,
            server_section=server_section
        )


def auth_type_predicate(event):
    """
    Implementor of the predicate interface used in conjunction with the FilteringLogObserver in proxy.py.

    This should detect events that should be directed to the auth log file.  Make sure the criteria here and in the
    log.auth method above match.
    """
    if event.get(MSG_TYPE_KEY) == MSG_TYPE_AUTH:
        return PredicateResult.yes
    else:
        return PredicateResult.no


def no_op_observer(event):
    """
    An ILogObserver implementation that simply ignores the event

    :param event: The event to ignore
    """
    pass


def format_auth_event(event):
    """
    Event formatter used to prepare a string suitable to go to the auth log file.

    The event will have properties based on the keyword arguments generated by the log.auth method above

    :param event: a twisted log event generated by the logging framework
    :return: a unicode string suitable for inclusion in the auth log file
    """
    auth_values = event.get(AUTH_VALUES_KEY)
    return json.dumps(auth_values) + '\n'


def _sanitize_config(config, scrub=lambda x: False):
    """
    Sanitize config is used to take a config and remove sensitive fields

    Args:
        config: (dict) an auth proxy config
        scrub: (lambda) lambda that evalulates which fields to scrub

    Returns:
        str: a pretty print string of the sanitized config
    """
    sanitized_config = dict(config)

    def sanitize(key, value):
        sanitized_keys = [
            'skey',
        ]

        if scrub(key):
            if key in sanitized_keys:
                sanitized_length = len(value)
                return '*****[{0}]'.format(sanitized_length)
            else:
                return '*****'

        return value

    sanitized_config = dict(
        (key, sanitize(key, value))
        for key, value in sanitized_config.items()
    )

    return pprint.pformat(sanitized_config)


class LoggerContainer(object):
    def __init__(self):
        self.loggers = {}

    def get_logger(self, key):
        """Return the appropriate logger object based on the type
        Args:
            key (str): Identifier for the logger
        Returns:
            Logger
        """
        return self.loggers[key]

    def add_logger(self, key, observer=None):
        """Add a twisted Logger to the dictionary
        Args:
            key (str): Identifier for the logger
            observer (ILogObserver): Optional observer to hook up. Will use globalLogPublisher otherwise
        """
        if observer:
            logger = Logger(observer=observer)
        else:
            logger = Logger()
        self.loggers[key] = logger


_instance = BufferingLogger()

ready = _instance.ready
msg = _instance.msg
err = _instance.err
config = _instance.config
auth = _instance.auth
auth_standard = _instance.auth_standard

logger_container = LoggerContainer()
