#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import hashlib
import hmac
import struct
import datetime
import os
import socket
from dpkt import Packet
from Crypto.Cipher import DES, ARC4

# Negotiate Flags
NTLMSSP_NEGOTIATE_56 = 0x80000000
NTLMSSP_NEGOTIATE_KEY_EXCH = 0x40000000
NTLMSSP_NEGOTIATE_128 = 0x20000000
NTLMSSP_R1 = 0x10000000
NTLMSSP_R2 = 0x08000000
NTLMSSP_R3 = 0x04000000
NTLMSSP_NEGOTIATE_VERSION = 0x02000000
NTLMSSP_R4 = 0x01000000
NTLMSSP_NEGOTIATE_TARGET_INFO = 0x00800000
NTLMSSP_REQUEST_NON_NT_SESSION_KEY = 0x00400000
NTLMSSP_R5 = 0x00200000
NTLMSSP_NEGOTIATE_IDENTIFY = 0x00100000
NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY = 0x00080000
NTLMSSP_R6 = 0x00040000
NTLMSSP_TARGET_TYPE_SERVER = 0x00020000
NTLMSSP_TARGET_TYPE_DOMAIN = 0x00010000
NTLMSSP_NEGOTIATE_ALWAYS_SIGN = 0x00008000
NTLMSSP_R7 = 0x00004000
NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED = 0x00002000
NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED = 0x00001000
NTLMSSP_ANONYMOUS = 0x00000800
NTLMSSP_R8 = 0x00000400
NTLMSSP_NEGOTIATE_NTLM = 0x00000200
NTLMSSP_R9 = 0x00000100
NTLMSSP_NEGOTIATE_LM_KEY = 0x00000080
NTLMSSP_NEGOTIATE_DATAGRAM = 0x00000040
NTLMSSP_NEGOTIATE_SEAL = 0x00000020
NTLMSSP_NEGOTIATE_SIGN = 0x00000010
NTLMSSP_R10 = 0x00000008
NTLMSSP_REQUEST_TARGET = 0x00000004
NTLMSSP_NEGOTIATE_OEM = 0x00000002
NTLMSSP_NEGOTIATE_UNICODE = 0x00000001


class FieldDescriptor(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('Len', 'H', 0),
        ('MaxLen', 'H', 0),
        ('BufferOffset', 'L', 0)
    )

    def set(self, position, value):
        """Set the contents of the field descriptor for a given
        field value, starting at the given buffer position. return
        the buffer position to be used by the next field"""
        self.Len = self.MaxLen = len(value)
        self.BufferOffset = position
        return position + self.Len

    def get_field(self, buf):
        return buf[self.BufferOffset:self.BufferOffset + self.Len]

_DefaultFieldDescriptor = FieldDescriptor()


class NTLMVersion(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('ProductMajorVersion', 'B', 6),
        ('ProductMinorVersion', 'B', 1),
        ('ProductBuild', 'H', 7600),
        ('_Reserved1', 'H', 0),
        ('_Reserved2', 'B', 0),
        ('NTLMRevisionCurrent', 'B', 0x0F),
    )

_DefaultNTLMVersion = NTLMVersion()


class NTLMMessage(Packet):
    _payload_fields = ()

    def __len__(self):
        ret = self.__hdr_len__
        for (field_name, desc_name) in self._payload_fields:
            ret += len(getattr(self, field_name))
        return ret

    def __str__(self):
        buffer = ''
        buffer_pos = self.__hdr_len__
        for (field_name, desc_name) in self._payload_fields:
            field = getattr(self, field_name)
            desc = FieldDescriptor()
            buffer_pos = desc.set(buffer_pos, field)
            buffer += field
            setattr(self, desc_name, str(desc))
        buffer = self.pack_hdr() + buffer
        return buffer

    def unpack(self, buf):
        super(NTLMMessage, self).unpack(buf)
        for (field_name, desc_name) in self._payload_fields:
            desc = FieldDescriptor(getattr(self, desc_name))
            setattr(self, field_name, desc.get_field(buf))


class NegotiateMessage(NTLMMessage):
    __byte_order__ = '<'
    __hdr__ = (
        ('Signature', '8s', 'NTLMSSP\0'),
        ('MessageType', 'L', 1),
        ('NegotiateFlags', 'L', 0),
        ('_DomainNameFields', '8s', str(_DefaultFieldDescriptor)),
        ('_WorkstationFields', '8s', str(_DefaultFieldDescriptor)),
        ('_Version', '8s', str(_DefaultNTLMVersion)),
    )
    DomainName = ''
    Workstation = ''
    _payload_fields = (
        ('DomainName', '_DomainNameFields'),
        ('Workstation', '_WorkstationFields')
    )


class ChallengeMessage(NTLMMessage):
    __byte_order__ = '<'
    __hdr__ = (
        ('Signature', '8s', 'NTLMSSP\0'),
        ('MessageType', 'L', 2),
        ('_TargetNameFields', '8s', str(_DefaultFieldDescriptor)),
        ('NegotiateFlags', 'L', 0),
        ('ServerChallenge', '8s', '\0' * 8),
        ('_Reserved1', '8s', '\0' * 8),
        ('_TargetInfoFields', '8s', str(_DefaultFieldDescriptor)),
        ('_Version', '8s', str(_DefaultNTLMVersion)),
    )
    TargetName = ''
    RawTargetInfo = ''
    _payload_fields = (('TargetName', '_TargetNameFields'),
                       ('RawTargetInfo', '_TargetInfoFields'))
    TargetInfo = []

    def unpack(self, buf):
        super(ChallengeMessage, self).unpack(buf)
        if self.RawTargetInfo:
            self.TargetInfo = AVPair.unpack_pairs(self.RawTargetInfo)

    def __str__(self):
        self.RawTargetInfo = AVPair.pack_pairs(self.TargetInfo)
        return super(ChallengeMessage, self).__str__()


class AuthenticateMessage(NTLMMessage):
    __byte_order__ = '<'
    __hdr__ = (
        ('Signature', '8s', 'NTLMSSP\0'),
        ('MessageType', 'L', 3),
        ('_LmChallengeResponseFields', '8s', str(_DefaultFieldDescriptor)),
        ('_NtChallengeResponseFields', '8s', str(_DefaultFieldDescriptor)),
        ('_DomainNameFields', '8s', str(_DefaultFieldDescriptor)),
        ('_UserNameFields', '8s', str(_DefaultFieldDescriptor)),
        ('_WorkstationFields', '8s', str(_DefaultFieldDescriptor)),
        ('_EncryptedRandomSessionKeyFields', '8s', str(_DefaultFieldDescriptor)),
        ('NegotiateFlags', 'L', 0),
        ('Version', '8s', str(_DefaultNTLMVersion)),
        ('MIC', '16s', '\0' * 16),
    )

    LmChallengeResponse = ''
    NtChallengeResponse = ''
    DomainName = ''
    UserName = ''
    Workstation = ''
    EncryptedRandomSessionKey = ''
    _payload_fields = (('DomainName', '_DomainNameFields'),
                       ('UserName', '_UserNameFields'),
                       ('Workstation', '_WorkstationFields'),
                       ('LmChallengeResponse', '_LmChallengeResponseFields'),
                       ('NtChallengeResponse', '_NtChallengeResponseFields'),
                       ('EncryptedRandomSessionKey', '_EncryptedRandomSessionKeyFields'))


class AVPair(Packet):
    AVID_MsvAvEOL = 0
    AVID_MsvAvNbComputerName = 1
    AVID_MsvAvNbDomainName = 2
    AVID_MsvAvDnsComputerName = 3
    AVID_MsvAvDnsDomainName = 4
    AVID_MsvAvDnsTreeName = 5
    AVID_MsvAvFlags = 6
    AVID_MsvAvTimestamp = 7
    AVID_MsAvRestrictions = 8
    AVID_MsvAvTargetName = 9
    AVID_MsvChannelBindings = 10

    __byte_order__ = '<'
    __hdr__ = (
        ('AvId', 'H', 0),
        ('AvLen', 'H', 0)
    )
    Value = ''

    @classmethod
    def unpack_pairs(klass, buf):
        pairs = []
        while True:
            pair = klass(buf)
            pair.data = pair.data[:len(pair)]
            pairs.append(pair)
            if pair.AvId == klass.AVID_MsvAvEOL:
                break
            buf = buf[len(pair):]
        return pairs

    @staticmethod
    def pack_pairs(pairs):
        return ''.join([str(pair) for pair in pairs])

    def __len__(self):
        return self.__hdr_len__ + len(self.Value)

    def __str__(self):
        self.AvLen = len(self.Value)
        return self.pack_hdr() + self.Value

    def unpack(self, buf):
        super(AVPair, self).unpack(buf)
        self.Value = buf[self.__hdr_len__: self.__hdr_len__ + self.AvLen]


class LMResponse(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('Response', '24s', '\0' * 24),
    )


class LMv2Response(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('Response', '16s', '\0' * 16),
        ('ChallengeFromClient', '8s', '\0' * 16)
    )


class NTLMResponse(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('Response', '24s', '\0' * 24),
    )


class NTLMv2ClientChallenge(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('RespType', 'B', 1),
        ('HiRespType', 'B', 1),
        ('_Reserved1', 'H', 0),
        ('_Reserved2', 'L', 0),
        ('TimeStamp', 'Q', 0),
        ('ChallengeFromClient', '8s', '\0' * 8),
        ('_Reserved3', 'L', 0),
    )

    AVPairs = []

    def __str__(self):
        return self.pack_hdr() + AVPair.pack_pairs(self.AVPairs)

    def unpack(self, buf):
        super(NTLMv2ClientChallenge, self).unpack(buf)
        self.AVPairs = AVPair.unpack_pairs(buf[self.__hdr_len__:])


class NTLMv2Response(Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('Response', '16s', '\0' * 24),
    )

    ClientChallenge = ''

    def __str__(self):
        return self.pack_hdr() + self.ClientChallenge

    def unpack(self, buf):
        super(NTLMv2Response, self).unpack(buf)
        self.ClientChallenge = buf[self.__hdr_len__:]


class NTLMSSPMessageSignature:
    __byte_order__ = '<'
    __hdr__ = (
        ('Version', 'L', 1),
        ('RandomPad', '4s', '\0' * 4),
        ('Checksum', '4s', '\0' * 4),
        ('SeqNum', 'L', 0)
    )


class NTLMSSPMessageSignature_Extended:
    __byte_order__ = '<'
    __hdr__ = (
        ('Version', 'L', 1),
        ('Checksum', '8s', '\0' * 8),
        ('SeqNum', 'L', 0)
    )


def create_negotiate_msg(domain='', workstation=''):
    flags = (NTLMSSP_NEGOTIATE_VERSION |
             NTLMSSP_REQUEST_TARGET |
             NTLMSSP_NEGOTIATE_NTLM |
             NTLMSSP_NEGOTIATE_ALWAYS_SIGN |
             NTLMSSP_NEGOTIATE_UNICODE |
             NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY)

    if domain:
        flags |= NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED
    if workstation:
        flags |= NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED

    msg = NegotiateMessage()
    msg.NegotiateFlags = flags
    msg.DomainName = domain.encode('utf-16le')
    msg.Workstation = workstation.encode('utf-16le')
    return str(msg)


def create_authenticate_msg(negotiate_buffer, challenge_buffer, username,
                            password, domain='', workstation='', ntlm_version=1,
                            client_challenge=None, session_key=None):

    challenge = ChallengeMessage(challenge_buffer)

    ClientChallenge = client_challenge or os.urandom(8)
    Time = get_ntlm_time()
    for pair in challenge.TargetInfo:
        # If NTLM v2 authentication is used, the client SHOULD send
        # the timestamp in the CHALLENGE_MESSAGE. [MS-NLMP] p.45
        if (ntlm_version == 2) and (pair.AvId == AVPair.AVID_MsvAvTimestamp):
            Time = struct.unpack('<Q', pair.Value)

        # If a domain was not specified, attempt to infer it from TargetInfo
        if (not domain) and (pair.AvId == AVPair.AVID_MsvAvNbDomainName):
            domain = pair.Value

    # if a workstation was not specified, infer it with gethostname
    if not workstation:
        workstation = socket.gethostname()

    workstation = workstation.upper().encode('ascii')
    domain = domain.upper().encode('ascii')

    TargetInfo = challenge.TargetInfo

    if ntlm_version == 1:
        ResponseKeyLM = LMOWFv1(password, username, domain)
        ResponseKeyNT = NTOWFv1(password, username, domain)
        response_vals = ComputeResponse_v1(challenge.NegotiateFlags, ResponseKeyNT, ResponseKeyLM,
                                           challenge.ServerChallenge, ClientChallenge, Time,
                                           TargetInfo)
    elif ntlm_version == 2:
        ResponseKeyLM = ResponseKeyNT = NTOWFv2(password, username, domain)
        response_vals = ComputeResponse_v2(challenge.NegotiateFlags, ResponseKeyNT, ResponseKeyLM,
                                           challenge.ServerChallenge, ClientChallenge, Time,
                                           TargetInfo)
    else:
        raise ValueError('Invalid NTLM version specified')

    msg = AuthenticateMessage()
    msg.LmChallengeResponse = response_vals[0]
    if ntlm_version == 2 and challenge.RawTargetInfo:
        # If NTLM v2 authentication is used and the CHALLENGE_MESSAGE
        # contains a TargetInfo field, the client SHOULD NOT send the
        # LmChallengeResponse and SHOULD set the
        # LmChallengeResponseLen and LmChallengeResponseMaxLen fields
        # in the AUTHENTICATE_MESSAGE to zero. [MS-NLMP] p.45

        msg.LmChallengeResponse = '\0' * 24
    msg.NtChallengeResponse = response_vals[1]
    msg.UserName = username
    msg.DomainName = domain
    msg.Workstation = workstation
    if challenge.NegotiateFlags & NTLMSSP_NEGOTIATE_UNICODE:
        msg.UserName = msg.UserName.encode('utf-16le')
        msg.DomainName = msg.DomainName.encode('utf-16le')
        msg.Workstation = msg.Workstation.encode('utf-16le')

    msg.NegotiateFlags = (NTLMSSP_NEGOTIATE_VERSION |
                          NTLMSSP_REQUEST_TARGET |
                          NTLMSSP_NEGOTIATE_NTLM |
                          NTLMSSP_NEGOTIATE_ALWAYS_SIGN |
                          NTLMSSP_NEGOTIATE_UNICODE |
                          NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY)

    if ntlm_version == 2:
        # Compute MIC, currently for NTLMv2 only
        msg.NegotiateFlags |= NTLMSSP_NEGOTIATE_TARGET_INFO

        KeyExchangeKey = KXKEY_v2(response_vals[2], response_vals[0], challenge.ServerChallenge)
        if not challenge.NegotiateFlags & NTLMSSP_NEGOTIATE_KEY_EXCH:
            ExportedSessionKey = KeyExchangeKey
        else:
            ExportedSessionKey = session_key or os.urandom(16)
            cipher = ARC4.new(KeyExchangeKey)
            msg.EncryptedRandomSessionKey = cipher.encrypt(ExportedSessionKey)

        msg.MIC = hmac.new(ExportedSessionKey, negotiate_buffer + challenge_buffer + str(msg), hashlib.md5).digest()

    return str(msg)


def KXKEY_v2(SessionBaseKey, LmChallengeResponse, ServerChallenge):
    return SessionBaseKey


def get_ntlm_time():
    ntlm_epoch = datetime.datetime(1601, 1, 1, 0, 0, 0)
    td = datetime.datetime.utcnow() - ntlm_epoch
    ntlm_time = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) * 10
    return ntlm_time


def NTOWFv1(password, user, domain):
    digest = hashlib.new('md4', password.encode('utf-16le')).digest()
    return digest


def LMOWFv1(password, user, domain):
    # make password 14 bytes long, upper-case
    lm_password = password.upper()
    lm_password = lm_password + '\0' * (14 - len(lm_password))

    cipher = DES.new(create_des_key(lm_password[0:7]), DES.MODE_ECB)
    res = cipher.encrypt('KGS!@#$%')
    cipher = DES.new(create_des_key(lm_password[7:14]), DES.MODE_ECB)
    res += cipher.encrypt('KGS!@#$%')

    return res


def DESL(k, d):
    cipher = DES.new(create_des_key(k[0:7]), DES.MODE_ECB)
    res = cipher.encrypt(d)
    cipher = DES.new(create_des_key(k[7:14]), DES.MODE_ECB)
    res += cipher.encrypt(d)
    cipher = DES.new(create_des_key(k[14:16] + '\0' * 5), DES.MODE_ECB)
    res += cipher.encrypt(d)
    return res


def ComputeResponse_v1(NegotiateFlags, ResponseKeyNT, ResponseKeyLM,
                       ServerChallenge, ClientChallenge, Time, TargetInfo):
    if (0):  
        LmChallengeResponse = DESL(ResponseKeyLM, ServerChallenge)
        NtChallengeResponse = ''
    elif (NegotiateFlags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY):
        challenge_digest = hashlib.new('md5', ServerChallenge + ClientChallenge).digest()
        NtChallengeResponse = DESL(ResponseKeyNT, challenge_digest[0:8])
        LmChallengeResponse = ClientChallenge + '\0' * 16
    else:
        NtChallengeResponse = DESL(ResponseKeyNT, ServerChallenge)
        if (0):  
            LmChallengeResponse = NtChallengeResponse
        else:
            LmChallengeResponse = DESL(ResponseKeyLM, ServerChallenge)
    SessionBaseKey = hashlib.new('md4', ResponseKeyNT).digest()
    return (LmChallengeResponse, NtChallengeResponse, SessionBaseKey)


def NTOWFv2(password, user, domain):
    key = hashlib.new('md4', password.encode('utf-16le')).digest()
    return hmac.new(key, (user.upper() + domain).encode('utf-16le'), hashlib.md5).digest()


def ComputeResponse_v2(NegotiateFlags, ResponseKeyNT, ResponseKeyLM,
                       ServerChallenge, ClientChallenge, Time, TargetInfo):
    challenge_struct = NTLMv2ClientChallenge(TimeStamp=Time, ChallengeFromClient=ClientChallenge,
                                             AVPairs=TargetInfo)
    temp = str(challenge_struct) + '\0' * 4
    NtProofStr = hmac.new(ResponseKeyNT, ServerChallenge + temp, hashlib.md5).digest()
    NtChallengeResponse = NtProofStr + temp
    LmChallengeResponse = (hmac.new(ResponseKeyLM, ServerChallenge + ClientChallenge, hashlib.md5).digest()
                           + ClientChallenge)
    SessionBaseKey = hmac.new(ResponseKeyNT, NtProofStr, hashlib.md5).digest()
    return (LmChallengeResponse, NtChallengeResponse, SessionBaseKey)


def create_des_key(key):
    """Convert 56-bit des key to 64-bit des key"""
    key_int = struct.unpack('>Q', '\0' + key)[0]
    key_int_new = 0
    for i in range(8):
        temp = (key_int & 0x7F) << 1
        temp |= ((count_ones(temp) ^ 0x1) & 0x1)
        key_int_new = (key_int_new << 8) | temp
        key_int >>= 7
    return struct.pack('<Q', key_int_new)


def count_ones(val):
    count = 0
    while(val):
        count += 1
        val &= (val - 1)
    return count

if __name__ == '__main__':
    User = 'User'
    Domain = 'Domain'
    Password = 'Password'
    Workstation = 'Workstation'.encode('utf-16le')
    RandomSessionKey = 'U' * 16
    Time = 0
    ClientChallenge = 'aa'.decode('hex') * 8
    ServerChallenge = '0123456789abcdef'.decode('hex')

    rkey_lmv1 = LMOWFv1(Password, User, Domain)
    rkey_ntv1 = NTOWFv1(Password, User, Domain)
    assert(rkey_lmv1 == 'e52cac67419a9a224a3b108f3fa6cb6d'.decode('hex'))
    assert(rkey_ntv1 == 'a4f49c406510bdcab6824ee7c30fd852'.decode('hex'))
    response_vals = ComputeResponse_v1(0, rkey_ntv1, rkey_lmv1, ServerChallenge,
                                       ClientChallenge, Time, None)
    assert(response_vals[0] == '98def7b87f88aa5dafe2df779688a172def11c7d5ccdef13'.decode('hex'))
    assert(response_vals[1] == '67c43011f30298a2ad35ece64f16331c44bdbed927841f94'.decode('hex'))

    response_vals = ComputeResponse_v1(NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY, rkey_ntv1, rkey_lmv1,
                                       ServerChallenge, ClientChallenge, Time, None)
    assert(response_vals[0] == 'aaaaaaaaaaaaaaaa00000000000000000000000000000000'.decode('hex'))
    assert(response_vals[1] == '7537f803ae367128ca458204bde7caf81e97ed2683267232'.decode('hex'))
    assert(response_vals[2] == 'd87262b0cde4b1cb7499becccdf10784'.decode('hex'))

    challenge_msg_v2 = ("4e544c4d53535000020000000c000c00"
                        "3800000033828ae20123456789abcdef"
                        "00000000000000002400240044000000"
                        "060070170000000f5300650072007600"
                        "6500720002000c0044006f006d006100"
                        "69006e0001000c005300650072007600"
                        "6500720000000000").decode('hex')
    parsed_challenge_v2 = ChallengeMessage(challenge_msg_v2)

    auth_msg_v2 = ("4e544c4d535350000300000018001800"
                   "6c00000054005400840000000c000c00"
                   "48000000080008005400000010001000"
                   "5c00000010001000d8000000358288e2"
                   "0501280a0000000f44006f006d006100"
                   "69006e00550073006500720043004f00"
                   "4d005000550054004500520086c35097"
                   "ac9cec102554764a57cccc19aaaaaaaa"
                   "aaaaaaaa68cd0ab851e51c96aabc927b"
                   "ebef6a1c010100000000000000000000"
                   "00000000aaaaaaaaaaaaaaaa00000000"
                   "02000c0044006f006d00610069006e00"
                   "01000c00530065007200760065007200"
                   "0000000000000000c5dad2544fc97990"
                   "94ce1ce90bc9d03e").decode('hex')
    parsed_auth_v2 = AuthenticateMessage(auth_msg_v2)

    rkey_lmv2 = rkey_ntv2 = NTOWFv2(Password, User, Domain)
    assert(rkey_lmv2 == '0c868a403bfd7a93a3001ef22ef02e3f'.decode('hex'))
    response_vals = ComputeResponse_v2(0, rkey_ntv2, rkey_lmv2, ServerChallenge, ClientChallenge,
                                       Time, parsed_challenge_v2.TargetInfo)
    assert(response_vals[0] == '86c35097ac9cec102554764a57cccc19aaaaaaaaaaaaaaaa'.decode('hex'))
    assert(response_vals[1] == '68cd0ab851e51c96aabc927bebef6a1c01010000000000000000000000000000aaaaaaaaaaaaaaaa0000000002000c0044006f006d00610069006e0001000c005300650072007600650072000000000000000000'.decode('hex'))
    assert(response_vals[2] == '8de40ccadbc14a82f15cb0ad0de95ca3'.decode('hex'))
