#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#
# taken mostly intact from
# http://stackoverflow.com/questions/463832/using-dpapi-with-python
#
"""Win32-specific implementation of data protection, isolating all
win32-ness.
You probably don't want to import this directly, but use the protect module
instead, which will use this implementation when running in a
win32 environment."""

import base64

from ctypes import *
from ctypes.wintypes import DWORD

LocalFree = windll.kernel32.LocalFree
# Note that CopyMemory is defined in term of memcpy:
memcpy = cdll.msvcrt.memcpy
CryptProtectData = windll.crypt32.CryptProtectData
CryptUnprotectData = windll.crypt32.CryptUnprotectData

# See http://msdn.microsoft.com/architecture/application/default.aspx?pull=
# /library/en-us/dnnetsec/html/SecNetHT07.asp
CRYPTPROTECT_UI_FORBIDDEN = 0x01
CRYPTPROTECT_LOCAL_MACHINE = 0x04

__E = "AU$HPRO$Y"


class DATA_BLOB(Structure):
    # See d:\vc98\Include\WINCRYPT.H
    # This will not work
    # _fields_ = [("cbData", DWORD), ("pbData", c_char_p)]
    # because accessing pbData will create a new Python string which is
    # null terminated.
    # cbData is a byte count for the blob
    # pbData is the pointer to the data buffer
    _fields_ = [("cbData", DWORD), ("pbData", POINTER(c_char))]


def _getData(blobOut):
    """Copies blob data to a c_buffer and frees the original blob data
    using LocalFree.
    This should ONLY be used on DATA_BLOBS returned by the Windows API calls.
    This allows us to keep the returned answer around as long as we need
    and not have to worry about free'ing the DLL alloc'ed memory later."""

    cbData = int(blobOut.cbData)
    pbData = blobOut.pbData
    buffer = c_buffer(cbData)
    memcpy(buffer, pbData, cbData)
    LocalFree(pbData)
    return buffer.raw


def win32CryptProtectData(plainText, entropy=__E):
    """Returns plainText encrypted and base64Encoded"""

    bufferIn = c_buffer(plainText, len(plainText))
    blobIn = DATA_BLOB(len(plainText), bufferIn)
    bufferEntropy = c_buffer(entropy, len(entropy))
    blobEntropy = DATA_BLOB(len(entropy), bufferEntropy)
    blobOut = DATA_BLOB()
    # The CryptProtectData function performs encryption on the data
    # in a DATA_BLOB structure.
    # BOOL WINAPI CryptProtectData(
    #    DATA_BLOB* pDataIn,
    #    LPCWSTR szDataDescr,
    #    DATA_BLOB* pOptionalEntropy,
    #    PVOID pvReserved,
    #    CRYPTPROTECT_PROMPTSTRUCT* pPromptStruct,
    #    DWORD dwFlags,
    #    DATA_BLOB* pDataOut
    if CryptProtectData(byref(blobIn), u"win32crypto.py", byref(blobEntropy),
        None, None, CRYPTPROTECT_UI_FORBIDDEN|CRYPTPROTECT_LOCAL_MACHINE, byref(blobOut)):
        return base64.b64encode(_getData(blobOut))
    else:
        return None


def win32CryptUnprotectData(cipherText, entropy=__E):
    """Returns cipherText, which should be encrypted and base64Encoded,
    as plain text."""

    cipherText = base64.b64decode(cipherText)
    bufferIn = c_buffer(cipherText, len(cipherText))
    blobIn = DATA_BLOB(len(cipherText), bufferIn)
    bufferEntropy = c_buffer(entropy, len(entropy))
    blobEntropy = DATA_BLOB(len(entropy), bufferEntropy)
    blobOut = DATA_BLOB()
    if CryptUnprotectData(byref(blobIn), None, byref(blobEntropy), None,
        None, CRYPTPROTECT_UI_FORBIDDEN|CRYPTPROTECT_LOCAL_MACHINE, byref(blobOut)):
        return _getData(blobOut)
    else:
        return None
