#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import copy
import random
import traceback

import twisted.internet.error
from twisted.internet import protocol, reactor, defer
from pyrad import packet

from .. import log
import base
from ..base import AuthError

import six

# create a strong PRNG
_generator = random.SystemRandom()


class ClientRequest(base.RadiusRequest):
    def __init__(self, request_packet, pw_codec='utf-8'):
        base.RadiusRequest.__init__(self, request_packet, pw_codec)

        # delayed call to resend
        self.retry_forward_dc = None

        # deferred for callback
        self.deferred = defer.Deferred()


class _IdentifierList(object):
    MAX_WAITING = 256

    def __init__(self):
        # not sure if deterministic id choice would be bad here, but
        # let's make it random
        self.unused_ids = range(0, 256)
        _generator.shuffle(self.unused_ids)

        # so that we can block if we run out of IDs
        self.wait_queue = []

    @defer.inlineCallbacks
    def request(self):
        # get an identifier for use with the dest server
        # this may not strictly be necessary, but should prevent collisions in cases
        # where the proxy may be serving multiple client devices...
        if len(self.unused_ids) < 1:
            # wait until an id becomes available. unless we've got way too many waiting already.
            if len(self.wait_queue) > self.MAX_WAITING:
                raise Exception('Too many concurrent requests!')
            deferred = defer.Deferred()
            self.wait_queue += [deferred]
            yield deferred

        id = self.unused_ids[0]
        self.unused_ids = self.unused_ids[1:]
        defer.returnValue(id)

    def release(self, id):
        insert_pos = _generator.randrange(len(self.unused_ids) + 1)
        self.unused_ids.insert(insert_pos, id)

        if len(self.wait_queue) > 0:
            deferred = self.wait_queue[0]
            self.wait_queue = self.wait_queue[1:]
            deferred.callback()


class RadiusClient(protocol.DatagramProtocol):
    def __init__(self, addrs, nas_ip, secret, retries,
                 retry_wait, debug, pw_codec='utf-8'):
        self.addrs = addrs
        if len(self.addrs) < 1:
            raise Exception('Need at least one destination (host, port)!')
        self.nas_ip = nas_ip
        self.secret = secret
        self.retries = retries
        self.retry_wait = retry_wait
        self.pw_codec = pw_codec

        self.requests = {}
        self.id_list = _IdentifierList()
        self.debug = debug

    @defer.inlineCallbacks
    def authenticate(self, username, password, client_ip, pass_through_attrs=None):
        if pass_through_attrs is None:
            pass_through_attrs = {}
        id = yield self.id_list.request()
        request_packet = packet.AuthPacket(code=packet.AccessRequest,
                                           id=id,
                                           secret=self.secret,
                                           dict=base.radius_dictionary())
        request_packet['NAS-IP-Address'] = self.nas_ip
        for name, value in pass_through_attrs.iteritems():
            if name not in request_packet:
                if isinstance(value, six.string_types):
                    request_packet.AddAttribute(name, value)
                else:
                    for v in value:
                        request_packet.AddAttribute(name, v)

        request = ClientRequest(request_packet, self.pw_codec)
        request.username = username
        if password is not None:
            request.password = password
        if client_ip:
            request.client_ip = client_ip

        log.msg('Sending request for user %r to %r with id %r'
                % (username, self.addrs[0], id))

        # Send request; wait for response
        self.requests[id] = request
        self._send_request(request)
        response_packet = yield request.deferred
        defer.returnValue(response_packet)

    @defer.inlineCallbacks
    def radius_proxy(self, orig_request):
        """Send the given RADIUS packet and return the response received.
        """
        # Create a new request packet with a different ID but the same
        # type and attributes.
        request_packet = copy.deepcopy(orig_request.packet)
        request_packet.id = yield self.id_list.request()
        request_packet.secret = self.secret
        request = ClientRequest(request_packet, self.pw_codec)

        # Send onward; wait for reply.
        log.msg('Sending proxied request for id %r to %r with id %r'
                % (orig_request.id, self.addrs[0], request_packet.id))
        self.requests[request_packet.id] = request
        self._send_request(request)

        # Make response match principal's expectations.
        response_packet = yield request.deferred
        response_packet.id = orig_request.id
        response_packet.secret = orig_request.secret
        defer.returnValue(response_packet)

    def cleanup_all(self):
        for request in self.requests.values():
            self._request_done(request)
            request.deferred.errback(AuthError('RADIUS client shutting down'))

    def _send_request(self, request, addrs=None, retry=0):
        if not addrs:
            # cycle through configured addresses
            addrs = self.addrs
        if retry > self.retries:
            log.msg('Request timeout for (outgoing) id %r to %r'
                    % (request.id, addrs[0]))
            self._request_done(request)
            request.deferred.errback(AuthError('RADIUS auth request timed out'))
            return

        # send request to first address
        raw = request.packet.RequestPacket()
        self.transport.write(raw, addrs[0])
        if self.debug:
            log.msg("Packet dump - sent to %s:" % (addrs[0][0]))
            log.msg(repr(raw))

        # set retry timer to re-send with next address
        request.retry_forward_dc = reactor.callLater(self.retry_wait,
                                                     self._send_request,
                                                     request,
                                                     addrs[1:],
                                                     retry + 1)

    def _handle_response(self, datagram, source):
        response_packet = packet.AuthPacket(packet=datagram, dict=base.radius_dictionary())
        response_packet.source = source

        # make sure it's from the correct host (or properly ip-spoofed :)
        if (source[0], int(source[1])) not in self.addrs:
            raise packet.PacketError(
                'response packet from unknown address: %s:%s'
                % source,
            )

        # look up id
        try:
            request = self.requests[response_packet.id]
        except KeyError:
            raise packet.PacketError, 'unrecognized id in response packet: %s' % response_packet.id

        # verify reply authentication
        if not request.packet.VerifyReply(response_packet, datagram):
            raise packet.PacketError, 'response packet has invalid authenticator'
        # Validate Message-Authenticator, if any
        response_packet.secret = self.secret
        if not response_packet.VerifyMessageAuthenticator(datagram, request_authenticator=request.packet.authenticator):
            raise packet.PacketError('Invalid Message-Authenticator from {0}'.format(source[0]))
        response_packet.authenticator = request.packet.authenticator

        # clear request state
        self._request_done(request)

        log.msg('Got response for id %r from %r; code %r'
                % (response_packet.id, response_packet.source, response_packet.code))

        # callback
        request.deferred.callback(response_packet)

    def _request_done(self, request):
        try:
            request.retry_forward_dc.cancel()
        except twisted.internet.error.AlreadyCalled:
            pass
        request.retry_forward_dc = None
        del self.requests[request.id]
        self.id_list.release(request.id)

    def datagramReceived(self, datagram, (host, port)):
        if self.debug:
            log.msg('Packet dump - received from %s:' % (host))
            log.msg(repr(datagram))
        try:
            self._handle_response(datagram, (host, port))
        except packet.PacketError, err:
            log.msg("dropping packet from %s:%s - %s" % (host, port, err))
        except Exception, err:
            traceback.print_exc()
            log.err(err)
