#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
import copy
import functools
import itertools

from twisted.internet import reactor, defer
import twisted.internet.protocol

import ldaptor.protocols.pureldap
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.ldap.distinguishedname import (
    DistinguishedName,
    InvalidRelativeDistinguishedName,
)

import six

from duoauthproxy.lib.config_provider import ConfigError
from . import ssl_server
from ..lib import base, log, duo_async, util
from ..lib.base import ServerModule
from ..lib.ldap.proxy import Proxy
from ..lib.const import DEFAULT_LDAP_PORT, DEFAULT_LDAPS_PORT

DEFAULT_DELIM = ','


class DuoAutoLdapServer(Proxy):
    def __init__(self, factors,
                 primary_ator,
                 failmode,
                 exempt_primary_bind=True,
                 exempt_ous=None,
                 allow_searches_after_bind=True,
                 allow_unlimited_binds=False,
                 **kwargs):
        """
        Duo LDAP server with secondary auth factor pre-selected by admin.

        factors: comma-separated list of factor names like "auto,push".

        exempt_primary_bind: If True, Duo auth will not be performed
        for the first bind request in a session. Otherwise, it will be
        performed for every successful bind.

        exempt_ous: List of base DNs (DistinguishedName objects or
        strings). Duo auth will not be performed for any bind request
        with a DN under one of the listed trees.

        resolving_bind_request: True if a valid BindRequest has been
        received and a BindResponse has not yet been sent back.

        Note: there is currently no support for self-enrollment here!
        """
        if exempt_ous is None:
            exempt_ous = []
        super(DuoAutoLdapServer, self).__init__(**kwargs)
        self.primary_ator = primary_ator
        self.factors = factors
        self.failmode = failmode
        self.exempt_primary_bind = exempt_primary_bind
        self.exempt_ous = [
            DistinguishedName(str(dn).lower())
            for dn in exempt_ous
        ]
        self.no_more_binds = False
        self.allow_searches_after_bind = allow_searches_after_bind
        self.allow_unlimited_binds = allow_unlimited_binds
        self.resolving_bind_request = False

    def log_request(self, request, log_msg):
        addr = self.transport.getPeer()
        log.msg('[Request from %s:%d] %s' % (
            addr.host,
            addr.port,
            log_msg,
        ))

    def _request_do_concat(self, request):
        """
        Check if request is possibly a concat-type BindRequest containing a delimiter and a factor.
        """
        if not self.factory.allow_concat:
            return False
        if request.sasl or not isinstance(request.auth,
                                          six.string_types):
            # Explicitly exempt SASL requests because no concat is
            # possible. They have ('SASL mechanism', 'sasl creds') for
            # auth instead of a PLAIN password string. Delim wouldn't
            # be found in the tuple anyway.
            return False
        if self.factory.delim not in request.auth:
            return False
        if not isinstance(request,
                          ldaptor.protocols.pureldap.LDAPBindRequest):
            return False
        # Not checking exempt_primary_bind. self.bound is only
        # incremented when a bind response is received, so
        # theoretically a second bind request could come in before the
        # primary bind response is received. The equivalent can't
        # happen during response processing because the responses are
        # never received out of sequence.
        #
        # Could artificially serialize bind request processing to work
        # around this, but there are enough other solutions. Service
        # users must either be in exempt_ous, not have delim in the
        # password, or fall through to the second primary auth
        # attempt.
        if self._is_under_exempt_ous(request.dn):
            return False
        return True

    def handle(self, msg):
        """
        Inherited from Proxy, which inherited it from ldaptor's BaseLDAPServer.
        Slightly overriden to drop msg if the protocol's already servicing a BindRequest.
        """
        assert isinstance(msg.value,
                          ldaptor.protocols.pureldap.LDAPProtocolRequest)
        if self.resolving_bind_request:
            msg.value.is_logging_insecure = self.is_logging_insecure
            log.msg("Received extraneous LDAP PDU while resolving a BindRequest: {0}".format(repr(msg)))
        else:
            super(DuoAutoLdapServer, self).handle(msg)

    def handle_LDAPBindRequest(self, orig_request,
                               controls,
                               orig_reply_fn):
        """
        Intercept (some) bind requests and check for concat passwords.
        """
        if self.no_more_binds:
            # We already handled and 2FA'd a bindRequest.
            # Panic!
            self.log_request(orig_request, 'Attempt to bindRequest multiple times in the same LDAP connection.  Disconnecting.')
            self.transport.loseConnection()
            return
        # Verify that it's OK to ignore controls
        self.checkControls(controls)
        # ignore controls during BIND requests
        controls = None
        self.resolving_bind_request = True

        if self._request_do_concat(orig_request):
            def reply_fn(response, response_controls=None):
                if not isinstance(response,
                                  ldaptor.protocols.pureldap.LDAPBindResponse):
                    # Sent a bind request, didn't get a bind
                    # response... What just happened?!
                    return orig_reply_fn(response)
                if response.resultCode != ldaperrors.Success.resultCode:
                    # Maybe it wasn't concat after all? Retry with
                    # un-split password.
                    return self.handleUnknown(orig_request,
                                              controls,
                                              orig_reply_fn)
                else:
                    # Successful primary auth with split password.
                    # relay_response() performed duo_auth using the
                    # smuggled factor if applicable.
                    return orig_reply_fn(response)
            request = copy.deepcopy(orig_request)
            request.auth, factor = request.auth.rsplit(self.factory.delim,
                                                       1)
            if request.auth:
                reply_fn.duo_factor = factor  # commence smuggling!
                return self.handleUnknown(request, controls, reply_fn)
            else:
                # Treating this as concat mode would transform a bind
                # request the real client thought was not anonymous
                # (because it had, at a minimum, the delimiter) into a
                # bind the backing LDAP server thinks is anonymous.
                # This is not safe -- the real client likely doesn't
                # check the result for e.g. "NT AUTHORITY\ANONYMOUS
                # LOGON" instead of the requested DN because it
                # doesn't know it sent an anonymous bind.
                #
                # Assume either 1. this is not actually concat mode
                # and sending the split password would fail of not for
                # anonymous bind or 2. this is concat mode and
                # therefore end user auth, for which anonymous bind is
                # contraindicated. In both cases the original request
                # should be sent instead of checking for concat mode.
                return self.handleUnknown(orig_request, controls, orig_reply_fn)
        else:
            # Allow anonymous bind if the real client did. Pass
            # through anonymous binds if it knowingly sent an empty,
            # non-concat password.
            return self.handleUnknown(orig_request, controls, orig_reply_fn)

    def _response_needs_duo_auth(self, request, response):
        if not isinstance(response,
                          ldaptor.protocols.pureldap.LDAPBindResponse):
            return False
        if not isinstance(request,
                          ldaptor.protocols.pureldap.LDAPBindRequest):
            return False
        if response.resultCode != ldaperrors.Success.resultCode:
            log.auth(
                msg='Primary authentication rejected',
                dn=request.dn,
                auth_stage=log.AUTH_PRIMARY,
                status=log.AUTH_REJECT,
                server_section=self.server_section_name)
            return False
        # At this point, we know we got a successful bind response to a bind request
        log.auth(
            msg='Primary authentication successful',
            dn=request.dn,
            auth_stage=log.AUTH_PRIMARY,
            status=log.AUTH_ALLOW,
            server_section=self.server_section_name)
        if self.exempt_primary_bind and self.bound < 2:
            # E.g. the usual LDAP auth "bind, search, bind" pattern.
            msg = 'Primary bind exempted from 2FA'
            self.log_request(request, msg)
            log.auth(
                msg=msg,
                dn=request.dn,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_ALLOW,
                server_section=self.server_section_name)
            return False
        if not request.dn:
            log.auth(
                msg='Duo authentication bypassed for anonymous bind',
                dn=request.dn,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_ALLOW,
                server_section=self.server_section_name)
            # No username. Can't 2FA. The real client might expect
            # anonymous bind so pass it through.
            return False
        if self._is_under_exempt_ous(request.dn):
            msg = 'Exempt OU: {0}'.format(request.dn)
            log.auth(
                msg=msg,
                dn=request.dn,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_ALLOW,
                server_section=self.server_section_name)
            self.log_request(request, msg)
            return False
        return True

    def _is_under_exempt_ous(self, dn_str):
        if not self.exempt_ous:
            return False
        dn_str = dn_str.lower()
        try:
            dn = DistinguishedName(dn_str)
        except Exception:
            # Response DNs shouldn't be invalid but a request's DN
            # field may be a "NetBIOS domain\username". Implementing
            # exempt_ous for such binds would require an additional
            # search.
            return False
        return any([base_dn.contains(dn) for base_dn in self.exempt_ous])

    @defer.inlineCallbacks
    def relay_response(self, response, reply_fn, request, **kwargs):
        """ Intercept (some) binds in a connection, and handle Duo auth. """

        if self._response_needs_duo_auth(request, response):
            # In concat mode the selected factor is smuggled through
            # ldaptor to this callback as an attribute of the
            # reply_fn.
            factor = getattr(reply_fn, 'duo_factor', None)
            try:
                response = yield self.duo_auth(request,
                                               response,
                                               factor=factor)
            except duo_async.DuoAPIError as e:
                if e.fail_open and self.failmode == base.FAILMODE_SAFE:
                    msg = base.FAILMODE_SAFE_MSG
                    # return primary response unchanged
                    log.auth(
                        msg=msg,
                        dn=request.dn,
                        auth_stage=log.AUTH_SECONDARY,
                        status=log.AUTH_ALLOW,
                        server_section=self.server_section_name)
                    self.log_request(request, msg)
                else:
                    msg = base.FAILMODE_SECURE_MSG
                    log.auth(
                        msg=msg,
                        dn=request.dn,
                        auth_stage=log.AUTH_SECONDARY,
                        status=log.AUTH_ERROR,
                        server_section=self.server_section_name)
                    response = self.create_bind_response(request,
                                                         False,
                                                         msg)
            if not self.allow_unlimited_binds:
                self.no_more_binds = True
        yield super(DuoAutoLdapServer, self).relay_response(
            response=response,
            request=request,
            reply_fn=reply_fn,
            **kwargs
        )
        self.resolving_bind_request = False
        if self.no_more_binds and not self.allow_searches_after_bind:
            self.transport.loseConnection()

    def make_connected_client(self):
        return self.primary_ator.ldap_proxy()

    def create_bind_response(self, request, success, msg):
        self.log_request(request, msg)
        if success:
            code = ldaperrors.Success.resultCode
        else:
            code = ldaperrors.LDAPInvalidCredentials.resultCode
        return ldaptor.protocols.pureldap.LDAPBindResponse(
            resultCode=code,
            errorMessage=msg,
        )

    @defer.inlineCallbacks
    def duo_auth(self, request, primary_response, factor=None):
        try:
            # Get the Duo username from the DN as requested. Still
            # want to apply 2FA to anonymous binds where the matchedDN
            # is empty.
            username = yield self.dn_to_username(request.dn)
        except Exception as e:
            log.msg('Username lookup failed: ' + repr(str(e)))
            username = None
        if not username:
            msg = 'Cannot find username'
            log.auth(
                msg=msg,
                dn=request.dn,
                username=log.AUTH_UNKNOWN,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_REJECT,
                client_ip=log.AUTH_UNKNOWN,
                server_section=self.server_section_name)
            response = self.create_bind_response(request, False, msg)
            defer.returnValue(response)

        preauth_res = yield self.factory.client.preauth(username)
        self.log_request(request, 'Got preauth result for %s: %r' % (
            username,
            preauth_res['result'],
        ))

        if preauth_res['result'] == duo_async.API_RESULT_ALLOW:
            msg = preauth_res['status']
            log.auth(
                msg='Duo preauth result was {0}'.format(msg),
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_ALLOW,
                server_section=self.server_section_name)
            self.log_request(request, msg)
            defer.returnValue(primary_response)
        elif preauth_res['result'] == duo_async.API_RESULT_DENY:
            msg = preauth_res['status']
            log.auth(
                msg='Duo preauth result was {0}'.format(msg),
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_REJECT,
                server_section=self.server_section_name)
            response = self.create_bind_response(request, False, msg)
            defer.returnValue(response)
        elif preauth_res['result'] == duo_async.API_RESULT_ENROLL:
            msg = preauth_res['status']
            log.auth(
                msg='User is not enrolled. Enrollment link sent.',
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_REJECT,
                server_section=self.server_section_name)
            response = self.create_bind_response(request, False, msg)
            defer.returnValue(response)

        if not factor:
            factor = util.factor_for_request(self.factors, preauth_res)
        if not factor or factor == 'passcode':
            msg = 'User has no Duo factors usable with this configuration'
            log.auth(
                msg=msg,
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_REJECT,
                server_section=self.server_section_name)
            response = self.create_bind_response(request, False, msg)
            defer.returnValue(response)

        # LDAP server never sends real client's IP (no equivalent of
        # Calling-Station-Id).
        auth_res = yield self.factory.client.auth(username,
                                                  factor,
                                                  client_ip=None)
        msg = auth_res.get('status', '')
        self.log_request(request, "Duo authentication returned '%s': '%s'"
                         % (auth_res['result'], msg))
        if auth_res['result'] == duo_async.API_RESULT_ALLOW:
            self.log_request(request, msg)
            log.auth(
                msg="Duo authentication succeeded - {0}".format(msg),
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_ALLOW,
                server_section=self.server_section_name
            )
            defer.returnValue(primary_response)
        else:
            log.auth(
                msg="Duo authentication was rejected - {0}".format(msg),
                dn=request.dn,
                username=username,
                auth_stage=log.AUTH_SECONDARY,
                status=log.AUTH_REJECT,
                server_section=self.server_section_name)
            response = self.create_bind_response(request, False, msg)
            defer.returnValue(response)


class DuoAutoLdapServerFactory(twisted.internet.protocol.ServerFactory):
    def __init__(self, duo_client,
                 delim=',',
                 allow_concat=True):
        """
        delim: delimiter for concat-formatted passwords.

        allow_concat: If true, will try to split concat-formatted
        passwords containing delim. Otherwise, concat-mode will not be
        tried.
        """
        self.client = duo_client
        self.delim = delim
        self.allow_concat = allow_concat


class Module(ServerModule):
    def __init__(self, config, primary_ator, server_section_name):
        log.msg('LDAP Automatic Factor Server Module Configuration:')
        log.config(config,
                   (lambda x: x == 'skey'))
        self.port = config.get_int('port', DEFAULT_LDAP_PORT)
        self.interface = config.get_str('interface', '')

        self.ssl_port = config.get_int('ssl_port', DEFAULT_LDAPS_PORT)
        self.ssl_key_path = config.get_str('ssl_key_path', '')
        self.ssl_cert_path = config.get_str('ssl_cert_path', '')
        self.cipher_list = config.get_str('cipher_list', '')
        if self.ssl_key_path and self.ssl_cert_path:
            self.ssl_key_path = util.resolve_file_path(self.ssl_key_path)
            self.ssl_cert_path = util.resolve_file_path(self.ssl_cert_path)
            try:
                self.sslctx = ssl_server.ChainingOpenSSLContextFactory(
                    privatekey_filename=self.ssl_key_path,
                    certificate_filename=self.ssl_cert_path,
                    cipher_list=self.cipher_list
                )
            except Exception as e:
                if e.message[0][1] == 'fopen':
                    if e.message[2][1] == 'SSL_CTX_use_certificate_file':
                        raise ConfigError("Could not open certificate file at %s" % (self.ssl_cert_path))
                    if e.message[2][1] == 'SSL_CTX_use_PrivateKey_file':
                        raise ConfigError("Could not open private key file at %s" % (self.ssl_key_path))
                if e.message[0][1] == 'SSL_CTX_set_cipher_list':
                    raise ConfigError("Invalid OpenSSL cipher list specified %s" % (self.cipher_list))
                raise e

        elif self.ssl_key_path and not self.ssl_cert_path:
            raise ConfigError('"ssl_key_path" cannot be used without "ssl_cert_path"')
        elif self.ssl_cert_path and not self.ssl_key_path:
            raise ConfigError('"ssl_cert_path" cannot be used without "ssl_key_path"')
        else:
            self.sslctx = None
            log.msg('SSL disabled. No server key and certificate configured.')

        exempt_ous = []
        for i in itertools.count(1):
            exempt_ou_key = 'exempt_ou_%d' % i
            if exempt_ou_key not in config:
                break
            exempt_ou = config.get_str(exempt_ou_key)
            try:
                # Syntax check by side effect of parsing. Server does
                # its own parsing and canonicalization.
                DistinguishedName(exempt_ou)
            except InvalidRelativeDistinguishedName:
                raise ConfigError(
                    'The value "%s" of "%s"'
                    ' is an invalid LDAP distinguished name.' % (
                        exempt_ou,
                        exempt_ou_key,
                    )
                )
            exempt_ous.append(exempt_ou)

        exempt_primary_bind = config.get_bool('exempt_primary_bind', True)
        allow_unlimited_binds = config.get_bool('allow_unlimited_binds', False)
        if allow_unlimited_binds and exempt_primary_bind:
            raise ConfigError('Exempt primary bind is not compatible with allow unlimited binds. Set \'exempt_primary_bind=false\' and use \'exempt_ou\' to exempt service accounts from 2fa')

        self.factory = DuoAutoLdapServerFactory(
            duo_client=self.make_duo_client(config),
            delim=config.get_str('delimiter', DEFAULT_DELIM),
            allow_concat=config.get_bool('allow_concat', True),
        )
        self.factory.protocol = functools.partial(
            DuoAutoLdapServer,
            primary_ator=primary_ator,
            failmode=config.get_enum('failmode', base.FAILMODES,
                                     base.FAILMODE_SAFE),
            factors=util.parse_factor_list(config.get_str('factors', 'auto')),
            sslctx=self.sslctx,
            network_timeout=config.get_int('network_timeout',  # 10m
                                           10 * 60),
            idle_timeout=config.get_int('idle_timeout',  # 1h
                                        60 * 60),
            debug=config.get_bool('debug', False),
            is_logging_insecure=config.get_bool('is_logging_insecure', False),
            exempt_primary_bind=exempt_primary_bind,
            exempt_ous=exempt_ous,
            allow_searches_after_bind=config.get_bool('allow_searches_after_bind', True),
            allow_unlimited_binds=allow_unlimited_binds,
            server_section_name=server_section_name,
        )
        self.listener = None
        self.listener_ssl = None
        self._bind_if_necessary()

    def startService(self):
        ServerModule.startService(self)
        self._bind_if_necessary()

    @defer.inlineCallbacks
    def stopService(self):
        super(Module, self).stopService()
        if self.listener:
            yield self.listener.stopListening()
            self.listener = None
        if self.listener_ssl:
            yield self.listener_ssl.stopListening()
            self.listener_ssl = None

    def _bind_if_necessary(self):
        if self.listener is None:
            self.listener = reactor.listenTCP(
                port=self.port,
                factory=self.factory,
                interface=self.interface,
            )
        if self.sslctx and self.listener_ssl is None:
            self.listener_ssl = reactor.listenSSL(
                port=self.ssl_port,
                factory=self.factory,
                contextFactory=self.sslctx,
                interface=self.interface,
            )
