#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#

from twisted.internet import defer

from pyrad import packet
import six
import os

from ..lib import base, log, duo_async, util
from ..lib.base import ServerModule
from ..lib.radius.base import set_message_authenticator
from ..lib.radius.challenge import ChallengeResponseRadiusServer
from ..lib.radius.server import parse_radius_secrets
from ..lib.radius.server import parse_client_ip_attribute
from ..lib.radius.duo_server import DuoSimpleRadiusServer
from ..lib import eap
from ..lib import mppe


def _format_short(preauth_res):
    factors = preauth_res['factors']
    if factors:
        sorted_factors = ('\'%s\'' % factors[k].rstrip('1')
                          for k in sorted(factors.iterkeys())
                          if k != 'default')
        msg = ('Choose a secondary factor from (%s) or enter passcode:'
               % ', '.join(sorted_factors))
        if len(msg) > 253:
            # sanity check. probably shouldn't ever be a problem in practice
            msg = 'Choose a secondary factor or enter passcode:'
    else:
        msg = 'Enter passcode:'
    return msg


class EAPSession(eap.EAPSession):
    def __init__(self, server, **kwargs):
        self.server = server
        super(EAPSession, self).__init__(**kwargs)

    def success(self, session):
        return self.server.accept_session(session)

    def errback(self, session, reason):
        return self.server.kill_session(session, reason)

    def gtc_received(self, session, passcode, prompt):
        return self.server.passcode_received(session, passcode, prompt)


class DuoEAPRadiusServer(DuoSimpleRadiusServer, ChallengeResponseRadiusServer):
    def __init__(self, factors,
                 delim, prompt,
                 pkey_file, cert_file,
                 concat_mode=False, cipher_list='',
                 **kwargs):
        """
        Duo RADIUS server which attempts to start a PEAP/EAP-GTC
        session to get password/factor

        factors: comma-separated list of factor names like "auto,push".

        delim: delimiter for concat-formatted passwords.

        concat_mode: If true, will try to split concat-formatted
        passwords containing delim from the first GTC response,
        otherwise send two prompts for password and factor
        seperately.

        Note: there is currently no support for self-enrollment here!
        """
        super(DuoEAPRadiusServer, self).__init__(**kwargs)
        self.factors = factors
        self.delim = delim
        self.concat_mode = concat_mode
        self.EAPSessions = {}
        self.prompt = prompt
        self.pkey_file = pkey_file
        self.cert_file = cert_file
        self.cipher_list = cipher_list
        if not os.path.isfile(pkey_file):
            self.pkey_file = os.path.join('conf', pkey_file)
        if not os.path.isfile(cert_file):
            self.cert_file = os.path.join('conf', cert_file)

    def fragment(self, x):
        stuff = []
        while x:
            stuff.append(str(x[:253]))
            x = x[253:]
        return stuff

    def create_eap_session(self, request):
        session = EAPSession(
            self,
            pkey=self.pkey_file,
            certs=self.cert_file,
            gtc_message=self.prompt,
            cipher_list=self.cipher_list
        )
        self.log_request(request, 'Creating EAP session for user %s' % (request.username))
        return session

    @defer.inlineCallbacks
    def get_initial_response(self, request):
        # Create an EAP session
        session = self.create_eap_session(request)
        response = yield self.get_challenge_response(request, None, session=session)
        defer.returnValue(response)

    @defer.inlineCallbacks
    def get_challenge_response(self, request, state, session=None):
        if 'EAP-Message' in request.packet:
            request_message = ''.join(request.packet[79])
            if session is None:
                session = state
            session.current_request = request

            add_response = yield session.add_message(request_message)
            if add_response:
                defer.returnValue(add_response)

            response_message = yield session.next_message()

            if isinstance(response_message, six.string_types):
                # Next EAP message
                response_packet = self.create_challenge(request, '', session)
                response_packet[79] = self.fragment(response_message)
            else:
                # response is a packet from kill_session or accept_session
                response_packet = response_message

            # EAP-Message (79) RADIUS packets MUST include a Message-Authenticator.
            set_message_authenticator(response_packet)
            defer.returnValue(response_packet)

        else:
            # If no EAP, treat it like a normal radius request
            try:
                password = request.password
            except Exception:
                password = None
            if not password:
                # Either PAP but blank or un-decryptable. (Not PAP?
                # Wrong shared secret?). Not EAP, either, or
                # EAP-Message would've been found.
                msg = 'Missing or improperly-formatted password'
                self.log_request(request, msg)
                log.auth_standard(msg=msg,
                                  username=request.username,
                                  auth_stage=log.AUTH_PRIMARY,
                                  status=log.AUTH_ERROR,
                                  client_ip=request.client_ip,
                                  server_section=self.server_section_name)
                defer.returnValue(
                    self.create_reject_packet(request, msg)
                )
            else:
                result = yield self.auto_auth(request.username, password)
                defer.returnValue(result)

    @defer.inlineCallbacks
    def passcode_received(self, session, passcode, prompt):
        request = session.current_request
        if self.concat_mode:
            # In concat mode, attempt auth on first password
            result = yield self.auto_auth(request, passcode)
            if result.code == packet.AccessAccept:
                session.state = eap.EAP_SESSION_PEAP_ACCEPT
            else:
                session.state = eap.EAP_SESSION_DENY
        else:
            # If we didn't get any input, just resend the last prompt
            if passcode == '':
                return

            if not session.password:
                # otherwise, if this is the first credential we got
                # from the user, then do primary auth
                session.password = passcode
                session.primary_res = yield self.primary_auth(request, session.password)

                # if primary auth fails, send back a reject
                if not session.primary_res.success:
                    session.state = eap.EAP_SESSION_DENY
                    return

                # otherwise, do preauth and proceed according to the response
                resp, preauth_res = yield self.preauth(request, session.primary_res)
                session.state = eap.EAP_SESSION_PEAP_GTC

                if preauth_res is None:
                    if resp.code == packet.AccessAccept:
                        session.state = eap.EAP_SESSION_PEAP_ACCEPT
                    else:
                        session.state = eap.EAP_SESSION_DENY
                else:
                    if preauth_res['result'] == duo_async.API_RESULT_ENROLL:
                        session.enrolling = True
                        msg = preauth_res['status'] + ' then try again.'
                        session.innerEAP.gtc_message = msg
                        session.state = eap.EAP_SESSION_PEAP_GTC
                    elif preauth_res['result'] == duo_async.API_RESULT_AUTH:
                        session.innerEAP.gtc_message = _format_short(preauth_res)
                        session.gtc_message = session.innerEAP.gtc_message
                    elif preauth_res['result'] == duo_async.API_RESULT_DENY:
                        session.state = eap.EAP_SESSION_DENY
                    elif preauth_res['result'] == duo_async.API_RESULT_ALLOW:
                        session.state = eap.EAP_SESSION_PEAP_ACCEPT

            elif session.enrolling:
                # Always deny an auth attempt after displaying enrollment link
                session.state = eap.EAP_SESSION_DENY
            else:
                result = yield self.auto_auth(
                    request,
                    session.password,
                    factor=passcode,
                    primary_res=session.primary_res
                )
                if result.code == packet.AccessAccept:
                    session.state = eap.EAP_SESSION_PEAP_ACCEPT
                elif passcode.rstrip('0123456789') == 'sms':
                    # Let them enter an SMSed passcode
                    if 'Reply-Message' in result:
                        session.innerEAP.gtc_message = result['Reply-Message'][0] + '\n' + session.innerEAP.gtc_message
                    else:
                        # Reject result and no reply message means primary auth failed, no sms sent
                        log.auth_standard(msg="Duo authentication failed",
                                          username=request.username,
                                          auth_stage=log.AUTH_SECONDARY,
                                          status=log.AUTH_REJECT,
                                          client_ip=request.client_ip,
                                          server_section=self.server_section_name)
                        session.state = eap.EAP_SESSION_DENY
                else:
                    log.auth_standard(msg="Unrecognized passcode",
                                      username=request.username,
                                      auth_stage=log.AUTH_SECONDARY,
                                      status=log.AUTH_REJECT,
                                      client_ip=request.client_ip,
                                      server_section=self.server_section_name)
                    session.state = eap.EAP_SESSION_DENY

    def kill_session(self, session, reason):
        request = session.current_request
        self.log_request(request, 'Destroying EAP session: %s' % (reason))
        response_packet = self.create_reject_packet(request)
        return response_packet

    def accept_session(self, session):
        request = session.current_request
        response_packet = self.create_accept_packet(request)
        eap_packet = eap.EAPPacket(eap.EAP_CODE_SUCCESS, session.id, 0, '')
        response_packet[79] = [eap_packet.render()]
        mppe_keys = mppe.get_mppe_keys(
            session.mk,
            session.cr,
            session.sr,
            'client EAP encryption',
            session.ssl_con
        )
        mppe.add_mppe(
            response_packet,
            mppe_keys,
            self.secret_for_host(request.source[0]),
            request.packet.authenticator
        )
        self.kill_session(session, 'Success')
        self.log_request(request, 'Authentication successful')
        return response_packet

    @defer.inlineCallbacks
    def auto_auth(self, request, password, factor=None, primary_res=None):
        if not factor:
            if self.delim in password:
                # User may specify passcode or factor with concat.
                # Speculatively split the password. If primary auth
                # succeeds, it was concat.
                #
                # Could reduce false-positives by checking if the part
                # after the last delim actually looks like a factor.
                # However, the general case is difficult because of
                # e.g. ModHex passcodes.
                password_part, factor = password.rsplit(self.delim, 1)
                if not factor:
                    factor = None
                    self.log_request(request, 'attempting authentication with factor auto')
                else:
                    self.log_request(request, 'attempting authentication with factor %s' % (factor))
                if primary_res is None:
                    primary_res = yield self.primary_auth(
                        request,
                        password_part
                    )
            else:
                primary_res = None
                factor = None

            if not (primary_res and primary_res.success):
                # (Re)try primary auth with the whole password if:
                # * No delim for concat.
                # * Delim found, but primary auth failed indicating it
                #   was not really concat.
                # * Concat was intended but the primary auth password
                #   part was wrong. If so, this will fail, too.
                primary_res = yield self.primary_auth(request, password)
                factor = None
        else:
            if primary_res is None:
                primary_res = yield self.primary_auth(request, password)

        if primary_res.success:
            result = yield self.duo_auth(request, primary_res, factor=factor)
        else:
            result = self.create_reject_packet(request)

        defer.returnValue(result)

    @defer.inlineCallbacks
    def duo_auth(self, request, primary_res,
                 factor=None,
                 preauth_res=None):
        if not preauth_res:
            # Must call preauth even if the factor is known.
            response_packet, preauth_res = yield self.preauth(request,
                                                              primary_res)
            if response_packet is not None:
                # E.g. enroll policy of deny or allow.
                defer.returnValue(response_packet)
        if factor is None:
            factor = util.factor_for_request(self.factors, preauth_res)
            if factor is None:
                msg = 'User has no Duo factors usable with this configuration'
                log.auth_standard(msg=msg,
                                  username=request.username,
                                  auth_stage=log.AUTH_SECONDARY,
                                  status=log.AUTH_ERROR,
                                  client_ip=request.client_ip,
                                  server_section=self.server_section_name)
                self.log_request(request, msg)
                defer.returnValue(self.create_reject_packet(request, msg))

        # Factor was either passed in or found above auto-push
        response_packet = yield super(DuoEAPRadiusServer, self).duo_auth(
            request,
            primary_res,
            factor,
        )
        defer.returnValue(response_packet)


class Module(ServerModule):
    def __init__(self, config, primary_ator, server_section_name):
        log.msg('RADIUS PEAP/EAP-GTC Automatic Factor Server Module Configuration:')
        log.config(
            config, lambda x: x.startswith('radius_secret') or
            x in ('skey', 'skey_protected'))

        self.protocol = DuoEAPRadiusServer(
            secrets=parse_radius_secrets(config),
            primary_ator=primary_ator,
            duo_client=self.make_duo_client(config),
            failmode=config.get_enum('failmode', base.FAILMODES,
                                     base.FAILMODE_SAFE),
            factors=util.parse_factor_list(config.get_str('factors', 'auto')),
            delim=config.get_str('delimiter', ','),
            concat_mode=config.get_bool('concat_mode', False),
            debug=config.get_bool('debug', False),
            pass_through_attr_names=config.get_str('pass_through_attr_names', ''),
            prompt=config.get_str('prompt', 'Enter your password: '),
            pkey_file=config.get_str('pkey'),
            cert_file=config.get_str('certs'),
            cipher_list=config.get_str('cipher_list', ''),
            pw_codec=config.get_str('pw_codec', 'utf-8'),
            client_ip_attr=parse_client_ip_attribute(config),
            server_section_name=server_section_name,
        )
