import functools
import operator

from twisted.internet import ssl
import OpenSSL

from ..lib import log

DEFAULT_SSL_ECDH_CURVE = 'prime256v1'

DEFAULT_METHOD = OpenSSL.SSL.SSLv23_METHOD


def _info_callback(conn, where, ret):
    """Use the info callback to gather information about attempted
    SSL connections and warn about incompatibilities. See the man
    page[1] for further information.

    [1] https://www.openssl.org/docs/manmaster/man3/SSL_CTX_set_info_callback.html
    """
    if where & OpenSSL.SSL.SSL_CB_WRITE_ALERT:
        if conn.get_cipher_name() is None:
            log.err(
                "Unable to establish SSL connection. "
                "Client may be attemping incompatible protocol version or cipher."
            )


class ChainingOpenSSLContextFactory(ssl.DefaultOpenSSLContextFactory, object):

    _handshake_done = object()

    def __init__(
            self,
            privatekey_filename=None,
            certificate_filename=None,
            privatekey_data=None,
            certificate_data=None,
            method=None,
            cipher_list=None,
            context_factory=None):
        if not bool(privatekey_filename) ^ bool(privatekey_data):
            raise ValueError("Please specify one of 'privatekey_filename' or 'privatekey_data'")
        if not bool(certificate_filename) ^ bool(certificate_data):
            raise ValueError("Please specify one of 'certificate_filename' or 'certificate_data'")

        self.privatekey_filename = privatekey_filename
        self.certificate_filename = certificate_filename
        self.privatekey_data = privatekey_data
        self.certificate_data = certificate_data
        self.cipher_list = cipher_list

        self.method = method if method is not None else DEFAULT_METHOD
        self.context_factory = context_factory if context_factory is not None else OpenSSL.SSL.Context

        self.cacheContext()

    def cacheContext(self):
        """The default twisted ssl context factory doesn't support certificate
        chains, which we need for our PEMs with intermediate certs.
        """
        if self._context is not None:
            return

        context = self.context_factory(self.method)

        ssl_options = [
            OpenSSL.SSL.OP_NO_SSLv2,
        ]
        options = functools.reduce(operator.ior, ssl_options)
        context.set_options(options)

        if self.privatekey_filename:
            context.use_privatekey_file(self.privatekey_filename)
        elif self.certificate_data:
            context.use_privatekey(self.privatekey_data)

        if self.certificate_filename:
            context.use_certificate_file(self.certificate_filename)
            context.use_certificate_chain_file(self.certificate_filename)
        elif self.certificate_data:
            context.use_certificate(self.certificate_data)

        if self.cipher_list:
            context.set_cipher_list(self.cipher_list)

        elliptic_curve = OpenSSL.crypto.get_elliptic_curve(DEFAULT_SSL_ECDH_CURVE)
        context.set_tmp_ecdh(elliptic_curve)

        context.set_info_callback(_info_callback)

        self._context = context
