#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import optparse
import os
import platform
import string
import sys
import types
import warnings

if os.name != 'nt':
    # This module is not supported on Windows
    import pwd


warnings.filterwarnings("ignore",
                        category=DeprecationWarning,
                        message="^twisted\.internet\.interfaces\.IStreamClientEndpointStringParser was deprecated in "
                                "Twisted 14\.0\.0: This interface has been superseded by "
                                "IStreamClientEndpointStringParserWithReactor\.$")

from twisted.application.service import Application, Service
from twisted.logger import FileLogObserver, FilteringLogObserver, textFileLogObserver, globalLogPublisher, LegacyLogObserverWrapper
from twisted.python.logfile import LogFile
from twisted.python.log import ILogObserver

from lib import (
    http,
    ssl_verify,
    base,
    log,
    forward_serv,
    util,
    config_provider,
    const,
)
from lib.radius.server import parse_radius_secrets

try:
    from twisted.python import syslog
    import syslog as pySyslog
except ImportError:
    syslog = None

import modules


def get_version():
    import duoauthproxy
    return duoauthproxy.__version__


class LogReadyService(Service):
    """Dumb 'service' whose sole purpose is to print out our version number
    on startup"""

    def startService(self):
        version = get_version()
        log.msg('Duo Security Authentication Proxy %s - Init Complete'
                % version)
        log.ready()
        Service.startService(self)


def create_application(args=None, twistd_user=None):
    home_dir = util.get_home_dir()
    os.chdir(home_dir)
    is_logging_insecure = False

    if syslog is not None:
        facilityDict = {
            'LOG_KERN': pySyslog.LOG_KERN,
            'LOG_USER': pySyslog.LOG_USER,
            'LOG_MAIL': pySyslog.LOG_MAIL,
            'LOG_DAEMON': pySyslog.LOG_DAEMON,
            'LOG_AUTH': pySyslog.LOG_AUTH,
            'LOG_LPR': pySyslog.LOG_LPR,
            'LOG_NEWS': pySyslog.LOG_NEWS,
            'LOG_UUCP': pySyslog.LOG_UUCP,
            'LOG_CRON': pySyslog.LOG_CRON,
            'LOG_SYSLOG': pySyslog.LOG_SYSLOG,
            'LOG_LOCAL0': pySyslog.LOG_LOCAL0,
            'LOG_LOCAL1': pySyslog.LOG_LOCAL1,
            'LOG_LOCAL2': pySyslog.LOG_LOCAL2,
            'LOG_LOCAL3': pySyslog.LOG_LOCAL3,
            'LOG_LOCAL4': pySyslog.LOG_LOCAL4,
            'LOG_LOCAL5': pySyslog.LOG_LOCAL5,
            'LOG_LOCAL6': pySyslog.LOG_LOCAL6,
            'LOG_LOCAL7': pySyslog.LOG_LOCAL7
        }

    # locate configfile
    config_filename = os.path.join('conf', 'authproxy.cfg')

    # parse command-line args, if appropriate
    if args:
        option_parser = optparse.OptionParser()
        option_parser.add_option(
            "-c",
            "--config-file",
            dest="config_filename",
            help="Path to config file",
            default=config_filename
        )
        option_parser.add_option(
            "--logging-insecure",
            action="store_true",
            help="This option enables debug, and prints logs containing passwords and possibly other secrets.",
            default=False
        )
        (options, args) = option_parser.parse_args()
        config_filename = options.config_filename
        is_logging_insecure = options.logging_insecure

    try:
        # open configfile and construct ConfigProvider
        with open(config_filename, "rb") as config_file:
            configuration = config_provider.ConfigProvider.from_ini_file(config_file, os.path.abspath(config_filename))
    except IOError:
        raise config_provider.ConfigError('Config file missing or unreadable: {0}'.format(config_filename))

    main_config = configuration.get_main_section_config()
    if len(main_config) > 0:
        log.msg('Main Configuration:')
        log.config(main_config)

    # handle log configuration
    log_to_file = main_config.get_bool('log_file', False)
    log_stdout = main_config.get_bool('log_stdout', False)
    log_syslog = main_config.get_bool('log_syslog', False)
    log_auth_events = main_config.get_bool('log_auth_events', False)

    if is_logging_insecure:
        globalLogPublisher.addObserver(textFileLogObserver(sys.stdout))
    else:
        if log_to_file or not (log_to_file or log_syslog or log_stdout):
            log_dir = main_config.get_str('log_dir', 'log')
            log_max_size = main_config.get_int('log_max_size', 10 * (1 << 20))
            log_max_files = main_config.get_int('log_max_files', 6)
            if log_max_files == 0:
                # we need to pass None explicitly if we want there to be no limit
                # 0 would just mean no logfiles would get kept...
                log_max_files = None

            log_file = create_log_file('authproxy.log', log_dir, log_max_size, log_max_files, twistd_user)
            log_file_observer = textFileLogObserver(log_file)

            if log_auth_events:
                auth_log_file = create_log_file('authevents.log', log_dir, log_max_size, log_max_files, twistd_user)
                auth_observer = FileLogObserver(auth_log_file, log.format_auth_event)
            else:
                auth_observer = log.no_op_observer

            filtering_observer = FilteringLogObserver(auth_observer,
                                                      [log.auth_type_predicate],
                                                      log_file_observer)

            globalLogPublisher.addObserver(filtering_observer)
        if log_stdout:
            std_out_observer = textFileLogObserver(sys.stdout)
            std_out_filter = FilteringLogObserver(log.no_op_observer,
                                                  [log.auth_type_predicate],
                                                  std_out_observer)
            globalLogPublisher.addObserver(std_out_filter)
        if log_syslog:
            if syslog is None:
                raise config_provider.ConfigError('syslog not supported on Windows')
            syslog_facilitystr = main_config.get_str('syslog_facility', 'LOG_USER')
            syslog_facility = facilityDict.get(syslog_facilitystr, None)
            if syslog_facility is None:
                raise config_provider.ConfigError('Unknown syslog_facility: {0}'.format(syslog_facilitystr))
            syslog_observer = syslog.SyslogObserver('Authproxy', facility=syslog_facility)
            wrapped_syslog_observer = LegacyLogObserverWrapper(syslog_observer.emit)
            syslog_filtering_observer = FilteringLogObserver(log.no_op_observer,
                                                             [log.auth_type_predicate],
                                                             wrapped_syslog_observer)
            globalLogPublisher.addObserver(syslog_filtering_observer)

    # Global debug mode
    if is_logging_insecure:
        debug_mode = True
    else:
        debug_mode = main_config.get_bool('debug', False)
    http.set_debug(debug_mode)

    # Create main application.
    application = Application('duoauthproxy')
    LogReadyService().setServiceParent(application)

    # get ca certs file
    # use const as default and then resolve file path unconditionally
    http_ca_certs_file = main_config.get_str('http_ca_certs_file', '')
    if http_ca_certs_file:
        http_ca_certs_file = util.resolve_file_path(http_ca_certs_file)
    else:
        http_ca_certs_file = os.path.join('conf', const.DEFAULT_HTTP_CERTS_FILE)

    # read ca certs
    if not os.path.isfile(http_ca_certs_file):
        http_ca_certs_file = os.path.join('conf', http_ca_certs_file)
    with open(http_ca_certs_file, 'r') as bundle_fp:
        http.set_ca_certs(ssl_verify.load_ca_bundle(bundle_fp))

    # get proxy settings
    http_proxy_host = main_config.get_str('http_proxy_host', '')
    http_proxy_port = main_config.get_int('http_proxy_port', 80)
    if http_proxy_host:
        http.set_proxy(http_proxy_host, http_proxy_port)

    servers = {}
    clients = {}
    cloud_module = None
    http_proxy_modules = {}
    for section_name_raw in configuration.list_sections():
        section_name_lowered = section_name_raw.lower()
        if section_name_lowered == 'main':
            continue
        section_type = section_name_lowered.rstrip(string.digits)

        # Make sure section is a valid module
        try:
            section_module = getattr(modules, section_type)
            if not isinstance(section_module, types.ModuleType):
                raise Exception()
            if issubclass(section_module.Module, base.ServerModule):
                module_type = 'server'
            elif issubclass(section_module.Module, base.ClientModule):
                module_type = 'client'
            elif issubclass(section_module.Module, modules.http_proxy.Module):
                pass
            else:
                raise Exception()
        except (Exception, AttributeError):
            raise config_provider.ConfigError('Invalid module: \'%s\'' % section_name_raw)

        # create server service instances
        section_config = configuration.get_section_config(section_name_raw)

        # Insecure logging flag cannot be set in configuration so we must set it manually
        if is_logging_insecure:
            section_config['is_logging_insecure'] = 'True'
            section_config['debug'] = 'True'

        if section_type == 'cloud':
            if cloud_module is None:
                cloud_module = section_module.Module(section_config)
                cloud_module.setServiceParent(application)
            else:
                raise config_provider.ConfigError("Configuring multiple [cloud] sections isn't supported.")
        elif section_type == 'http_proxy':
            http_proxy_modules[section_name_lowered] = section_module.Module(section_config)
        elif module_type == 'server':
            port = section_config.get_int("port", 1812)
            interface = section_config.get_str('interface', '')
            servers.setdefault((port, interface), []).append((section_name_lowered, section_module, section_config))
        else:
            clients[section_name_lowered] = section_module.Module(section_config)

    if not (servers or clients or cloud_module or http_proxy_modules):
        raise config_provider.ConfigError('No integrations in config file.')

    for client_name in clients.keys():
        clients[client_name].setServiceParent(application)

    for proxy_name in http_proxy_modules.keys():
        http_proxy_modules[proxy_name].setServiceParent(application)

    # For backwards compatibility, track the client specified in [main], if any
    main_client = main_config.get_str('client', '')

    # Setup forwarding/server pairs by port
    for port, interface in servers.keys():
        server_networks = {}
        server_names = {}
        for section_name, server_module, server_config in servers[(port, interface)]:
            client_name = server_config.get_str("client", "").lower()
            if not client_name:
                if 'ldap_server' in section_name and 'ad_client' in clients:
                    client_name = 'ad_client'
                elif main_client:
                    client_name = main_client
                elif len(clients) == 1:
                    client_name = clients.keys()[0]
                elif server_module.Module.no_client:
                    clients['no_client'] = None
                    client_name = 'no_client'
                else:
                    raise config_provider.ConfigError('Neither module %s or main has "client" value' % section_name)

            if client_name not in clients:
                raise config_provider.ConfigError('Invalid "client" value %s' % client_name)

            if client_name != 'radius_client' \
                    and server_config.get_str('pass_through_attr_names', ''):
                raise config_provider.ConfigError('Can only pass through radius attributes if using a radius client')
            server_instance = server_module.Module(server_config, clients[client_name], section_name)
            server_instance.setServiceParent(application)

            if 'ldap_server' in section_name:
                if 'ad_client' not in client_name:
                    raise config_provider.ConfigError('LDAP servers must use ad_client')
            elif section_name.startswith('radius_server_'):
                server_networks[server_instance] = parse_radius_secrets(server_config).keys()
                server_names[server_instance] = section_name

        if server_names:
            forward_module = forward_serv
            forward_instance = forward_module.Module(
                port=port,
                servers=server_networks,
                server_names=server_names,
                interface=interface,
                debug=debug_mode,
            )
            forward_instance.setServiceParent(application)

    # set user-agent
    sections = ','.join(sorted(set(configuration.list_sections())))
    user_agent = "duoauthproxy/{0} ({1}; Python{2}; {3})".format(
            get_version(),
            platform.platform(),
            platform.python_version(),
            sections)
    http.set_user_agent(user_agent)

    # Authproxy uses globalLogPublisher to emit events. Defining a no-op emitter will squelch the creation
    # of the unwatned twistd default logging mechanisms.
    def no_op_emitter(eventDict):
        pass

    application.setComponent(ILogObserver, no_op_emitter)

    return application


def create_log_file(filename, directory, max_size, max_files, twistd_user):
    """Helper function to create twisted LogFiles and set file permissions

    Change the log file permissions to match our service user if one is defined.
    This is needed so that the service can rotate the log files.
    """
    log_file = LogFile(filename, directory, rotateLength=max_size, maxRotatedFiles=max_files)

    if twistd_user is not None:
        uid, gid = _parse_user(twistd_user)
        os.chown(os.path.join(directory, filename), uid, gid)

    return log_file


def _parse_user(user):
    if os.name == 'nt':
        raise Exception('This function is not supported on windows')

    row = None
    try:
        try:
            # attempt to treat the 'user' option as a uid
            uid = int(user)
        except ValueError:
            # if that failed, look up the username and get the associated uid
            row = pwd.getpwnam(user)
        else:
            row = pwd.getpwuid(uid)
    except KeyError:
        print >>sys.stderr,  'Warning: user \'%s\' does not exist - ' \
              'duoauthproxy will not drop privileges' % user
    if row:
        return (row.pw_uid, row.pw_gid)
    return (None, None)
