#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import os
import subprocess
import win32serviceutil
import win32evtlogutil
import win32event
import traceback
import servicemanager

from twisted.application import app
from twisted.internet import reactor

import proxy
from duoauthproxy.lib import util
from duoauthproxy.lib import config_provider


class MyService(win32serviceutil.ServiceFramework):
    _svc_name_ = "DuoAuthProxy"
    _svc_display_name_ = "Duo Security Authentication Proxy Service"
    _svc_deps_ = ["EventLog"]

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        self.duo_root = util.get_home_dir()

    def SvcStop(self):
        reactor.callFromThread(reactor.stop)
        win32event.WaitForSingleObject(self.hWaitStop, win32event.INFINITE)

    def SvcDoRun(self):
        win32evtlogutil.AddSourceToRegistry(self._svc_name_,
                                            servicemanager.__file__)

        # Write a 'started' event to the event log...
        win32evtlogutil.ReportEvent(
            self._svc_name_,
            servicemanager.PYS_SERVICE_STARTED,
            0,  # category
            servicemanager.EVENTLOG_INFORMATION_TYPE,
            (self._svc_name_, ''))

        try:
            self.run_connectivity_tool()

            application = proxy.create_application()
            app.startApplication(application, 0)
            reactor.run(installSignalHandlers=0)
        except Exception:
            # service did not shut down cleanly
            message = traceback.format_exc()
            win32evtlogutil.ReportEvent(
                self._svc_name_,
                servicemanager.PYS_SERVICE_STOPPED,
                0,  # category
                servicemanager.EVENTLOG_ERROR_TYPE,
                (self._svc_name_,
                 ' due to a failure:\n\n%s\n' \
                 'If this was due to a connectivity problem, the connectivity tool at %s can be run ' \
                 'independently to test for connectivity problems.' % (message, self.conn_tool_path)))
        else:
            # service shut down cleanly
            win32evtlogutil.ReportEvent(
                self._svc_name_,
                servicemanager.PYS_SERVICE_STOPPED,
                0,  # category
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                (self._svc_name_, ''))

        win32event.SetEvent(self.hWaitStop)

    def run_connectivity_tool(self):
        config_filename = os.path.join(self.duo_root, 'conf', 'authproxy.cfg')
        try:
            with open(config_filename, "rb") as config_file:
                configuration = config_provider.ConfigProvider.from_ini_file(config_file, os.path.abspath(config_filename))
        except IOError:
            msg = ' but failed to run the connectivity tool because the configuration file is missing or unreadable:\n\n{0}\n'.format(
                    os.path.abspath(config_filename))
            self.log_error_during_startup(msg)
            return

        main_config = configuration.get_main_section_config()
        if not main_config.get_bool('test_connectivity_on_startup', False):
            return

        conn_tool = self.conn_tool_path
        try:
            subprocess.call(conn_tool)
        except OSError as e:
            msg = ' but failed to run the connectivity tool due to a failure:\n\n{0}\n'.format(str(e))
            self.log_error_during_startup(msg)

    def log_error_during_startup(self, msg):
        """ Log an error that occurs during startup process
        Args:
            msg (str): Fill in the blank! "The DuoAuthProxy service is starting ____
        """
        win32evtlogutil.ReportEvent(
            self._svc_name_,
            servicemanager.PYS_SERVICE_STARTING,
            0,  # category
            servicemanager.EVENTLOG_ERROR_TYPE,
            (self._svc_name_, msg)
        )

    @property
    def conn_tool_path(self):
        return os.path.join(self.duo_root, "bin", "authproxy_connectivity_tool")


if __name__ == '__main__':
    # Note that this code will not be run in the 'frozen' exe-file!!!
    win32serviceutil.HandleCommandLine(MyService)
