#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import mock

from twisted.internet import defer

from ...base import ModuleConnectivityTest, MockConnectivityTestToolbox, MockProtocol, ConnectivityMockException
from duoauthproxy.lib.connectivity.check import ad_client
from duoauthproxy.lib.ldap.client import AD_TRANSPORT_CLEAR
from duoauthproxy.lib.config_provider import ConfigDict, ConfigError
from duoauthproxy.lib.connectivity.connectivity_results import (
    AdClientResult,
    ConfigCheckResult,
    MissingConfigKeyProblem,
    ConfigErrorProblem,
    LdapHostResult,
    LdapConnectResult,
    UnmetPrerequisiteSkippedTestResult,
)


class TestCheckADClient(ModuleConnectivityTest):
    def setUp(self):
        super(TestCheckADClient, self).setUp()

        self.BASIC_CONFIG_INPUT = ConfigDict({
            'host': '127.0.0.1',
            'service_account_username': 'user',
            'service_account_password': 'pass',
            'search_dn': 'dc=example,dc=com',
        })

    @defer.inlineCallbacks
    def test_no_host(self):
        expected_result = AdClientResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('host')
            ]),
            []
        )

        del self.BASIC_CONFIG_INPUT['host']
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_no_username(self):
        config_error = ConfigError("Missing required configuration item: 'service_account_username'")
        expected_result = AdClientResult(
            ConfigCheckResult([
                ConfigErrorProblem(config_error)
            ]),
            []
        )

        del self.BASIC_CONFIG_INPUT['service_account_username']
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_no_password(self):
        config_error = ConfigError("Missing required configuration item: 'service_account_password'")
        expected_result = AdClientResult(
            ConfigCheckResult([
                ConfigErrorProblem(config_error)
            ]),
            []
        )

        del self.BASIC_CONFIG_INPUT['service_account_password']
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_no_search_dn(self):
        config_error = ConfigError("Missing required configuration item: 'search_dn'")
        expected_result = AdClientResult(
            ConfigCheckResult([
                ConfigErrorProblem(config_error)
            ]),
            []
        )

        del self.BASIC_CONFIG_INPUT['search_dn']
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_error_in_parse_ad_client(self):
        config_error = ConfigError('Invalid LDAP Filter: asdf')
        """Create exception in parse_ad_client and make sure we handle it well"""
        expected_result = AdClientResult(
            ConfigCheckResult([
                ConfigErrorProblem(config_error)
            ]),
            []
        )

        self.BASIC_CONFIG_INPUT['ldap_filter'] = 'asdf'
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_can_connect_to_host(self):
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [
                LdapHostResult(
                    '127.0.0.1',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.GOOD_CONNECT_TCP_RESULT,
                    LdapConnectResult(True, '127.0.0.1', 389),
                    self.GOOD_BIND_SERVICE_ACCOUNT,
                    self.GOOD_SEARCH_HAS_RESULTS),
                LdapHostResult(
                    '192.168.22.22',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.GOOD_CONNECT_TCP_RESULT,
                    LdapConnectResult(True, '192.168.22.22', 389),
                    self.GOOD_BIND_SERVICE_ACCOUNT,
                    self.GOOD_SEARCH_HAS_RESULTS),
            ])

        self.BASIC_CONFIG_INPUT['host_2'] = '192.168.22.22'
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)
        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_cant_connect(self):
        e = ConnectivityMockException()
        mock_connect_host = mock.Mock(side_effect=e)
        self.patch(ad_client, '_connect_host', mock_connect_host)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [LdapHostResult(
                '127.0.0.1',
                389,
                AD_TRANSPORT_CLEAR,
                self.GOOD_CONNECT_TCP_RESULT,
                LdapConnectResult(False, '127.0.0.1', 389, exception=e),
                UnmetPrerequisiteSkippedTestResult('bind', 'ldap connection'),
                UnmetPrerequisiteSkippedTestResult('search', 'ldap connection')
            )]
        )

        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(
            actual_result,
            expected_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_cant_bind(self):
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [LdapHostResult(
                '127.0.0.1',
                389,
                AD_TRANSPORT_CLEAR,
                self.GOOD_CONNECT_TCP_RESULT,
                LdapConnectResult(True, '127.0.0.1', 389),
                self.BAD_BIND_SERVICE_ACCOUNT,
                UnmetPrerequisiteSkippedTestResult('search', 'bind')
            )]
        )

        toolbox = MockConnectivityTestToolbox(fail_bind_service_account=True)
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=toolbox)

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_cant_bind_multiple_hosts(self):
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [
                LdapHostResult(
                    '127.0.0.1',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.GOOD_CONNECT_TCP_RESULT,
                    LdapConnectResult(True, '127.0.0.1', 389),
                    self.BAD_BIND_SERVICE_ACCOUNT,
                    UnmetPrerequisiteSkippedTestResult('search', 'bind')
                ),
                LdapHostResult(
                    '192.168.22.22',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.GOOD_CONNECT_TCP_RESULT,
                    LdapConnectResult(True, '192.168.22.22', 389),
                    self.BAD_BIND_SERVICE_ACCOUNT,
                    UnmetPrerequisiteSkippedTestResult('search', 'bind')
                )
            ]
        )

        self.BASIC_CONFIG_INPUT['host_2'] = '192.168.22.22'
        toolbox = MockConnectivityTestToolbox(fail_bind_service_account=True)
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=toolbox)

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_cant_search(self):
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [LdapHostResult(
                '127.0.0.1',
                389,
                AD_TRANSPORT_CLEAR,
                self.GOOD_CONNECT_TCP_RESULT,
                LdapConnectResult(True, '127.0.0.1', 389),
                self.GOOD_BIND_SERVICE_ACCOUNT,
                self.BAD_SEARCH_HAS_RESULTS
            )]
        )

        toolbox = MockConnectivityTestToolbox(fail_search_has_results=True)
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=toolbox)

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_mix_successful_connect(self):
        """If several hosts are listed in config test one working and one failing
        The first host will fail TCP check and also creating connected client
        The second host will pass all checks
        """
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        mock_can_connect_tcp = mock.Mock(side_effect=[self.BAD_CONNECT_TCP_RESULT, self.GOOD_CONNECT_TCP_RESULT])
        self.patch(MockConnectivityTestToolbox, 'test_connect_tcp', mock_can_connect_tcp)

        expected_result = AdClientResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            [
                LdapHostResult(
                    '127.0.0.1',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.BAD_CONNECT_TCP_RESULT,
                    UnmetPrerequisiteSkippedTestResult('ldap connection', 'tcp connection'),
                    UnmetPrerequisiteSkippedTestResult('bind', 'tcp connection'),
                    UnmetPrerequisiteSkippedTestResult('search', 'tcp connection'),
                ),
                LdapHostResult(
                    '192.168.22.22',
                    389,
                    AD_TRANSPORT_CLEAR,
                    self.GOOD_CONNECT_TCP_RESULT,
                    LdapConnectResult(True, '192.168.22.22', 389),
                    self.GOOD_BIND_SERVICE_ACCOUNT,
                    self.GOOD_SEARCH_HAS_RESULTS
                )
            ]
        )

        self.BASIC_CONFIG_INPUT['host_2'] = '192.168.22.22'
        actual_result = yield ad_client.check_ad_client(self.BASIC_CONFIG_INPUT, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))


if __name__ == '__main__':
    import unittest
    unittest.main()
