#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
from twisted.internet import defer
import mock
import os

from ...base import ModuleConnectivityTest, MockConnectivityTestToolbox
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.connectivity.check import http_proxy
from duoauthproxy.lib import const, util
from duoauthproxy.lib.connectivity.connectivity_results import (
    HttpProxyResult,
    ConfigCheckResult,
    MissingConfigKeyProblem,
    InvalidConfigKeyProblem,
    ConfigProblemSkippedTestResult,
    UnmetPrerequisiteSkippedTestResult,
)


class TestCheckHttpProxy(ModuleConnectivityTest):

    def setUp(self):
        super(TestCheckHttpProxy, self).setUp()

        self.GOOD_CONFIG = ConfigDict({
            'api_host': 'a.totally.legit.api.host',
        })

        self.GOOD_CONFIG_RESULT = HttpProxyResult(
            ConfigCheckResult([]),
            self.GOOD_LISTEN_TCP_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT
        )

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()

    @defer.inlineCallbacks
    def test_good_minimal_config(self):
        expected_result = self.GOOD_CONFIG_RESULT

        config = self.GOOD_CONFIG
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_good_maximal_config(self):
        expected_result = self.GOOD_CONFIG_RESULT

        config = ConfigDict({
            'api_host': 'this.is.so.legit.guys.super.serial',
            'port': '8080',
            'client_ip': '1.2.3.4,10.20.30.40,100.200.100.200'
        })
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_bad_config_missing_api_host(self):
        expected_result = HttpProxyResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('api_host')
            ]),
            self.GOOD_LISTEN_TCP_RESULT,
            ConfigProblemSkippedTestResult('ping', 'api_host'),
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping')
        )

        config = ConfigDict({})
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_bad_config_invalid_ip(self):
        expected_result = HttpProxyResult(
            ConfigCheckResult([
                InvalidConfigKeyProblem('client_ip', '100.200.300.400')
            ]),
            self.GOOD_LISTEN_TCP_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT
        )

        config = ConfigDict({
            'api_host': 'this.is.so.legit.guys.super.serial',
            'client_ip': '1.2.3.4,10.20.30.40,100.200.300.400'
        })
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_bad_api_host_unreachable(self,):
        expected_result = HttpProxyResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_LISTEN_TCP_RESULT,
            self.BAD_PING_RESULT,
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping')
        )
        toolbox = MockConnectivityTestToolbox(fail_ping=True)

        config = self.GOOD_CONFIG
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=toolbox)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_listen_failure(self):
        expected_result = HttpProxyResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.BAD_LISTEN_TCP_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT
        )
        toolbox = MockConnectivityTestToolbox(fail_listen_tcp=True)

        config = self.GOOD_CONFIG
        actual_result = yield http_proxy.check_http_proxy(config, toolbox=toolbox)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_ca_certs_default_used(self):
        """If no ca certs file is specified or passed down from main, Duo calls like ping
        should use an absolute path to a default ca certs bundle
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.GOOD_CONFIG

        yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', const.DEFAULT_HTTP_CERTS_FILE)
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_absolute_path(self):
        """Calls to Duo such as ping will resolve its http-ca-certs file path first.
        If that path is already absolute, it shouldn't change.
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        os_root = os.path.abspath(os.sep)
        cert_path = os.path.join(
            os_root,
            'tmp',
            'mycerts.pem',
        )

        config = self.GOOD_CONFIG
        config.update({'http_ca_certs_file': cert_path})

        yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            cert_path,
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_relative_path(self):
        """Duo calls like ping that use a relative CA cert path should resolve it
        to be absolute at $DUO_INSTALL_DIR/conf/cert.file during checking
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.GOOD_CONFIG
        config.update({'http_ca_certs_file': 'mycerts.pem'})

        yield http_proxy.check_http_proxy(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', 'mycerts.pem')
        )


if __name__ == '__main__':
    import unittest
    unittest.main()
