#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import mock
import os

from twisted.internet import defer

from ...base import ModuleConnectivityTest, MockConnectivityTestToolbox
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.connectivity.check import ldap_server_auto
from duoauthproxy.lib import const, util
from duoauthproxy.modules import ssl_server
from duoauthproxy.lib.connectivity.connectivity_results import (
    LdapServerResult,
    ConfigCheckResult,
    MissingConfigKeyProblem,
    ConfigProblemSkippedTestResult,
    UnmetPrerequisiteSkippedTestResult,
    BAD_PROTECTED_SKEY_RESULT,
    NOT_APPLICABLE_TEST_RESULT
)


class TestCheckLdapServerAuto(ModuleConnectivityTest):
    def setUp(self):
        super(TestCheckLdapServerAuto, self).setUp()
        self.BLANK_CONFIG = ConfigDict({})

        self.BASIC_CONFIG = ConfigDict({
            'ikey': 'thisisagoodikey',
            'skey': 'thisisagoodskey',
            'api_host': 'this.is.a.good.api.host',
            'client': 'ad_client',
        })

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()

    @defer.inlineCallbacks
    def test_good_minimal_config(self):
        """Tests a good config with the bare minimum (no ssl key/cert paths)
        """
        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        config = self.BASIC_CONFIG
        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_good_maximal_config(self):
        """Test a good maximal config that includes ssl_key_path and ssl_cert_path
        A manual mock is required in this case as opposed to a patch decorator due to
        twisted + mock weirdness with async functions
        """
        mock_create_ctx = mock.Mock()
        mock_create_ctx.return_value = None
        self.patch(ssl_server.ChainingOpenSSLContextFactory, "__init__", mock_create_ctx)

        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.GOOD_SSL_CREDS_RESULT,
            self.GOOD_LISTEN_SSL_RESULT
        )

        config = ConfigDict({
            'ikey': 'anothergoodikey',
            'skey': 'totesgoodskeyfam',
            'api_host': 'good-api.host.name',
            'client': 'ad_client',
            'factors': 'push,auto,phone,passcode',
            'delimiter': '}',
            'allow_concat': 'false',
            'api_timeout': '2001',
            'failmode': 'secure',
            'ssl_port': '636',
            'interface': '192.168.0.1',
            'pass_through_attr_names': 'NAS-Identifier',
            'pass_through_all': 'true',
            'exempt_username_1': 'exempty_mcexemptface',
            'exempt_username_2': 'flintheard_glomgold',
            'ssl_key_path': 'path/to/ssl/key',
            'ssl_cert_path': 'path/to/ssl/cert',
            'cipher_list': 'DEFAULT',
        })

        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )

        self.assertEqual(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_bad_api_host_validate_and_ping_failure(self):
        """A config with a bad api-host should fail when trying to reach out to Duo,
        ie. should fail Duo ping and creds validation
        """
        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.BAD_PING_RESULT,
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping'),
            self.BAD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT,
        )

        config = self.BASIC_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_ping=True, fail_validate_api_creds=True)
        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=toolbox
        )

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_bad_api_creds(self):
        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.BAD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        config = self.BASIC_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_validate_api_creds=True)
        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=toolbox
        )

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_listen_tcp_failure(self):
        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.BAD_LISTEN_TCP_RESULT,
        )

        config = self.BASIC_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_listen_tcp=True)
        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=toolbox)

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_listen_ssl_failure(self):
        """A config with a bad SSL key and cert path should fail on SSL creds validation,
        and therefore skip listening over SSL
        """
        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.BAD_SSL_CREDS_RESULT,
            UnmetPrerequisiteSkippedTestResult('listen', 'ssl configuration')
        )

        config = ConfigDict({
            'ikey': 'thisisagoodikey',
            'skey': 'thisisagoodskey',
            'api_host': 'this.is.a.good.api.host',
            'client': 'ad_client',
            'ssl_key_path': 'path/to/bad/ssl/key',
            'ssl_cert_path': 'path/to/bad/ssl/cert',
        })
        toolbox = MockConnectivityTestToolbox(
            fail_test_ssl_creds=True,
            fail_listen_ssl=True,
        )
        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=toolbox
        )

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_blank_config(self):
        expected_result = LdapServerResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('ikey'),
                MissingConfigKeyProblem('api_host'),
                MissingConfigKeyProblem('skey / skey_protected')
            ]),
            ConfigProblemSkippedTestResult('ping', 'api_host'),
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping'),
            ConfigProblemSkippedTestResult('validate api credentials', 'api credentials'),
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            self.BLANK_CONFIG,
            toolbox=self.SUCCESS_TOOLBOX
        )

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_ssl_port_default_used(self):
        """If a config specifies ssl_key_path and ssl_cert_path, it should
        fall back to the AP's default LDAPS port if ssl_port isn't specified,
        even if the regular port option is specified
        """
        mock_create_ctx = mock.Mock()
        mock_create_ctx.return_value = None
        self.patch(ssl_server.ChainingOpenSSLContextFactory, "__init__", mock_create_ctx)

        config = ConfigDict({
            'ikey': 'anothergoodikey',
            'skey': 'totesgoodskeyfam',
            'api_host': 'good-api.host.name',
            'client': 'ad_client',
            'ssl_key_path': 'path/to/ssl/key',
            'ssl_cert_path': 'path/to/ssl/cert',
            'port': '389',
            'interface': '',
        })

        mock_test_listen_ssl = mock.Mock(return_value=self.GOOD_LISTEN_SSL_RESULT)
        self.patch(MockConnectivityTestToolbox, 'test_listen_ssl', mock_test_listen_ssl)

        yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )
        mock_test_listen_ssl.assert_called_with(
            const.DEFAULT_LDAPS_PORT,
            mock.ANY,
            config['interface'],
        )

    @defer.inlineCallbacks
    def test_non_ssl_port_default_used(self):
        """Even if ssl_port is specified for a non-SSL config, the default
        LDAP (not LDAPS) port should be used
        """
        config = ConfigDict({
            'ikey': 'anothergoodikey',
            'skey': 'totesgoodskeyfam',
            'api_host': 'good-api.host.name',
            'client': 'ad_client',
            'ssl_port': '789',
            'interface': '',
        })

        mock_test_listen_tcp = mock.Mock(return_value=self.GOOD_LISTEN_TCP_RESULT)
        self.patch(MockConnectivityTestToolbox, 'test_listen_tcp', mock_test_listen_tcp)

        yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )
        mock_test_listen_tcp.assert_called_with(
            const.DEFAULT_LDAP_PORT,
            config['interface'],
        )

    @defer.inlineCallbacks
    def test_ssl_key_and_cert_paths(self):
        mock_test_ssl_credentials = mock.Mock(return_value=self.GOOD_SSL_CREDS_RESULT)
        self.patch(MockConnectivityTestToolbox, 'test_ssl_credentials', mock_test_ssl_credentials)
        mock_create_ctx = mock.Mock()
        mock_create_ctx.return_value = None
        self.patch(ssl_server.ChainingOpenSSLContextFactory, "__init__", mock_create_ctx)

        config = ConfigDict({
            'ikey': 'anothergoodikey',
            'skey': 'totesgoodskeyfam',
            'api_host': 'good-api.host.name',
            'client': 'ad_client',
            'ssl_key_path': 'key',
            'ssl_cert_path': 'cert',
        })
        yield ldap_server_auto.check_ldap_server_auto(config, toolbox=self.SUCCESS_TOOLBOX)

        mock_test_ssl_credentials.assert_called_with(
                util.resolve_file_path('key'),
                util.resolve_file_path('cert'),
                'DEFAULT')

    @defer.inlineCallbacks
    def test_config_missing_ssl_cert(self):
        """An ldap_server_auto section with ssl_key_path needs ssl_cert_path to be present
        """
        expected_result = LdapServerResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('ssl_cert_path')
            ]),
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        config = self.BASIC_CONFIG
        config.update({'ssl_key_path': 'path/to/ssl/key'})

        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )
        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_config_missing_ssl_key(self):
        """An ldap_server_auto section with ssl_cert_path needs ssl_key_path to be present
        """
        expected_result = LdapServerResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('ssl_key_path')
            ]),
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        config = self.BASIC_CONFIG
        config.update({'ssl_cert_path': 'path/to/ssl/cert'})

        actual_result = yield ldap_server_auto.check_ldap_server_auto(
            config,
            toolbox=self.SUCCESS_TOOLBOX
        )
        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    def test_protected_config(self):
        """A good config with skey_protected instead of skey should check out fine
        """
        expected_result = self.GOOD_CONFIG_CHECK_RESULT

        config = ConfigDict({
            'ikey': 'thisisagoodikey',
            'skey_protected': 'thisisagoodskey',
            'api_host': 'this.is.a.good.api.host',
            'client': 'ad_client',
        })

        actual_result = ldap_server_auto.validate_ldap_server_auto_config(config)

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_skey_fail_decrypt(self):
        mock_get_protected_str = mock.Mock(return_value=None)
        config = self.BASIC_CONFIG
        config.get_protected_str = mock_get_protected_str

        expected_result = LdapServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            BAD_PROTECTED_SKEY_RESULT,
            NOT_APPLICABLE_TEST_RESULT,
            self.GOOD_LISTEN_TCP_RESULT
        )

        actual_result = yield ldap_server_auto.check_ldap_server_auto(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEquals(
            expected_result,
            actual_result,
            "{0} should equal {1}".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_ca_certs_default(self):
        """Calls to Duo like ping use the http-ca-certs file passed down by main section;
        if no certs file was specified, it should use our default one during checking
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.BASIC_CONFIG

        yield ldap_server_auto.check_ldap_server_auto(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', const.DEFAULT_HTTP_CERTS_FILE)
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_absolute_path(self):
        """Calls to Duo such as ping will resolve its http-ca-certs file path first.
        If that path is already absolute, it shouldn't change.
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        os_root = os.path.abspath(os.sep)
        cert_path = os.path.join(
            os_root,
            'tmp',
            'mycerts.pem',
        )

        config = self.BASIC_CONFIG
        config.update({'http_ca_certs_file': cert_path})

        yield ldap_server_auto.check_ldap_server_auto(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            cert_path,
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_relative_path(self):
        """Duo calls like ping that use a relative CA cert path should resolve it
        to be absolute at $DUO_INSTALL_DIR/conf/cert.file during checking
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.BASIC_CONFIG
        config.update({'http_ca_certs_file': 'mycerts.pem'})

        yield ldap_server_auto.check_ldap_server_auto(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', 'mycerts.pem')
        )


if __name__ == '__main__':
    import unittest
    unittest.main()
