# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

import mock
import os

from twisted.internet import defer

from ...base import ModuleConnectivityTest, MockConnectivityTestToolbox
from duoauthproxy.lib import util, const
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.connectivity.check import main


class TestCheckMain(ModuleConnectivityTest):

    @defer.inlineCallbacks
    def test_blank_main_success(self):
        """An empty main config should return a success. It should also default to using
        an absolute path to Duo's provided http ca certs file.
        """
        mock_test_certs = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ssl_certs', mock_test_certs)

        config = ConfigDict({})
        yield main.check_main(config, self.SUCCESS_TOOLBOX)

        mock_test_certs.assert_called_with(
            os.path.join(util.get_home_dir(), 'conf', const.DEFAULT_HTTP_CERTS_FILE)
        )

    @defer.inlineCallbacks
    def test_relative_main_cert_file_specified(self):
        """If you specify a relative path to your http ca certs, this path should
        be resolved into an absolute one during the check
        """
        mock_test_certs = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ssl_certs', mock_test_certs)
        file_path = 'file1.crt'

        config = ConfigDict({'http_ca_certs_file': file_path})
        yield main.check_main(config, self.SUCCESS_TOOLBOX)

        mock_test_certs.assert_called_with(
            os.path.join(util.get_home_dir(), 'conf', file_path)
        )

    @defer.inlineCallbacks
    def test_absolute_main_cert_file_specified(self):
        """If you specify an absolute path to your http ca certs, this path should
        remain unchanged when passed around during the check
        """
        mock_test_certs = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ssl_certs', mock_test_certs)

        os_root = os.path.abspath(os.sep)
        cert_path = os.path.join(
            os_root,
            'tmp',
            'mycerts.pem',
        )
        config = ConfigDict({'http_ca_certs_file': cert_path})
        yield main.check_main(config, self.SUCCESS_TOOLBOX)

        mock_test_certs.assert_called_with(cert_path)

    @defer.inlineCallbacks
    def test_good_cert_file_success(self):
        config = ConfigDict({'http_ca_certs_file': 'good_file.crt'})

        result = yield main.check_main(config, self.SUCCESS_TOOLBOX)

        self.assertEqual(result.certs_result, self.GOOD_CERTS_RESULT)

    @defer.inlineCallbacks
    def test_bad_cert_file_failure(self):
        config = ConfigDict({'http_ca_certs_file': 'bad_file.crt'})

        result = yield main.check_main(config, MockConnectivityTestToolbox(fail_ssl_certs=True))

        self.assertEqual(result.certs_result, self.BAD_CERTS_RESULT)

    @defer.inlineCallbacks
    def test_http_proxy_success(self):
        """Test communicating to Duo through a http_proxy"""
        config = ConfigDict({'http_proxy_host': '1.2.3.4', 'http_proxy_port': '80'})

        result = yield main.check_main(config, self.SUCCESS_TOOLBOX)

        self.assertEqual(result.http_proxy_result, self.GOOD_CONNECT_WITH_HTTP_PROXY_RESULT)

    @defer.inlineCallbacks
    def test_http_proxy_failure(self):
        """Test communicating to Duo through a http_proxy"""
        config = ConfigDict({'http_proxy_host': 'bad_host', 'http_proxy_port': '80'})

        result = yield main.check_main(config, MockConnectivityTestToolbox(fail_connect_with_http_proxy=True))

        self.assertEqual(result.http_proxy_result, self.BAD_CONNECT_WITH_HTTP_PROXY_RESULT)


if __name__ == '__main__':
    import unittest
    unittest.main()
