#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import os
import copy

from twisted.internet import defer
import mock

from ...base import ModuleConnectivityTest, MockConnectivityTestToolbox
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.connectivity.check import radius_server
from duoauthproxy.lib import const, util
from duoauthproxy.lib.connectivity.connectivity_results import (
    RadiusServerResult,
    ConfigCheckResult,
    MissingConfigKeyProblem,
    InvalidConfigKeyProblem,
    ConfigProblemSkippedTestResult,
    UnmetPrerequisiteSkippedTestResult,
    BAD_PROTECTED_SKEY_RESULT
)


class TestCheckRadiusServer(ModuleConnectivityTest):

    def setUp(self):
        super(TestCheckRadiusServer, self).setUp()

        self.BLANK_CONFIG = ConfigDict({})

        self.GOOD_CONFIG = ConfigDict({
            'ikey': 'thisisagoodikey',
            'skey': 'thisisagoodskey',
            'api_host': 'this.is.a.good.api.host',
            'radius_ip_1': '1.2.3.4',
            'radius_secret_1': 'secret',
            'client': 'radius_client',
        })

        self.GOOD_SECTION_RESULT = RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.GOOD_LISTEN_UDP_RESULT,
        )

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()

    @defer.inlineCallbacks
    def test_good_minimal_config(self):
        expected_result = self.GOOD_SECTION_RESULT

        config = self.GOOD_CONFIG
        actual_result = yield radius_server.check_radius_server(
            config,
            toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_good_maximal_config(self):
        expected_result = self.GOOD_SECTION_RESULT

        config = ConfigDict({
            'ikey': 'anothergoodikey',
            'skey': 'totesgoodskeyfam',
            'api_host': 'good-api.host.name',
            'radius_ip_1': '1.2.3.4',
            'radius_secret_protected_1': '5678',
            'radius_ip_2': '10.20.30.40',
            'radius_secret_2': '8675309',
            'radius_ip_3': '10.10.10.10',
            'radius_secret_3': 'TEN',
            'client': 'some_client',
            'factors': 'push,auto,phone,passcode',
            'delimiter': '}',
            'allow_concat': 'false',
            'api_timeout': '2001',
            'failmode': 'secure',
            'port': '2812',
            'interface': '192.168.0.1',
            'pass_through_attr_names': 'NAS-Identifier',
            'pass_through_all': 'true',
            'exempt_username_1': 'exempty_mcexemptface',
            'exempt_username_2': 'flintheard_glomgold',
        })
        actual_result = yield radius_server.check_radius_server(
            config,
            toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_bad_api_host_unreachable(self):
        expected_result = RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.BAD_PING_RESULT,
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping'),
            self.BAD_VALIDATE_RESULT,
            self.GOOD_LISTEN_UDP_RESULT,
        )

        config = self.GOOD_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_ping=True, fail_validate_api_creds=True)
        actual_result = yield radius_server.check_radius_server(
            config,
            toolbox=toolbox)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_bad_api_creds(self):
        expected_result = RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.BAD_VALIDATE_RESULT,
            self.GOOD_LISTEN_UDP_RESULT,
        )

        config = self.GOOD_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_validate_api_creds=True)
        actual_result = yield radius_server.check_radius_server(
            config,
            toolbox=toolbox)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_listen_failure(self):
        expected_result = RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.BAD_LISTEN_UDP_RESULT,
        )

        config = self.GOOD_CONFIG
        toolbox = MockConnectivityTestToolbox(fail_listen_udp=True)
        actual_result = yield radius_server.check_radius_server(
            config,
            toolbox=toolbox)

        self.assertEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_blank_config(self):
        expected_result = RadiusServerResult(
            ConfigCheckResult([
                MissingConfigKeyProblem('ikey'),
                MissingConfigKeyProblem('api_host'),
                MissingConfigKeyProblem('radius_ip_1'),
                MissingConfigKeyProblem('skey / skey_protected'),
                MissingConfigKeyProblem('radius_secret_1 / radius_secret_protected_1'),

            ]),
            ConfigProblemSkippedTestResult('ping', 'api_host'),
            UnmetPrerequisiteSkippedTestResult('time drift', 'ping'),
            ConfigProblemSkippedTestResult('validate api credentials', 'api credentials'),
            self.GOOD_LISTEN_UDP_RESULT,
        )

        actual_result = yield radius_server.check_radius_server(
            self.BLANK_CONFIG,
            toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result, "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    def test_protected_config(self):
        expected_result = self.GOOD_CONFIG_CHECK_RESULT

        config = ConfigDict({
            'ikey': 'thisisagoodikey',
            'skey_protected': 'thisisagoodskey',
            'api_host': 'this.is.a.good.api.host',
            'radius_ip_1': '1.2.3.4',
            'radius_secret_protected_1': 'secret',
            'client': 'radius_client',
        })

        actual_result = radius_server.validate_radius_server_config(config)

        self.assertEqual(expected_result, actual_result)

    def test_ip_validation_success(self):
        expected_result = self.GOOD_CONFIG_CHECK_RESULT

        config = copy.deepcopy(self.GOOD_CONFIG)
        config.update({'radius_ip_2': '10.20.30.0/24',
                       'radius_secret_2': 'secret',
                       'radius_ip_3': '100.200.100.200-100.200.200.200',
                       'radius_secret_3': 'secret'})

        actual_result = radius_server.validate_radius_server_config(config)

        self.assertEqual(expected_result, actual_result)

    def test_ip_validation_failure(self):
        expected_result = ConfigCheckResult([
                InvalidConfigKeyProblem('radius_ip_2', '10.20.30.400/60'),
                InvalidConfigKeyProblem('radius_ip_3', '100.100.100.100-100.200.300.400'),
            ])

        config = copy.deepcopy(self.GOOD_CONFIG)
        config.update({'radius_ip_2': '10.20.30.400/60',
                       'radius_secret_2': 'secret',
                       'radius_ip_3': '100.100.100.100-100.200.300.400',
                       'radius_secret_3': 'secret'})

        actual_result = radius_server.validate_radius_server_config(config)

        self.assertEqual(expected_result, actual_result, "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    @defer.inlineCallbacks
    def test_time_drift(self):
        mock_test_time_drift = mock.Mock(return_value=self.GOOD_TIME_DRIFT_RESULT)
        self.patch(MockConnectivityTestToolbox, 'test_time_drift', mock_test_time_drift)

        config = self.GOOD_CONFIG
        yield radius_server.check_radius_server(config, toolbox=self.SUCCESS_TOOLBOX)
        mock_test_time_drift.assert_called()

    @defer.inlineCallbacks
    def test_ca_certs_default_used(self):
        """If no ca certs file is specified or passed down from main, Duo calls like ping
        should use an absolute path to a default ca certs bundle"""
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.GOOD_CONFIG

        yield radius_server.check_radius_server(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', const.DEFAULT_HTTP_CERTS_FILE)
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_absolute_path(self):
        """Calls to Duo such as ping will resolve its http-ca-certs file path first.
        If that path is already absolute, it shouldn't change.
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        os_root = os.path.abspath(os.sep)
        cert_path = os.path.join(
            os_root,
            'tmp',
            'mycerts.pem',
        )

        config = self.GOOD_CONFIG
        config.update({'http_ca_certs_file': cert_path})

        yield radius_server.check_radius_server(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            cert_path,
        )

    @defer.inlineCallbacks
    def test_ca_certs_defined_relative_path(self):
        """Duo calls like ping that use a relative CA cert path should resolve it
        to be absolute at $DUO_INSTALL_DIR/conf/cert.file during checking
        """
        duo_client_mock = mock.Mock()
        self.patch(MockConnectivityTestToolbox, 'test_ping_duo', duo_client_mock)

        config = self.GOOD_CONFIG
        config.update({'http_ca_certs_file': 'mycerts.pem'})

        yield radius_server.check_radius_server(config, toolbox=self.SUCCESS_TOOLBOX)

        args, kwargs = duo_client_mock.call_args
        duo_client_auth = args[0]

        duo_client_mock.assert_called()
        self.assertEqual(
            duo_client_auth.ca_certs,
            os.path.join(util.get_home_dir(), 'conf', 'mycerts.pem')
        )

    @defer.inlineCallbacks
    def test_skey_fail_decrypt(self):
        mock_get_protected_str = mock.Mock(return_value=None)
        config = self.GOOD_CONFIG
        config.get_protected_str = mock_get_protected_str

        expected_result = RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            BAD_PROTECTED_SKEY_RESULT,
            self.GOOD_LISTEN_UDP_RESULT,
        )

        actual_result = yield radius_server.check_radius_server(config, toolbox=self.SUCCESS_TOOLBOX)

        self.assertEqual(expected_result, actual_result)


if __name__ == '__main__':
    import unittest
    unittest.main()
