#
# Copyright (c) 2017 Duo Security
# All Rights Reserved
# All Wrongs Reversed
#

import unittest
import socket
import mock

from duoauthproxy.lib.connectivity.connect import duo


class DuoConnectTestCase(unittest.TestCase):
    MOCKED_TIME_VALUE = 1513961155

    def setUp(self):
        """Set up mock for the duo client. This object will mock results returned
        from Duo Cloud"""
        self.duo_client_mock = mock.Mock()


class TestPing(DuoConnectTestCase):
    @mock.patch('timeit.default_timer')
    def test_can_ping_duo_success(self, timer_mock):
        """A successful pong from Duo is the server timestamp so, we will force
        that return value"""
        self.duo_client_mock.ping.return_value = {"time": 1}

        # set timer mock to return start and end values in seconds
        timer_mock.side_effect = [99.25, 100]

        can_ping = duo.can_ping_duo(self.duo_client_mock)

        self.assertTrue(can_ping.is_successful())
        self.assertEqual(can_ping.latency, 750)

    def test_can_ping_duo_failure(self):
        """Duo client python raises RuntimeErrors on failed API calls so, we will force
        that side effect"""
        self.duo_client_mock.ping.side_effect = RuntimeError('failed')

        can_ping = duo.can_ping_duo(self.duo_client_mock)
        self.assertFalse(can_ping.is_successful())

    def test_can_ping_duo_socketerror(self):
        """If a badly formatted/unresolvable api-host is passed in, duo_client can
        throw an uncaught socket.gaierror exception in _make_request during the call
        to check(), so we have to catch this ourselves."""
        self.duo_client_mock.ping.side_effect = socket.gaierror('failed')

        can_ping = duo.can_ping_duo(self.duo_client_mock)
        self.assertFalse(can_ping.is_successful())


class TestValidateDuoCreds(DuoConnectTestCase):
    def test_can_validate_duo_creds_success(self):
        """A successful call to Duo's /check is the server timestamp, so we'll force
        that return value."""
        self.duo_client_mock.check.side_effect = {"time": 1}

        valid_creds = duo.can_validate_duo_creds(self.duo_client_mock)
        self.assertTrue(valid_creds.is_successful())

    def test_can_validate_duo_creds_failure_runtimeerror(self):
        """duo_client_python raises RuntimeErrors on failed API calls, so we'll force
        that side effect."""
        self.duo_client_mock.check.side_effect = RuntimeError('failed')

        valid_creds = duo.can_validate_duo_creds(self.duo_client_mock)
        self.assertFalse(valid_creds.is_successful())

    def test_can_validate_duo_creds_failure_socketerror(self):
        """If a badly formatted/unresolvable api-host is passed in, duo_client can
        throw an uncaught socket.gaierror exception in _make_request during the call
        to check(), so we have to catch this ourselves."""
        self.duo_client_mock.check.side_effect = socket.gaierror('failed')

        valid_creds = duo.can_validate_duo_creds(self.duo_client_mock)
        self.assertFalse(valid_creds.is_successful())


class TestTimeDriftSuccess(DuoConnectTestCase):
    """Success cases for detecting the time drift between the Duo cloud and the
    server running the Auth Proxy."""

    @mock.patch('timeit.default_timer')
    def test_has_acceptable_time_drift_success(self, mock_time):
        self.duo_client_mock.ping.return_value = {"time": self.MOCKED_TIME_VALUE}
        mock_time.return_value = self.MOCKED_TIME_VALUE

        response = duo.has_acceptable_time_drift(self.duo_client_mock)

        self.assertTrue(response.is_successful())
        self.assertEqual(response.drift, 0)


class TestTimeDriftFailure(DuoConnectTestCase):
    """Failure cases for detecting the time drift between the Duo cloud and the
    server running the Auth Proxy."""

    @mock.patch('timeit.default_timer')
    def test_time_drift_failure(self, mock_time):
        self.duo_client_mock.ping.return_value = {"time": self.MOCKED_TIME_VALUE}

        mock_time.return_value = self.MOCKED_TIME_VALUE + duo.MAX_TIME_DRIFT_IN_SECONDS + 1

        response = duo.has_acceptable_time_drift(self.duo_client_mock)

        self.assertFalse(response.is_successful())
        self.assertEqual(response.drift, duo.MAX_TIME_DRIFT_IN_SECONDS + 1)

    @mock.patch('timeit.default_timer')
    def test_unreachable_duo_cloud_host_not_found(self, mock_time):
        """Setup mock for a duo_client which will raise an exception which
        mimics not being able to contact Duo Cloud"""
        self.duo_client_mock.ping.side_effect = socket.gaierror
        mock_time.return_value = 0
        response = duo.has_acceptable_time_drift(self.duo_client_mock)

        self.assertFalse(response.is_successful())

    @mock.patch('timeit.default_timer')
    def test_unreachable_duo_cloud_runtime(self, mock_time):
        """Setup mock for a duo_client which will raise an exception which
        mimics not being able to contact Duo Cloud"""
        self.duo_client_mock.ping.side_effect = RuntimeError
        mock_time.return_value = 0
        response = duo.has_acceptable_time_drift(self.duo_client_mock)

        self.assertFalse(response.is_successful())


class TestApiTimeout(DuoConnectTestCase):
    def test_api_timeout(self):
        exc = socket.timeout
        self.duo_client_mock.ping.side_effect = exc

        ping_result = duo.can_ping_duo(self.duo_client_mock)
        self.assertFalse(ping_result.is_successful())
        self.assertNotEqual(ping_result.exception, None)


if __name__ == '__main__':
    import unittest
    unittest.main()
