#
# Copyright (c) 2017 Duo Security
# All Rights Reserved
# All Wrongs Reversed
#
from twisted.internet import defer
from twisted.internet.error import DNSLookupError, UnknownHostError, ConnectionRefusedError
from twisted.python.failure import Failure
import pyrad
import mock

from duoauthproxy.lib.connectivity.connect import ip
from ...base import TestCase


class TestConnectTCP(TestCase):
    @defer.inlineCallbacks
    def test_can_connect_tcp(self):
        def mock_establish_connection(endpoint):
            d = defer.Deferred()
            d.callback(True)
            return d

        can_connect = yield ip.can_connect_tcp('duo.com', 80, mock_establish_connection)
        self.assertTrue(can_connect.is_successful())

    @defer.inlineCallbacks
    def test_unknown_host(self):
        def mock_establish_connection(endpoint):
            d = defer.Deferred()
            d.errback(Failure(UnknownHostError('Cannot identify host')))
            return d

        can_connect = yield ip.can_connect_tcp('duo.com', 80, mock_establish_connection)
        self.assertFalse(can_connect.is_successful())

    @defer.inlineCallbacks
    def test_connection_refused(self):
        def mock_establish_connection(endpoint):
            d = defer.Deferred()
            d.errback(Failure(ConnectionRefusedError('Connection refused by other side')))
            return d

        can_connect = yield ip.can_connect_tcp('duo.com', 80, mock_establish_connection)
        self.assertFalse(can_connect.is_successful())

    @defer.inlineCallbacks
    def test_cannot_resolve_dns(self):
        def mock_establish_connection(endpoint):
            d = defer.Deferred()
            d.errback(Failure(DNSLookupError('Could not resolve hostname')))
            return d

        can_connect = yield ip.can_connect_tcp('duo.com', 80, mock_establish_connection)
        self.assertFalse(can_connect.is_successful())


class TestConnectRadius(TestCase):
    def test_can_connect_radius(self):
        client = mock.Mock()
        client.SendPacket.return_value = 'got reply'
        can_connect = ip.can_connect_radius(client, 'mock_packet')

        self.assertTrue(can_connect.is_successful())

    def test_timeout(self):
        client = mock.Mock()
        client.SendPacket.side_effect = pyrad.client.Timeout("Timed out")
        can_connect = ip.can_connect_radius(client, 'mock_packet')

        self.assertFalse(can_connect.is_successful())


if __name__ == '__main__':
    import unittest
    unittest.main()
