#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import mock

from ldaptor.protocols.ldap import ldapclient
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry
from twisted.internet import defer

from duoauthproxy.lib.connectivity.connect import ldap_operations
from ...base import TestCase, MockProtocol


class TestFilterSearch(TestCase):
    # A dummy LDAPClient
    LDAP_CLIENT = ldapclient.LDAPClient()

    # A dummy search_dn
    SEARCH_DN = 'dc=duo,dc=local'

    # A sample well-formed filter text
    FILTER_TEXT = '(|(&(objectClass=user)(objectCategory=person))(objectClass=inetOrgPerson)(objectClass=organizationalPerson))'

    @mock.patch.object(LDAPEntry, 'search')
    @defer.inlineCallbacks
    def test_filter_no_result(self, mock_search):
        mock_search.return_value = []

        search_test = yield ldap_operations.ldap_search_has_results(self.LDAP_CLIENT, self.SEARCH_DN, self.FILTER_TEXT)

        self.assertFalse(search_test.is_successful(), "Search should not have any result")

    @mock.patch.object(LDAPEntry, 'search')
    @defer.inlineCallbacks
    def test_filter_some_result(self, mock_search):
        mock_search.return_value = [{}, {}, {}]

        search_test = yield ldap_operations.ldap_search_has_results(self.LDAP_CLIENT, self.SEARCH_DN, self.FILTER_TEXT)

        self.assertTrue(search_test.is_successful(), "Search should have a result")

    @mock.patch.object(LDAPEntry, 'search')
    @defer.inlineCallbacks
    def test_filter_search_error(self, mock_search):
        mock_search.side_effect = Exception()

        search_test = yield ldap_operations.ldap_search_has_results(self.LDAP_CLIENT, self.SEARCH_DN, self.FILTER_TEXT)

        self.assertIsNotNone(search_test.get_exception(), "Search should have raised an Exception.")

    def test_validate_filter_bad_filter(self):
        self.assertFalse(ldap_operations.validate_filter("badly formatted filter").is_successful())


class TestBindServiceAccount(TestCase):
    @defer.inlineCallbacks
    def test_successful_bind(self):
        mock_protocol = MockProtocol(bind_success=True)
        bind_test = yield ldap_operations.can_bind_service_account(mock_protocol)
        self.assertTrue(bind_test.is_successful())

    @defer.inlineCallbacks
    def test_failed_bind(self):
        mock_protocol = MockProtocol(bind_success=False)
        bind_test = yield ldap_operations.can_bind_service_account(mock_protocol)
        self.assertFalse(bind_test.is_successful())

    @defer.inlineCallbacks
    def test_attempt_bind_good(self):
        mock_protocol = MockProtocol(bind_success=True)
        bind_result = yield ldap_operations._attempt_bind(mock_protocol)
        self.assertTrue(bind_result)

    @defer.inlineCallbacks
    def test_attempt_bind_failed(self):
        mock_protocol = MockProtocol(bind_success=False)
        with self.assertRaises(ldaperrors.LDAPException):
            yield ldap_operations._attempt_bind(mock_protocol)

    @defer.inlineCallbacks
    def test_keep_bound(self):
        mock_protocol = MockProtocol(bind_success=True)
        yield ldap_operations.can_bind_service_account(mock_protocol, keep_bound=True)
        self.assertTrue(mock_protocol.bound)

    @defer.inlineCallbacks
    def test_bind_cleanup(self):
        mock_protocol = MockProtocol(bind_success=True)
        yield ldap_operations.can_bind_service_account(mock_protocol, keep_bound=False)
        self.assertFalse(mock_protocol.bound)


if __name__ == '__main__':
    import unittest
    unittest.main()
