#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
# All Wrongs Reversed
#
from collections import namedtuple

from twisted.internet import defer
from twisted.internet.error import CannotListenError
import psutil
import mock

from duoauthproxy.lib.connectivity.connect import listen
from duoauthproxy.lib.connectivity.connectivity_results import ListenResult, AuthproxyUsingPortSkippedTestResult
from ...base import TestCase


class MockListener(object):
    def stopListening(self):
        pass


class TestListen(TestCase):
    @mock.patch('twisted.internet.reactor.listenTCP')
    @defer.inlineCallbacks
    def test_can_listen_tcp(self, mock_listen_tcp):
        mock_listen_tcp.return_value = MockListener()

        can_listen = yield listen.can_listen_tcp(1812)
        self.assertTrue(can_listen.is_successful())

    @mock.patch('twisted.internet.reactor.listenTCP')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._reason_port_is_used')
    @defer.inlineCallbacks
    def test_cannot_listen_tcp(self, mock_reason, mock_listen_tcp):
        mock_listen_tcp.side_effect = CannotListenError('interface', 'port', 'socket_error')
        mock_reason.return_value = ListenResult(False, 1812, '')

        can_listen = yield listen.can_listen_tcp(1812)
        self.assertFalse(can_listen.is_successful())

    @mock.patch('twisted.internet.reactor.listenUDP')
    @defer.inlineCallbacks
    def test_can_listen_udp(self, mock_listen_udp):
        mock_listen_udp.return_value = MockListener()

        can_listen = yield listen.can_listen_udp(1812)
        self.assertTrue(can_listen.is_successful())

    @mock.patch('twisted.internet.reactor.listenUDP')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._reason_port_is_used')
    @defer.inlineCallbacks
    def test_cannot_listen_udp(self, mock_reason, mock_listen_udp):
        mock_listen_udp.side_effect = CannotListenError('interface', 'port', 'socket_error')
        mock_reason.return_value = ListenResult(False, 1812, '')

        can_listen = yield listen.can_listen_udp(1812)
        self.assertFalse(can_listen.is_successful())

    @mock.patch('twisted.internet.reactor.listenSSL')
    @defer.inlineCallbacks
    def test_can_listen_ssl(self, mock_listen_ssl):
        mock_listen_ssl.return_value = MockListener()

        can_listen = yield listen.can_listen_ssl(1812, 'ssl_ctx_factory')
        self.assertTrue(can_listen.is_successful())

    @mock.patch('twisted.internet.reactor.listenSSL')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._reason_port_is_used')
    @defer.inlineCallbacks
    def test_cannot_listen_ssl(self, mock_reason, mock_listen_ssl):
        mock_listen_ssl.side_effect = CannotListenError('interface', 'port', 'socket_error')
        mock_reason.return_value = ListenResult(False, 1812, '')

        can_listen = yield listen.can_listen_ssl(1812, 'ssl_ctx_factory')
        self.assertFalse(can_listen.is_successful())


class TestGetListenerPid(TestCase):
    def setUp(self):
        sconn = namedtuple('sconn', 'laddr pid')
        addr = namedtuple('addr', 'ip port')
        self.basic_net_connections = [
                sconn(laddr=addr(ip='127.0.0.1', port=1812), pid=1),
                sconn(laddr=addr(ip='127.0.0.1', port=2812), pid=2),
                sconn(laddr=addr(ip='0.0.0.0', port=389), pid=3),
                sconn(laddr=addr(ip='127.0.0.1', port=3000), pid=None),
        ]

    @mock.patch('psutil.net_connections')
    def test_no_listener(self, mock_net_connections):
        mock_net_connections.return_value = self.basic_net_connections

        result = listen._get_listener_pid(1234, '127.0.0.1', 'mock_kind')

        self.assertIsNone(result)

    @mock.patch('psutil.net_connections')
    def test_regular_listener(self, mock_net_connections):
        mock_net_connections.return_value = self.basic_net_connections
        expected_result = 1

        actual_result = listen._get_listener_pid(1812, '127.0.0.1', 'mock_kind')

        self.assertEqual(expected_result, actual_result)

    @mock.patch('psutil.net_connections')
    def test_unknown_listener(self, mock_net_connections):
        mock_net_connections.return_value = self.basic_net_connections
        expected_result = -1

        actual_result = listen._get_listener_pid(3000, '127.0.0.1', 'mock_kind')

        self.assertEqual(expected_result, actual_result)

    @mock.patch('psutil.net_connections')
    def test_all_interfaces(self, mock_net_connections):
        mock_net_connections.return_value = self.basic_net_connections
        expected_result = 3

        actual_result = listen._get_listener_pid(389, '', 'mock_kind')

        self.assertEqual(expected_result, actual_result)


class TestGetListenerName(TestCase):
    def setUp(self):
        class MockProcObj(object):
            def name(self):
                return 'twistd'

        self.mockProcObj = MockProcObj()

    @mock.patch('psutil.Process')
    def test_get_listener(self, mock_process):
        mock_process.return_value = self.mockProcObj
        expected_result = 'twistd'

        actual_result = listen._get_listener_name(2)

        self.assertEqual(expected_result, actual_result)

    def test_pid_not_found(self):
        expected_result = "Unknown Process"

        actual_result = listen._get_listener_name(-1)

        self.assertEqual(expected_result, actual_result)

    @mock.patch('psutil.Process')
    def test_process_finder_fails(self, mock_process):
        mock_process.side_effect = psutil.NoSuchProcess('no match for pid')
        expected_result = "Unknown Process"

        actual_result = listen._get_listener_name(7)

        self.assertEqual(expected_result, actual_result)


class TestReasonPortIsUsed(TestCase):
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_pid')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_name')
    def test_authproxy_is_using_port_linux(self, mock_name, mock_pid):
        mock_pid.return_value = 1
        mock_name.return_value = 'twistd'
        expected_result = AuthproxyUsingPortSkippedTestResult('listen tcp')

        actual_result = listen._reason_port_is_used(1234, '', None, 'tcp')
        self.assertEqual(expected_result, actual_result)

    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_pid')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_name')
    def test_authproxy_is_using_port_windows(self, mock_name, mock_pid):
        mock_pid.return_value = 1
        mock_name.return_value = 'proxy_svc.exe'
        expected_result = AuthproxyUsingPortSkippedTestResult('listen tcp')

        actual_result = listen._reason_port_is_used(1234, '', None, 'tcp')

        self.assertEqual(expected_result, actual_result)

    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_pid')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_name')
    def test_other_proc_is_using_port(self, mock_name, mock_pid):
        mock_pid.return_value = 1
        mock_name.return_value = 'other_proc'

        result = listen._reason_port_is_used(1234, '', None, 'tcp')

        self.assertFalse(result.is_successful())
        self.assertEqual(result.port_user, 'other_proc')
        self.assertEqual(result.pid, 1)

    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_pid')
    @mock.patch('duoauthproxy.lib.connectivity.connect.listen._get_listener_name')
    def test_cant_determine_port_user(self, mock_name, mock_pid):
        mock_pid.return_value = None
        mock_name.return_value = 'other_proc'

        result = listen._reason_port_is_used(1234, '', None, 'tcp')

        self.assertFalse(result.is_successful())
        self.assertIsNone(result.port_user)


if __name__ == '__main__':
    import unittest
    unittest.main()
