#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

import unittest

import mock
from OpenSSL import crypto

from duoauthproxy.lib.connectivity.connect import ssl_operations
from ...mock_ldap_server import ca_cert_pem as EXTRA_CERT_PEM

GOOD_HOST_CERT_PEM = """-----BEGIN CERTIFICATE-----
MIICnzCCAYcCAhABMA0GCSqGSIb3DQEBCwUAMBIxEDAOBgNVBAMMB0ludGVybS4w
HhcNMTgwMjE5MjAzOTUzWhcNMjgwMjE3MjAzOTUzWjAYMRYwFAYDVQQDDA0xLmV4
YW1wbGUuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwZ7D/8aI
QCOQoB5HFaH3dBJrRN2u7gKZITUKlY0ziC1KFzoFo34ePAf7uSW4ESFOgWzsvfsY
E4cCK1d8tDBnEeZEb5uYx8boJKYbSp4zSsz1hmtTjVV1eWfPYb34q34XukuUf7nR
FpXmGEPpXo+4DpwtiErqDm+kXwai5Ukyu7F0JDJsI+a76B+uuTA6l+Bu2hUpcijG
f3m8GMVJButdcPPpCJIERZkv0isSUAlkoeqQL5dNqRKTY0Ic5b0G9EjTK5BHXRJ4
ya7RT5DoJ3rCsokW9iyou4iYC70alayz+gpGHqINtK6j6SrEc6lKS9s7qne6TG3f
pwY8q7ECIMwX5QIDAQABMA0GCSqGSIb3DQEBCwUAA4IBAQBABDzHeyLC3+cV7eyN
PAhZug11uBM5ASDv/jMz3m6l2Fu3jdSoV61mJ7QFDhL7iR0J1phHKTs+2veCHo5X
O3NlO0zRos5My6T1wqZF/qNWfPVyDURmKNa2CiVw3qxwUnNaWsPTrYOboHgFz2R2
27lWCZgwx9wm7DHIX/L4xmyvbe34GGvDg8EPjyZ1nk5xXwwZZE+GMr3CX3Eakyom
f63sHiN9F4NeU1nHMiAcasOEgMygZ0bApWvjFPbDxcsLuTplUNC5wLHnVmUWgpRC
TwCG0T5fDfckWm4pSg8wC5QOYXNfRFZuKE9cp8MopWnR+VrbX20QwUuwie4yIxx0
EYVY
-----END CERTIFICATE-----
"""  # expires Feb 17 20:39:53 UTC 2028

GOOD_HOST_KEY_PEM = """-----BEGIN PRIVATE KEY-----
MIIEwAIBADANBgkqhkiG9w0BAQEFAASCBKowggSmAgEAAoIBAQDBnsP/xohAI5Cg
HkcVofd0EmtE3a7uApkhNQqVjTOILUoXOgWjfh48B/u5JbgRIU6BbOy9+xgThwIr
V3y0MGcR5kRvm5jHxugkphtKnjNKzPWGa1ONVXV5Z89hvfirfhe6S5R/udEWleYY
Q+lej7gOnC2ISuoOb6RfBqLlSTK7sXQkMmwj5rvoH665MDqX4G7aFSlyKMZ/ebwY
xUkG611w8+kIkgRFmS/SKxJQCWSh6pAvl02pEpNjQhzlvQb0SNMrkEddEnjJrtFP
kOgnesKyiRb2LKi7iJgLvRqVrLP6CkYeog20rqPpKsRzqUpL2zuqd7pMbd+nBjyr
sQIgzBflAgMBAAECggEBAKtu0qXmd/8qPkS3akMqLHqWUGoiSl+s5z6ywxNVeXoq
d36HPjTbVxPaEbSGRCrOrb8E3bz81P19PZt4lA7LCIedtZyITUcdw6+nb2SrHEkv
nnkNEGaEc0YydzlYxhO5ynKFtw88THhN8vQrNTcn01UETYBurDxp90S4EIUpQYVS
9XAp54y12HLgwX8vxG0inPNto39MouEJkMh1+DGRmT4euGJu09qhGQ+q0CQk/CcP
Wd7GclFyu4doUSKaODQGp5UyVJRJO53wsx0XL9nnxgRrxcAM9zul1FGtz2rQCAVt
1nXqB6Gf38QdQKrgIU1vDvz5DzALnCVRstvGZZjQJKECgYEA/BmCYMt5fN+s3vPM
cnHYSakP9cBGxRH7wYMMF4eKcnM/jX+iAJPy0OYVvJNYR9yS1r2XZT33dRXSWFNT
vTGhf7npdBob78u9Zk7PJm0dHjkaKF4OjwrmlagWqWS3A8Fc3Y5Ij+Pde+9Hp0Jw
HFprMkwrWoKov3bpCHh3dRRtzYkCgYEAxJ2jBIkPXRvqpWIg/oTuvNlozlyknY1B
Mdd+pYzn6L8nsQHcxgiQ1qUFBEiHup0nZyXuWCaK+UUM8UflO33hMAk2gx19lDJv
2pFk+CRYo9DuqhJlBI830sVnqIQguuMPMxX6n5kQSnQz2q9LSxSFyf06BCbGWNE9
hOSkS7fl3H0CgYEAis2wz4K43Vx6Xjd19f1gzQ3cz76SOKdyKlXxLnNecy2pbEs9
83uCUiqdBTtLfCWiY8H3vbKksT7WlfMrRBwzg5PlShumDh3rpoz12X5ybdWa5WcE
NRYXZ4cpExN1AQhSDVrt8OvRYw+YMCuQdswCAauWozsVlnXuUM+jVj9tRaECgYEA
tSylGPrLQ3sD/52q1Z3c9/xGjVHlBqExICJ4u0haMqhihMM4kus1u/SX6NiE6bsC
ohPbyzLJAOVWBcfAPc/C/6adqrg0DalkVL2Bkv67B0jxg5orp8XpJLluyWPnXQ75
5JMuPH6o+lKVrb+129tvZS+IlGa+gv1ceObPVk6A62kCgYEA21ZQfUBC7HpDW9IC
C5YzUvGW6NxS+YbfitYpE0Vr6KZN7eHCHhc3M7FTV3ft4bSkpurXgchGaKG4DONn
rHrI+tlpiAgYbs87T3rXmxIToyuKSFyPtTPOxoI6c/rTj73w4UfYaDVCQ2WJIcTG
X/ZB/K+btDHQsAbeFGNBkpi2br4=
-----END PRIVATE KEY-----
"""

GOOD_INTERMEDIATE_CERT_PEM = """-----BEGIN CERTIFICATE-----
MIICsDCCAZigAwIBAgICEAAwDQYJKoZIhvcNAQELBQAwEjEQMA4GA1UEAwwHUm9v
dC1jYTAeFw0xODAyMTkyMDM5NTNaFw0yODAyMTcyMDM5NTNaMBIxEDAOBgNVBAMM
B0ludGVybS4wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC/i8wNvhN7
64kU7Y9utIB8KLiThJSNCy1Pg4OUQJYpCkai6gyApWUpAoZsfE1ySFn2F82ZKcPl
EKf8fO77SyoUPH8aAqt4IXsaaZ53UyHvgY3V297XA4UoFaT9GOsCE01OAbAYtIGw
Kz+GrrCWQowsHQz/bebP3JLcN32lhiWjQ2xLYOlSSHD5fx/WkjyuSW/nSxL3tz+b
a/LlZSLG4zbuDU8ojoWiS8P1shphyPZvoCihWaziq3h6yq3jdRilgA/mjPsW389o
HST9NMH3nImbboL7VpbzhxegXhzQx0Bqg4sVg5CQnSLXfpCj0L4kAyw88oVbJe2w
GgW9rJnm2ZkpAgMBAAGjEDAOMAwGA1UdEwQFMAMBAf8wDQYJKoZIhvcNAQELBQAD
ggEBAH6NBtNYF85fTSUCSWxfT7rSS4Z+95GgX+QDnjAPC533XUC4/cmlj1TmNVg3
0jfO+N4KfkiULIZxCBC6tJbZhodci00/30T6Mh3wCT/wXbRl3cbM58b0dG8SusPG
ktpjjfs3ltMv8+5C4Qriz88vysBeIrLgEvsvGqN/Vt6X0VkhYihff3E/F6vX9me1
gEwJqEwjKC/gCC3KqqXi39KfU6+rX+UJC1Y8uDgKcHdrXivflUNb5mszO/+aXFii
S9sTkDVeZAEw9p9pjkdVmFEkiMpr+UxR56wZo9GbPJY41qznia9fLx0GQ19gvGpi
YFWDb2Z4fysIpMAV9oHBv/aiEqw=
-----END CERTIFICATE-----
"""  # expires Feb 17 20:39:53 UTC 2028

GOOD_INTERMEDIATE_KEY_PEM = """-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAv4vMDb4Te+uJFO2PbrSAfCi4k4SUjQstT4ODlECWKQpGouoM
gKVlKQKGbHxNckhZ9hfNmSnD5RCn/Hzu+0sqFDx/GgKreCF7Gmmed1Mh74GN1dve
1wOFKBWk/RjrAhNNTgGwGLSBsCs/hq6wlkKMLB0M/23mz9yS3Dd9pYYlo0NsS2Dp
Ukhw+X8f1pI8rklv50sS97c/m2vy5WUixuM27g1PKI6FokvD9bIaYcj2b6AooVms
4qt4esqt43UYpYAP5oz7Ft/PaB0k/TTB95yJm26C+1aW84cXoF4c0MdAaoOLFYOQ
kJ0i136Qo9C+JAMsPPKFWyXtsBoFvayZ5tmZKQIDAQABAoIBAEvz4Bb27vztWa8N
wBlSw7v33bDJGqqO6wmjiZJIxgJCC95f1lNQwjsvwCPWalNNszZtG96G/CGhqCrr
U0wb8vSRRIkI0/tPs2O2y9pfAY7aHWbWx5cU00EgzAp8TfvFsZqj+rLhKKe1jnwW
WvlKLBD1+Ah1CNgVi7p82nq/Jb+FtcGSroHTOyxBncMhQcjGpMRWa5TJddpbnkkD
xkQ3Pw8b9WMFdW4ER3ju4kPLJI0fFmKTWPf6kyzyf0pvY3bzbcZdKZ8tily4GDZq
AzmDaqPtN+9jpnXWdy3OO/w9BbFfQAg8VjV0Aoy2yitGn5geMIKhPszVb1SUhgr4
MoVi1LECgYEA64Ijew43Cof+G3y+In2I3E7uS9hkeME48nkFdbOcdVbxNz9Ydrj8
H0+QuxkLxNpYJ7bxJxuUdWN5hiAaOn1JOnOkEAHFMYIR4zDQgNlRMhwIgWhxD7/r
RQQSPnvk4JihXhiWUjcJDq8Fzeo/5ZU7lTzXuDvOgUNybFDHunRXUL0CgYEA0DZq
SCqxtPEbaav3dWvEDvhjjmDLD97ODSKFBMycC01p5HB1SbKrh3pRsJaZE4PZMZLO
JpprdDm/8C8ZcBt6UP5IlQ84pbzNh+TH2jLtDjtWt/htB/1jmjhkeoGE+hV46bhU
fQPDdmBGVYjsWBKZXMW/0oKbx09ymq5zWqEZ3t0CgYEA3pjK4Gy3KYUQcPqe3Aq7
aUBRi+6kSF6ZbNVmi7GQmHfs1sNQ4iZGZ3HU4oLx1MZrvF314Xj4AVxcaKG/KLly
lYTn9Rk3QWwnbA2fhaniTPKiN254LuK/hSQr+lMwKfkZzM6AhoTuUqqrWc2nR2Ce
SJiRbNlafmM5q5dOsLV2MiECgYEAufuhPA1G50VnIyEEzdIRMv4Uj51ORULT80+e
Wgcaa84ehHsfiVEbfsx4I7Ch05i2bV2g30KYMR1UoYqwwj8JJ0e8Ncl+5CyduoFV
M0bCETBX6cTyP2zLh3pP5k3DJk//zWVlsrWl5xtgOA8ybglPDFtnWPZph8Mjr5+c
UpwMq9ECgYBzjoUM8mOIeFQ9Xqnk7nI9gseAJFvlGzrCfgUaJpPj2WgDWgC6hKG3
EM/62EqEat2BJvpHTu0kEDCFpazUGo2IdqJVtShsJrvTXqRpvN2Ww9NyLj4PUeXb
wJaNeDcvx7pFpnQRm24sZx11phnBKJCfP/y4+0YAWnxmsI45ZIapWQ==
-----END RSA PRIVATE KEY-----
"""

GOOD_ROOT_CERT_PEM = """-----BEGIN CERTIFICATE-----
MIICtzCCAZ+gAwIBAgIJAMu27BXA15fMMA0GCSqGSIb3DQEBCwUAMBIxEDAOBgNV
BAMMB1Jvb3QtY2EwHhcNMTgwMjE5MjAzOTUzWhcNMjgwMjE3MjAzOTUzWjASMRAw
DgYDVQQDDAdSb290LWNhMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
xXc+e++NPim7JZNuY4uADWvK8JdcDDkDsZI7ZM1d0+YunU1cxEV1iPEZkf7aId2+
Wl78FBeu2hn9AzN0cweFtWcVs5FuXJ8+OEYKk3GqZOtslZfKzzNEI8FAN1YDeNrC
nlLu7fLlC/b1VXeR9QVd7pl7h+SDvzJMvKluCN60JWZk5jIs31/fjKYm7kC+qTut
roclt323bGpzZ3eQcRRxBoFX6c7KrSNXk4PfZc94GsMnMf1ooBRxQrT4mkGhuOaB
d1+jD4fOO9YMwJaFjrOEknSOvH/hb43ck3b021wW9YACLOBY7cSynobb5ENupFTQ
0RARg81FUsLof3gMQ3En/wIDAQABoxAwDjAMBgNVHRMEBTADAQH/MA0GCSqGSIb3
DQEBCwUAA4IBAQBr8aMfazkQ/XyC6HLnNk/VF+GPJqi2pZan0S1nf4wupjXrftoM
HauOENotz9uDgrMUte7mZ2poHNMYrSETGBHvMMgIQMrQcxwySQym4icocjzm8ONw
Sjw0Sw0R2jw/QebJK8Z1VE5St/y08FCnOWXyaL0rMn9HRkA46rBUrd6cFJYYeW4y
Qf/UjiRMYONUGsOmgg1QDUyK+QHs5/ahBomxwctafpi1oGFDKgMvc4Bz7Ek/gxJI
9AU2aJakOB0KP2pY/euNSxWKAHVrCTjcLIuhK0qa0xpk4U09sy60hqFzn//4LKX2
DDwf2/sCd3bPA2z1SDIt1Z+T7T2p5tmvE5nr
-----END CERTIFICATE-----
"""  # expires Feb 17 20:39:53 UTC 2028

EXPIRED_CERT_PEM = """-----BEGIN CERTIFICATE-----
MIIDWjCCAkKgAwIBAgIJAOtYLTcKfcwnMA0GCSqGSIb3DQEBCwUAMEIxCzAJBgNV
BAYTAlhYMRUwEwYDVQQHDAxEZWZhdWx0IENpdHkxHDAaBgNVBAoME0RlZmF1bHQg
Q29tcGFueSBMdGQwHhcNMTgwMTI2MTg0MjU0WhcNMTgwMTI5MTg0MjU0WjBCMQsw
CQYDVQQGEwJYWDEVMBMGA1UEBwwMRGVmYXVsdCBDaXR5MRwwGgYDVQQKDBNEZWZh
dWx0IENvbXBhbnkgTHRkMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
xXoHx9XhSqCWNMrwYPhfx5/01SGIfypFyw85FLfWGvY+P8PqhWxM6QCDIVoMmFsp
lJviyJzw6AlHUyyWoN+mKkVs8b9wYSr9nU/YatrmKeFFqm/41DKHeAdX+z27Ivi+
MdEZEC384Z5L8yp6qBoiW4xL8naVOldCxhM4x6vYs0SQPNmDN0mamEZyN6ouHTO9
rRqq/xEdOGXntTi/GnhXiDQ5Pp/keBACgGXwIqHES5lhuCZgaze7uSHp+44HFHkb
/2ACcXcJu+QNQ2Q5/MQEN4wzyrck/8Vyb3AC2YGlaxn8WQfA1YoQ97TD9cDWNB1i
U/hBJIBG+qcu6+LI3G8N/QIDAQABo1MwUTAdBgNVHQ4EFgQUZnUwKHqizP3FBUv5
KX65CKN6CeswHwYDVR0jBBgwFoAUZnUwKHqizP3FBUv5KX65CKN6CeswDwYDVR0T
AQH/BAUwAwEB/zANBgkqhkiG9w0BAQsFAAOCAQEAGpmtbH+1XkCplwtDPoz7kSPV
BZK1K/vhnmdJI7lNyT+PCGo/ze6lzzPdH93DkSjH5ryt0X3lxTh8WvHK0D8Au9Tx
Eb1BQzh2pkS9f44aQeLX/6iM1T9K9wWVf0ooQjnUJJRP+X90eiCJofm6BxdFojq2
V02UeaepIqLJVcj0KudF5jfz0PKUZLlkOhfIOI2W0oq0fSmEGwgKx1U7GIwiHlYJ
2cJnqf/1FUNNN7oAiZ0g0wPT8TfV80hkTUHFpxO3CuZf1vBHkhPIO25GPPsWIU1f
Qb+VF3eZ8U+RmZZYLUUmfy6TXoO5sjKXtQaeFG+PgAeRpT7w7iEyc1++tvWgAg==
-----END CERTIFICATE-----
"""

SELF_SIGNED_CERT_PEM = """-----BEGIN CERTIFICATE-----
MIIDWTCCAkGgAwIBAgIJALDvE1bKwMMHMA0GCSqGSIb3DQEBCwUAMEIxCzAJBgNV
BAYTAlhYMRUwEwYDVQQHDAxEZWZhdWx0IENpdHkxHDAaBgNVBAoME0RlZmF1bHQg
Q29tcGFueSBMdGQwIBcNMTgwMjIxMTYyMzA0WhgPMjExODAxMjgxNjIzMDRaMEIx
CzAJBgNVBAYTAlhYMRUwEwYDVQQHDAxEZWZhdWx0IENpdHkxHDAaBgNVBAoME0Rl
ZmF1bHQgQ29tcGFueSBMdGQwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
AQCls3st9uZLaguHxPiLKrdR0DwU7nP0FS73Loq6MdEc9WoRMdqP21WLgiUtmPzC
IScaiq8jFmu+soJdaZcbUu8qJwYfGw3b+UHLa/Oy21xRrfX38Jt0U1wVXvyQtsBe
fsXvQEKAbui6xQ43sJjrIgbA0mzUMSHYyBqJBmj8Bt2l+cvkrW9xj+NHA3OqUPmQ
AgeOflUBCYKcT6wfvxmprL+beyVVVybm9+71Bi+1B+NaPlVqs3XO9nfYqFlsj3QR
4sJDEgHEg/SYArKA2AdhSSEQNJ1cHc99J/cJk9qWZ6EWnhpchLp4mxEy93eIvWC2
HtUHpo0Mx7zNMVakz5B6SYJZAgMBAAGjUDBOMB0GA1UdDgQWBBSomE4XLQOu9LW+
26QWNxk1VGfpkTAfBgNVHSMEGDAWgBSomE4XLQOu9LW+26QWNxk1VGfpkTAMBgNV
HRMEBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBAQB4qcL/GPXqiC3UuX6+7fwzKfgj
T5UpLtw16b8IAaCV0XQVLSHlwKiidcz82SxLfgNgqP3vyQn2ShaVMQgMNzC/EeB2
Dy40ZMqUoJJXwDBcQ/yTwFkTYPqqkJcSZQR3IyGPXjC5XKBsKAXBghkD3dX5+IPn
uyaxzjWsd3uk29aSbEKf0ixyYnC2+fRGGh0BhtZWR4eD/9goFG76fLkz6bLHJ73M
gQ0GOcMqT5IUXFu1pp6G+LtOscVWqYDrNdgrhhlS+vpD74p1jkL3NoAjkd7QSu4u
7J9/088tURAXhAqGLyKjp34MDaGJ9dzn4k7T4noI7uFWPbOUD82oyI3QDkHv
-----END CERTIFICATE-----
"""  # expires Jan 28 16:23:04 UTC 2118

SELF_SIGNED_KEY_PEM = """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCls3st9uZLaguH
xPiLKrdR0DwU7nP0FS73Loq6MdEc9WoRMdqP21WLgiUtmPzCIScaiq8jFmu+soJd
aZcbUu8qJwYfGw3b+UHLa/Oy21xRrfX38Jt0U1wVXvyQtsBefsXvQEKAbui6xQ43
sJjrIgbA0mzUMSHYyBqJBmj8Bt2l+cvkrW9xj+NHA3OqUPmQAgeOflUBCYKcT6wf
vxmprL+beyVVVybm9+71Bi+1B+NaPlVqs3XO9nfYqFlsj3QR4sJDEgHEg/SYArKA
2AdhSSEQNJ1cHc99J/cJk9qWZ6EWnhpchLp4mxEy93eIvWC2HtUHpo0Mx7zNMVak
z5B6SYJZAgMBAAECggEBAJTPIwiceyWvDrvXgP46Vnf68no5y5igjZUfDyc4JfxD
ygW3J/mZmLG5J3pTsmsFKwEDG1X/hLO1zK3nbfTz/iV09mJpAYFT0GrHgAxYRAh2
mjcYtgM5+bCXpWqxNQGoCEpdA1wkPo28+o+44RwJatcwItxNY4Az9OCqCQ6Ne/DZ
et+WtlCIrcCb0aOWCQdT7WURXpC4gr1de3tqM78reQE13ReyW5KtvvsmMIEG7cr6
S88LubvEMIwx++vqJETBPQr8H0/pGIfb3zZVWA+gi3rT/79ZfDDjpBLtb512+kHR
sAwOXrTC320wKMbQxfqiud3pKituz6+ZufhGz8H8QqECgYEA0w9OxcEhMsZOcZPZ
J5SeKpD1pdpOasDBclToSr1GgQ1iVSJKqsrkRv2CIN0OuikZcgZLOyPGZT+lp7qp
dorgFt5cR0F7k1EWdbDI0mZCMAI/z7i2I0J/DaKitgi2itn9EHY8aMdWkaEM2UsB
j6V7Uv/bddb6T64GulDaLWTP1eMCgYEAyPu1fW5SppR/5gJkR3eppSMerWcBKKx5
e7GNSUw7r+cTxm6fjn7KH489/wO3HymF8RioE/EsVKPs+vxxCZtOzAyZbcRy7QRN
dr1jdPwsOvK/WB1ulPl8dGMA3CD1OF54HAnaCTteghM5IOhV6SoSdI+yR7tQP6Y0
51dUl0aOW5MCgYB5G+GRrD7GifBdT/9drPQNbNzobJpuu2jFmF8q8+1PBEUWoPTD
iGTxL2rKPM7El+zX9zCYIPEpDBqcYon3Ub1AohvFb5ffcro2kljKEo+MSN4g7dbl
6LRLOLWxStKkXm3OUmpbnXbZXdanoVil+skHWXWu6sr7idqmctqfDagNDQKBgCqR
rcIwcM8ZmvFjxkug28y/YIsHyyYLSfvaAnvo6qd6QSUBX0g2MKd4loACEHJIQDdt
qoP2zBXtVo9ekq6fN5RegszOFC2BgMrKmFs54kBmFAMc0MQ7B5I7Of5l+KqEz9nY
eFcGPSE94xbGPxQQtjH8EZ3B3eBtWr05MRCNgC+DAoGAf4nG4yoINgPXXqCM0frc
KgCwZYIYgtpg0TfSeglexjjzBXq/NqPVDPYNildkzBjl74oX39op1gK8Y79xdIkE
qkjMaL6+byvfF96CgtFF+Vn0IEhOEcGHWLyptsx38oHbhQCbHbEhieSckWUdmLww
QKAm0EI8egF6hBICAgk9iY8=
-----END PRIVATE KEY-----
"""

GOOD_CHAINED_CERT_PEM = GOOD_HOST_CERT_PEM + "\n" + GOOD_INTERMEDIATE_CERT_PEM + "\n" + GOOD_ROOT_CERT_PEM
DUPLICATE_CONTAINING_CHAIN_CERT_PEM = GOOD_HOST_CERT_PEM + "\n" + GOOD_HOST_CERT_PEM + "\n" + GOOD_INTERMEDIATE_CERT_PEM + "\n" + GOOD_ROOT_CERT_PEM
OUT_OF_ORDER_CHAIN_CERT_PEM = GOOD_INTERMEDIATE_CERT_PEM + "\n" + GOOD_ROOT_CERT_PEM + "\n" + GOOD_HOST_CERT_PEM
MISSING_INTERMEDIATE_CHAIN_PEM = GOOD_HOST_CERT_PEM + "\n" + GOOD_ROOT_CERT_PEM
UNTRUSTED_CERT_CONTAINING_CHAIN_CERT_PEM = GOOD_HOST_CERT_PEM + "\n" + EXTRA_CERT_PEM + "\n" + GOOD_ROOT_CERT_PEM
TRUNCATED_CERT_PEM = GOOD_CHAINED_CERT_PEM[:len(GOOD_CHAINED_CERT_PEM)/2]

good_key = crypto.load_privatekey(crypto.FILETYPE_PEM, SELF_SIGNED_KEY_PEM)
good_cert = [crypto.load_certificate(crypto.FILETYPE_PEM, SELF_SIGNED_CERT_PEM)]


def _mock_open(filename, **kwargs):
    """To prevent the creation of temporary files (and to make up for mock.mock_open's
    inability to specify multiple read_datas), we can mock out open()'s output instead."""
    if filename == "/path/to/ssl/key":
        content = SELF_SIGNED_KEY_PEM
    elif filename == "/path/to/ssl/cert":
        content = SELF_SIGNED_CERT_PEM
    elif filename == "/path/to/expired/cert":
        content = EXPIRED_CERT_PEM
    elif filename == "/path/to/intermediate/key":
        content = GOOD_INTERMEDIATE_KEY_PEM
    elif filename == "/path/to/intermediate/cert":
        content = GOOD_INTERMEDIATE_CERT_PEM
    elif filename == "/path/to/chained/cert":
        content = GOOD_CHAINED_CERT_PEM
    elif filename == "/path/to/chained/key":
        content = GOOD_HOST_KEY_PEM  # corresponds with first cert in GOOD_CHAINED_CERT_PEM
    elif filename == "/path/to/out/of/order/chain":
        content = OUT_OF_ORDER_CHAIN_CERT_PEM
    elif filename == "/path/to/missing/intermediate":
        content = MISSING_INTERMEDIATE_CHAIN_PEM
    elif filename == "/path/to/duplicate/containing/cert":
        content = DUPLICATE_CONTAINING_CHAIN_CERT_PEM
    elif filename == "/path/to/truncated/cert":
        content = TRUNCATED_CERT_PEM
    elif filename == "/path/to/untrusted/containing/cert":
        content = UNTRUSTED_CERT_CONTAINING_CHAIN_CERT_PEM
    elif filename == "/path/to/bad/key" or filename == "/path/to/bad/cert":
        content = ""
    else:
        raise IOError("Non-existent or unreadable file")
    file_object = mock.mock_open(read_data=content).return_value
    file_object.__iter__.return_value = content.splitlines(True)
    return file_object


class TestCanValidateSSLCreds(unittest.TestCase):
    def test_load_ssl_key_failure(self):
        """A bad path or unreadable private key file should return None"""
        file_result, key_data = ssl_operations._load_ssl_key("/path/to/nonexistent/key")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(key_data)

    def test_load_ssl_certs_failure(self):
        """A bad path or unreadable cert file should return None"""
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/nonexistent/cert")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    def test_contains_truncated_certs_true(self):
        input = TRUNCATED_CERT_PEM
        self.assertTrue(ssl_operations._contains_truncated_certs(input))

    def test_contains_truncated_certs_false(self):
        input = GOOD_CHAINED_CERT_PEM
        self.assertFalse(ssl_operations._contains_truncated_certs(input))

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_single_expired(self):
        """A cert past its NotAfter date should not validate
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/expired/cert")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_contains_truncated(self):
        """A multi-cert containing a truncated cert should not validate
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/truncated/cert")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_intermediate_only(self):
        """An intermediate cert on its own without its root cert should not verify
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/intermediate/cert")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_missing_intermediate(self):
        """An intermediate cert that signs off on a host's cert is needed to fully
        validate the root -> intermediate -> host chain
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/missing/intermediate")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_chain_extra_untrusted_cert_present(self):
        """A cert chain containing a random extra cert shouldn't validate
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/untrusted/containing/cert")
        self.assertFalse(file_result.is_successful())
        self.assertIsNone(cert_data)

    @mock.patch('__builtin__.open', _mock_open)
    def test_load_ssl_certs_duplicate_cert_present(self):
        """A duplicate cert present in an otherwise valid chain should be de-duped
        and validate successfully
        """
        file_result, cert_data = ssl_operations.load_ssl_certs("/path/to/duplicate/containing/cert")
        self.assertTrue(file_result.is_successful())
        self.assertIsNotNone(cert_data)

    def test_can_create_ssl_context_bad_ciphers(self):
        """Passing in a badly-formatted cipher-list string should return None"""
        res = ssl_operations._can_create_ssl_context(
            good_key,
            good_cert,
            "BAD:CIPHER:LIST"
        )
        self.assertFalse(res.is_successful())

    def test_can_create_ssl_context_bad_key(self):
        """Attempting to validate without a key should return False"""
        res = ssl_operations._can_create_ssl_context(
            None,
            good_cert,
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    def test_can_create_ssl_context_bad_cert(self):
        """Attempting to validate without a cert should return False"""
        res = ssl_operations._can_create_ssl_context(
            good_key,
            None,
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    def test_can_create_ssl_context_no_creds(self):
        """Attempting to validate with neither a cert nor key should return False"""
        res = ssl_operations._can_create_ssl_context(
            None,
            None,
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_success(self):
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/ssl/key",
            "/path/to/ssl/cert",
            "DEFAULT"
        )
        self.assertTrue(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_nonexistent_key(self):
        """A private key file that can't be found or has bad permissions should cause an IOError"""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/nonexistent/key",
            "/path/to/ssl/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_nonexistent_cert(self):
        """A certificate file that can't be found or has bad permissions should cause an IOError"""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/ssl/key",
            "/path/to/nonexistent/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_bad_privatekey(self):
        """An existing but undecodable private key file should not validate"""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/bad/key",
            "/path/to/ssl/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_bad_certificate(self):
        """An existing but undecodable cert file should not validate"""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/ssl/key",
            "/path/to/bad/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_expired_cert(self):
        """An existing but expired certificate should not validate"""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/ssl/key",
            "/path/to/expired/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_mismatched_key_cert(self):
        """A private key that doesn't match its certificate should not validate.
        In this case, the intermediate-only key is given instead of the self-signed
        key."""
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/intermediate/key",
            "/path/to/ssl/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_intermediate_cert_only(self):
        """Attempting to validate an intermediate cert without its root/CA
        cert present should fail
        """
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/intermediate/key",
            "/path/to/intermediate/cert",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_chained_cert(self):
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/chained/key",
            "/path/to/chained/cert",
            "DEFAULT"
        )
        self.assertTrue(res.is_successful())

    @mock.patch('__builtin__.open', _mock_open)
    def test_can_validate_ssl_creds_chain_out_of_order(self):
        """Certs are expected to be in host, then intermediate(s), then root order in-file;
        it should not validate otherwise
        """
        res = ssl_operations.can_validate_ssl_creds(
            "/path/to/chained/key",
            "/path/to/out/of/order/chain",
            "DEFAULT"
        )
        self.assertFalse(res.is_successful())


if __name__ == "__main__":
    unittest.main()
