#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
from OpenSSL import crypto

from twisted.internet import defer
from twisted.internet import error
from twisted.internet import protocol
from twisted.internet import reactor
from twisted.python import components
from twisted.python.failure import Failure

import ldaptor.protocols.pureldap
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.ldap import ldapserver
import ldaptor.interfaces
import ldaptor.inmemory

from base import sleep

from duoauthproxy.modules import ssl_server


class MoreADLikeLDAPServer(ldapserver.LDAPServer,
                           object):
    @defer.inlineCallbacks
    def handle(self, *args, **kwargs):
        if self.factory.response_delay:
            yield sleep(self.factory.response_delay)
            self.factory.delayed.callback(None)
            self.factory.delayed = defer.Deferred()
        try:
            r = super(MoreADLikeLDAPServer, self).handle(*args,
                                                         **kwargs)
            defer.returnValue(r)
        except ldapserver.LDAPServerConnectionLostException:
            pass

    def _cbHandle(self, *args, **kwargs):
        try:
            return super(MoreADLikeLDAPServer,
                         self)._cbHandle(*args, **kwargs)
        except ldapserver.LDAPServerConnectionLostException:
            pass

    # Override generation of search results to inject LDAPSearchResultRef:
    def _cbSearchGotBase(self, base, dn, request, reply):
        def _sendEntryToClient(entry):
            reply(ldaptor.protocols.pureldap.LDAPSearchResultEntry(
                objectName=str(entry.dn),
                attributes=entry.items(),
            ))
        d = base.search(filterObject=request.filter,
                        attributes=request.attributes,
                        scope=request.scope,
                        derefAliases=request.derefAliases,
                        sizeLimit=request.sizeLimit,
                        timeLimit=request.timeLimit,
                        typesOnly=request.typesOnly,
                        callback=_sendEntryToClient)

        def _done(_):
            # Ensure proxy does not crash when round-tripping
            # LDAPSearchResultRef because AD sends them but upsteam
            # ldaptop neither generates nor parses them.
            reply(ldaptor.protocols.pureldap.LDAPSearchResultReference(''))

            return ldaptor.protocols.pureldap.LDAPSearchResultDone(
                resultCode=ldaperrors.Success.resultCode,
            )
        d.addCallback(_done)
        return d

    def handle_LDAPSearchRequest(self, request, controls, reply):
        callee = super(MoreADLikeLDAPServer, self).handle_LDAPSearchRequest
        if self.factory.handle_LDAPSearchRequest is not None:
            callee = self.factory.handle_LDAPSearchRequest
        return callee(request, controls, reply)

    def handle_LDAPBindRequest(self, request, controls, reply):
        self.factory.bind_requests_received += 1
        if self.factory.handle_LDAPBindRequest:
            return self.factory.handle_LDAPBindRequest(
                self, request, controls, reply)
        return super(MoreADLikeLDAPServer, self).handle_LDAPBindRequest(
            request, controls, reply)


class LDAPServerFactory(protocol.ServerFactory):
    protocol = MoreADLikeLDAPServer

    def __init__(self, root):
        self.root = root
        self.response_delay = None
        self.delayed = defer.Deferred()
        self.bind_requests_received = 0
        self.handle_LDAPBindRequest = None
        self.handle_LDAPSearchRequest = None


components.registerAdapter(lambda x: x.root,
                           LDAPServerFactory,
                           ldaptor.interfaces.IConnectedLDAPEntry)

MEGA_GROUP_SIZE = 3003
MEGA_GROUP_MEMBERS = [
    "CN=test_user.{0},CN=Users,DC=bongos,DC=test".format(i + 1)
    for i in range(0, MEGA_GROUP_SIZE)]

LDIF_DATA = """\
# extended LDIF
#
# LDAPv3
# base <DC=bongos,DC=test> with scope subtree
# filter: (cn=*)
# requesting: ALL
#
# bongos.test
dn: DC=bongos,DC=test
objectClass: top
objectClass: container
cn: Users
distinguishedName: DC=bongos,DC=test
instanceType: 4
name: Users
objectGUID:: gmiU4dEoqkS1CzrgfsyhuA==
systemFlags: -1946157056
objectCategory: CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test
isCriticalSystemObject: TRUE

# subdomain.bongos.test
dn: DC=subdomain,DC=bongos,DC=test
objectClass: top
objectClass: container
cn: Users
distinguishedName: DC=subdomain,DC=bongos,DC=test
instanceType: 4
name: Users
objectGUID:: gmiU4dEoqkS1CzrgfsyhuX==
systemFlags: -1946157056
objectCategory: CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test
isCriticalSystemObject: TRUE

# Users, bongos.test
dn: CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: container
cn: Users
distinguishedName: CN=Users,DC=bongos,DC=test
instanceType: 4
name: Users
objectGUID:: gmiU4dEoqkS1CzrgfsyhuA==
systemFlags: -1946157056
objectCategory: CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test
isCriticalSystemObject: TRUE

# Users, subdomain.bongos.test
dn: CN=Users,DC=subdomain,DC=bongos,DC=test
objectClass: top
objectClass: container
cn: Users
distinguishedName: CN=Users,DC=subdomain,DC=bongos,DC=test
instanceType: 4
name: Users
objectGUID:: gmiU4dEoqkS1CzrgfsyhuX==
systemFlags: -1946157056
objectCategory: CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test
isCriticalSystemObject: TRUE

# Fred Person, Users, bongos.test
dn: CN=Fred Person,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Fred Person
sn: Person
givenName: Fred
distinguishedName: CN=Fred Person,CN=Users,DC=bongos,DC=test
instanceType: 4
displayName: Fred Person
name: Fred Person
objectGUID:: ic4Cx19ZZ0ytdFqEFgu+Lw==
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
objectSid:: AQUAAAAAAAUVAAAAsrQ07CvakI19EZdAVAQAAA==
sAMAccountName: fred
userPrincipalName: fred@bongos.test
objectCategory: CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test
objectCategory: person
userPassword: {SSHA}xMrqdPlDSqoNLo/troPdSUlWg/q2HoacYPjU0A==
msDS-PrincipalName: bongos\\fred

# Jane Blackwood, Users, bongos.test
dn: CN=Jane Blackwood,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Jane Blackwood
sn: Blackwood
givenName: Jane
distinguishedName: CN=Jane Blackwood,CN=Users,DC=bongos,DC=test
instanceType: 4
displayName: Jane Blackwood
memberOf: CN=Users,DC=bongos,DC=test
name: Jane Blackwood
objectGUID:: kbC3e0RzokWZ9xpbFyFfiQ==
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
objectSid:: AQUAAAAAAAUVAAAAsrQ07CvakI19EZdAVQQAAA==
sAMAccountName: jane
userPrincipalName: jane@bongos.test
objectCategory: user
objectCategory: person
userPassword: {SSHA}fd13FaJjxVpjZMDhJuhwmhsTj9FtbW0gc2FsdHk=
msDS-PrincipalName: bongos\\jane

# Dupe User 1, Users, subdomain.bongos.test
dn: CN=Dupe User 1,CN=Users,DC=subdomain,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Dupe User 1
sn: User
givenName: Dupe
distinguishedName: CN=Dupe User 1,CN=Users,DC=subdomain,DC=bongos,DC=test
instanceType: 4
displayName: Dupe User 1
name: Dupe User 1
objectGUID:: kbC3e0RzokWZ9xpbFyFfiX==
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
objectSid:: AQUAAAAAAAUVAAAAsrQ07CvakI19EXdAVQQAAA==
sAMAccountName: dupe
userPrincipalName: dupe@subdomain.bongos.test
objectCategory: user
objectCategory: person
userPassword: {SSHA}p+aBX4U0x6n17O6kr05lMx+M+TQicCENtnyOIg==
msDS-PrincipalName: subdomain\\dupe

# Dupe User 2, Users, bongos.test
dn: CN=Dupe User 2,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Dupe User 1
sn: User
givenName: Dupe
distinguishedName: CN=Dupe User 2,CN=Users,DC=bongos,DC=test
instanceType: 4
displayName: Dupe User 2
name: Dupe User 2
objectGUID:: kbC3e0RzokWZ9xpbFyFfiZ==
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
objectSid:: AQUAAAAAAAUVAAAAsrQ07CvakI19EXdAFQQAAA==
sAMAccountName: dupe
userPrincipalName: dupe@bongos.test
objectCategory: user
objectCategory: person
userPassword: {SSHA}fd13FaJjxVpjZMDhJuhwmhsTj9FtbW0gc2FsdHk=
msDS-PrincipalName: bongos\\dupe

# adminer adminer, Users, bongos.test
dn: CN=adminer adminer,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: adminer adminer
sn: adminer
givenName: adminer
distinguishedName: CN=adminer adminer,CN=Users,DC=bongos,DC=test
instanceType: 4
displayName: adminer adminer
name: adminer adminer
objectGUID:: S3bP8XTr3Uu2rpAjw+1kcQ==
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
objectSid:: AQUAAAAAAAUVAAAAsrQ07CvakI19EZdAVwQAAA==
sAMAccountName: adminer
userPrincipalName: adminer@bongos.test
objectCategory: CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test
objectCategory: person
userPassword: {SSHA}0Ap2Wgcf2vjM4hGLdBgWbbgTqhNtbW0gc2FsdHk=
msDS-PrincipalName: bongos\\adminer

# Principal Skinner, Users, bongos.test
dn: CN=Principal Skinner,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Principal Skinner
sn: pskinner
givenName: Principal
distinguishedName: CN=Principal Skinner,CN=Users,DC=bongos,DC=test
instanceType: 4
displayName: Principal Skinner
name: Principal Skinner
userAccountControl: 66048
codePage: 0
countryCode: 0
primaryGroupID: 513
sAMAccountName: pskinner
userPrincipalName: pskinner@bongos.test
objectCategory: CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test
objectCategory: person
userPassword: {SSHA}fd13FaJjxVpjZMDhJuhwmhsTj9FtbW0gc2FsdHk=
msDS-PrincipalName: S-1-336566957-3618751888-1237994806-1297237956-58068105-1020576660

# Non AD, Users, bongos.test
dn: CN=Non AD,CN=Users,DC=bongos,DC=test
distinguishedname: CN=Non AD,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: user
cn: Non AD
sn: AD
givenName: Non
objectCategory: CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test
objectCategory: person
userPassword: {SSHA}xMrqdPlDSqoNLo/troPdSUlWg/q2HoacYPjU0A==

# OpenLDAPer, Users, bongos.test
dn: CN=OpenLDAPer,CN=Users,DC=bongos,DC=test
objectClass: top
objectClass: person
objectClass: organizationalPerson
cn: OpenLDAPer
sn: OpenLDAPer
userPassword: {SSHA}fd13FaJjxVpjZMDhJuhwmhsTj9FtbW0gc2FsdHk=

# mega users group (3K + 1 users - not even increment of max_range)
# also contains dn with characters that need to be escaped.
""" + r"""
dn: CN=Mega Users\,\+\"\\<\>\;,DC=bongos,DC=test
objectClass: top
objectClass: container
cn: Mega Users,+"\<>;
distinguishedName: CN=Mega Users\,\+\"\\<\>\;,DC=bongos,DC=test
instanceType: 4
name: Mega Users,+"\<>;
objectGUID:: mig4UdoqEk1CSzrfgshyAu==
systemFlags: -1946157056
objectCategory: CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test
isCriticalSystemObject: TRUE
""" + "\n".join(["member: " + m for m in MEGA_GROUP_MEMBERS]) + "\n\n"

ca_cert_pem = """\
-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----
"""

server_key_pem = """\
-----BEGIN RSA PRIVATE KEY-----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==
-----END RSA PRIVATE KEY-----
"""
server_key = crypto.load_privatekey(crypto.FILETYPE_PEM, server_key_pem)

server_cert_pem = """\
-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----
"""
server_cert = crypto.load_certificate(crypto.FILETYPE_PEM, server_cert_pem)


@defer.inlineCallbacks
def make_server_factory(port,
                        factory_class=LDAPServerFactory,
                        ssl=False,
                        method=None,
                        cipher_list=''):
    p = ldaptor.inmemory.InMemoryLDIFProtocol()
    p.dataReceived(LDIF_DATA)
    p.connectionLost(Failure(error.ConnectionDone()))
    db = yield p.completed
    factory = factory_class(db)
    if ssl:
        context_factory = ssl_server.ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            certificate_data=server_cert,
            method=method,
            cipher_list=cipher_list
        )
        listener = reactor.listenSSL(port, factory, context_factory)
    else:
        listener = reactor.listenTCP(port, factory)
    defer.returnValue((factory, listener))
