#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import mock
import socket

from twisted.logger import Logger
from ldaptor.protocols import pureldap
from OpenSSL import SSL

from base import TestCase, ConnectivityMockException

from duoauthproxy.lib.connectivity import connectivity_results
from duoauthproxy.lib.config_provider import ConfigError


class TestResultObjects(TestCase):

    def test_config_result_is_good(self):
        config_result = connectivity_results.ConfigCheckResult([])

        self.assertTrue(config_result.is_successful(), "Config result should be good")

    def test_config_result_is_bad(self):
        config_result = connectivity_results.ConfigCheckResult([connectivity_results.MissingConfigKeyProblem('key')])

        self.assertFalse(config_result.is_successful(), "Config result should not be good")


class ResultOutputTest(TestCase):
    def setUp(self):
        super(ResultOutputTest, self).setUp()

        self.mock_config_result = mock.Mock(spec=connectivity_results.ConfigCheckResult)
        self.mock_config_result.is_successful.return_value = True
        self.mock_ping_result = mock.Mock(spec=connectivity_results.DuoPingResult)
        self.mock_ping_result.is_successful.return_value = True
        self.mock_drift_result = mock.Mock(spec=connectivity_results.TimeDriftResult)
        self.mock_drift_result.is_successful.return_value = True
        self.mock_validate_result = mock.Mock(spec=connectivity_results.ValidateApiCredentialsResult)
        self.mock_validate_result.is_successful.return_value = True
        self.mock_listen_result = mock.Mock(spec=connectivity_results.ListenResult)
        self.mock_listen_result.is_successful.return_value = True
        self.mock_ssl_result = mock.Mock(spec=connectivity_results.SslResult)
        self.mock_ssl_result.is_successful.return_value = True
        self.mock_tcp_connect_result = mock.Mock(spec=connectivity_results.ConnectResult)
        self.mock_tcp_connect_result.is_successful.return_value = True
        self.mock_ldap_connect_result = mock.Mock(spec=connectivity_results.LdapConnectResult)
        self.mock_ldap_connect_result.is_successful.return_value = True
        self.mock_bind_result = mock.Mock(spec=connectivity_results.LdapBindResult)
        self.mock_bind_result.is_successful.return_value = True
        self.mock_search_result = mock.Mock(spec=connectivity_results.LdapSearchResult)
        self.mock_search_result.is_successful.return_value = True
        self.mock_logger = mock.Mock(spec=Logger)

    def _assert_info_only_called(self, *args, **kwargs):
        """
        Assert that only the info method on the mock logger was called, with the provided arguments

        Args:
            *args: expected args for the info call
            **kwargs: expected kwargs for the info call
        """
        self.mock_logger.info.assert_called_with(*args, **kwargs)
        self.mock_logger.warn.assert_not_called()
        self.mock_logger.error.assert_not_called()
        self.mock_logger.debug.assert_not_called()

    def _assert_not_logged(self, mocks):
        """
        Assert that none of the mocks' to_log_output methods were called

        Args:
            mocks [Mock]: the mocks to test
        """
        for one_mock in mocks:
            one_mock.to_log_output.assert_not_called()

    def _assert_logged(self, mocks):
        for one_mock in mocks:
            one_mock.to_log_output.assert_called()


class TestConfigCheckResultOutput(ResultOutputTest):
    def test_good_config_result_output(self):
        result = connectivity_results.ConfigCheckResult([])

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.ConfigCheckResult.SUCCESS_MESSAGE
        )

    def test_config_one_problem_result_output(self):
        key = 'missing_key'
        result = connectivity_results.ConfigCheckResult([
            connectivity_results.MissingConfigKeyProblem(key)
        ])

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.ConfigCheckResult.FAILURE_MESSAGE)
        self.mock_logger.error.assert_called_with(
            connectivity_results.MissingConfigKeyProblem.MESSAGE,
            config_key=key)

    def test_config_multiple_problems_result_output(self):
        key_1 = 'something'
        key_2 = 'something_else'
        key_3 = 'and_another_thing'
        value_3 = 'garbage'
        result = connectivity_results.ConfigCheckResult([
            connectivity_results.MissingConfigKeyProblem(key_1),
            connectivity_results.MissingConfigKeyProblem(key_2),
            connectivity_results.InvalidConfigKeyProblem(key_3, value_3)
        ])

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.ConfigCheckResult.FAILURE_MESSAGE)
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.MissingConfigKeyProblem.MESSAGE, config_key=key_1),
            mock.call(connectivity_results.MissingConfigKeyProblem.MESSAGE, config_key=key_2),
            mock.call(connectivity_results.InvalidConfigKeyProblem.MESSAGE, config_key=key_3, value=value_3)
        ])


class TestMainSectionResultOutput(ResultOutputTest):

    def test_good_main_section(self):
        good_cert = connectivity_results.SslCertFileResult(True, 'good_path')
        good_http_proxy = connectivity_results.HttpWebProxyResult(True, '1.2.3.4', 80)
        good_main = connectivity_results.MainSectionResult(good_cert, good_http_proxy)

        good_main.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(connectivity_results.MainSectionResult.SUCCESS_MESSAGE)

    def test_bad_main_section(self):
        path = 'bad_path'
        bad_cert = connectivity_results.SslCertFileResult(False, path, ConnectivityMockException())

        bad_host = 'bad_proxy'
        bad_port = 80
        bad_http_proxy = connectivity_results.HttpWebProxyResult(False, bad_host, bad_port, ConnectivityMockException())

        bad_main = connectivity_results.MainSectionResult(bad_cert, bad_http_proxy)

        bad_main.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls(
            [
                mock.call(connectivity_results.MainSectionResult.FAILURE_MESSAGE),
                mock.call(connectivity_results.SslCertFileResult.FAILURE_MESSAGE, file_path=path),
                mock.call(connectivity_results.HttpWebProxyResult.FAILURE_MESSAGE, http_proxy_host=bad_host, http_proxy_port=bad_port)
            ]
        )


class TestHttpProxyResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestHttpProxyResultOutput, self).setUp()

        self.mocks = [
            self.mock_config_result,
            self.mock_listen_result,
            self.mock_ping_result,
            self.mock_drift_result,
        ]

        self.http_proxy_result = connectivity_results.HttpProxyResult(*self.mocks)

    def test_good_http_proxy_success_output(self):
        self.http_proxy_result.to_log_output(self.mock_logger)

        self._assert_info_only_called(connectivity_results.HttpProxyResult.SUCCESS_MESSAGE)
        self._assert_not_logged(self.mocks)

    def test_partially_good_http_proxy_failure_output(self):
        self.mock_listen_result.is_successful.return_value = False
        self.mock_drift_result.is_successful.return_value = False

        self.http_proxy_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.HttpProxyResult.FAILURE_MESSAGE)
        self._assert_logged(self.mocks)

    def test_fully_bad_http_proxy_failure_output(self):
        for one_mock in self.mocks:
            one_mock.is_successful.return_value = False

        self.http_proxy_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.HttpProxyResult.FAILURE_MESSAGE)
        self._assert_logged(self.mocks)


class TestListenResultOutput(ResultOutputTest):

    def test_success_result_output_specified_interface(self):
        result = connectivity_results.ListenResult(True, 1812, '127.0.0.1')

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.ListenResult.SUCCESS_MESSAGE,
            port=1812,
            interface='127.0.0.1')

    def test_success_result_output_all_interfaces(self):
        result = connectivity_results.ListenResult(True, 1812, '')

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.ListenResult.SUCCESS_MESSAGE,
            port=1812,
            interface='all interfaces')

    def test_failure_result_object(self):
        result = connectivity_results.ListenResult(False, 1813, '127.0.0.1', port_user="netcat", pid=1234)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.ListenResult.FAILURE_MESSAGE,
            port=1813,
            interface='127.0.0.1',
            port_user="netcat",
            pid=1234,
            )


class TestConnectResultOutput(ResultOutputTest):
    def test_connect_success_output(self):
        host = 'host'
        port = 389
        result = connectivity_results.ConnectResult(True, host, port)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.ConnectResult.SUCCESS_MESSAGE,
            host=host,
            port=port)

    def test_connect_failure_output(self):
        host = 'host'
        port = 389
        exception = ConnectivityMockException()
        result = connectivity_results.ConnectResult(False, host, port, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.ConnectResult.FAILURE_MESSAGE,
            host=host,
            port=port)
        self.mock_logger.debug.assert_called_with(
            connectivity_results.ConnectResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))

    def test_ldap_connect_success_output(self):
        host = 'host'
        port = 389
        result = connectivity_results.LdapConnectResult(True, host, port)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.LdapConnectResult.SUCCESS_MESSAGE,
            host=host,
            port=port)

    def test_ldap_connect_failure_output(self):
        host = 'host'
        port = 389
        exception = ConnectivityMockException()
        result = connectivity_results.LdapConnectResult(False, host, port, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.LdapConnectResult.FAILURE_MESSAGE,
            host=host,
            port=port)
        self.mock_logger.debug.assert_called_with(
            connectivity_results.ConnectResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))


class TestLdapBindResultOutput(ResultOutputTest):
    def test_success_result_output(self):
        bind_username = 'DN=gooduser mcgee'
        result = connectivity_results.LdapBindResult(True, bind_username)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.LdapBindResult.SUCCESS_MESSAGE,
            bind_username=bind_username)

    def test_failure_result_output(self):
        bind_username = 'DN=baduser mcbaddy'
        exception = ConnectivityMockException()
        result = connectivity_results.LdapBindResult(False, bind_username, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.LdapBindResult.FAILURE_MESSAGE, bind_username=bind_username),
            mock.call(connectivity_results.LdapBindResult.FAILURE_MESSAGE_CREDS),
        ])
        self.mock_logger.debug.assert_called_with(
            connectivity_results.LdapBindResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))

    def test_failure_certs_result_output(self):
        bind_username = 'DN=gooduser mcgee'
        exception = SSL.Error()
        result = connectivity_results.LdapBindResult(False, bind_username, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.LdapBindResult.FAILURE_MESSAGE, bind_username=bind_username),
            mock.call(connectivity_results.LdapBindResult.FAILURE_MESSAGE_CERTS),
        ])
        self.mock_logger.debug.assert_called_with(
            connectivity_results.LdapBindResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))


class TestLdapFilterParseResultOutput(ResultOutputTest):
    def test_success_result_output(self):
        filter_text = 'an exemplary filter'
        result = connectivity_results.LdapFilterParseResult(True, filter_text)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.LdapFilterParseResult.SUCCESS_MESSAGE,
            filter_text=filter_text)

    def test_failure_result_output(self):
        filter_text = 'a deplorable filter'
        result = connectivity_results.LdapFilterParseResult(False, filter_text)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.LdapFilterParseResult.FAILURE_MESSAGE,
            filter_text=filter_text)


class TestValidateResultOutput(ResultOutputTest):
    def test_success_result_output(self):
        result = connectivity_results.ValidateApiCredentialsResult(True, 'good_host', 'good_ikey')

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(connectivity_results.ValidateApiCredentialsResult.SUCCESS_MESSAGE)

    def test_failure_result_output(self):
        """A failure not due to bad CA certs (eg. a bad api host, a bad ikey) should emit
        a failure and a debug message as well as a notification to check creds.
        """
        bad_api_host = 'bad_host'
        bad_ikey = 'bad ikey'
        exception = ConnectivityMockException()
        result = connectivity_results.ValidateApiCredentialsResult(False, bad_api_host, bad_ikey, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE_CHECK_CREDS, api_host=bad_api_host, ikey=bad_ikey),
        ])
        self.mock_logger.debug.assert_called_with(
            connectivity_results.ValidateApiCredentialsResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))

    def test_failure_result_output_ca_certs(self):
        """A failure due to bad CA certs passed down from main should indicate during logging
        with an extra error log alongside the normal failure and debug messages
        """
        bad_api_host = 'bad_host'
        bad_ikey = 'bad ikey'
        exception = SSL.Error()
        result = connectivity_results.ValidateApiCredentialsResult(
            False,
            bad_api_host,
            bad_ikey,
            exception=exception,
        )

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE_CHECK_CERTS),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.ValidateApiCredentialsResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )

    def test_failure_result_socket_error(self):
        """socket.error is a subclass of IOError, but such an error indicates a problem with
        a bad api host rather than bad CA certs, as a type IOError exception would normally indicate-
        make sure our error output reflects this.
        """
        bad_api_host = 'bad_api_host'
        exception = socket.gaierror("can't resolve this host")
        result = connectivity_results.ValidateApiCredentialsResult(
            False,
            bad_api_host,
            'ikey',
            exception,
        )

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.ValidateApiCredentialsResult.FAILURE_MESSAGE_CHECK_CREDS, api_host=bad_api_host, ikey='ikey'),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.ValidateApiCredentialsResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )


class TestLdapSearchResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestLdapSearchResultOutput, self).setUp()

        ldap_filter_attr = pureldap.LDAPAttributeDescription('objectclass')
        ldap_filter_value = pureldap.LDAPAssertionValue('someclass')
        self.ldap_filter_object = pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_equalityMatch(ldap_filter_attr, ldap_filter_value)]
        )

    def test_success_result_output(self):
        result = connectivity_results.LdapSearchResult(True, 'dn=good dn', self.ldap_filter_object)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(connectivity_results.LdapSearchResult.SUCCESS_MESSAGE)

    def test_no_results_output(self):
        search_dn = 'dn=bad dn'
        result = connectivity_results.LdapSearchResult(False, search_dn, self.ldap_filter_object)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.LdapSearchResult.NO_RESULTS_MESSAGE,
            search_dn=search_dn,
            filter=self.ldap_filter_object.asText())

    def test_error_result_output(self):
        search_dn = 'dn=bad dn'
        exception = ConnectivityMockException()
        result = connectivity_results.LdapSearchResult(False, search_dn, self.ldap_filter_object, exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.LdapSearchResult.FAILED_SEARCH_MESSAGE,
            search_dn=search_dn)
        self.mock_logger.debug.assert_called_with(
            connectivity_results.LdapSearchResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))


class TestSslFileResultOutput(ResultOutputTest):
    def test_ssl_key_file_success_result_output(self):
        path = 'good/path'
        result = connectivity_results.SslKeyFileResult(True, path)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info._assert_called_once_with(
            connectivity_results.SslKeyFileResult.SUCCESS_MESSAGE,
            file_path=path
        )

    def test_ssl_cert_file_success_result_output(self):
        path = 'good/path'
        result = connectivity_results.SslCertFileResult(True, path)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info._assert_called_once_with(
            connectivity_results.SslCertFileResult.SUCCESS_MESSAGE,
            file_path=path
        )

    def test_ssl_key_file_failure_result_output(self):
        path = 'bad/path'
        exception = ConnectivityMockException()
        result = connectivity_results.SslKeyFileResult(False, path, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error._assert_called_once_with(
            connectivity_results.SslKeyFileResult.FAILURE_MESSAGE,
            file_path=path,
        )
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslFileResult.EXCEPTION_MESSAGE,
            exception_string=str(exception),
        )

    def test_ssl_cert_file_failure_result_output(self):
        path = 'bad/path'
        exception = ConnectivityMockException()
        result = connectivity_results.SslCertFileResult(False, path, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error._assert_called_once_with(
            connectivity_results.SslCertFileResult.FAILURE_MESSAGE,
            file_path=path,
        )
        self.mock_logger.debug.assert_called_once_with(
            connectivity_results.SslFileResult.EXCEPTION_MESSAGE,
            exception_string=str(exception),
        )


class TestSslContextResultOutput(ResultOutputTest):
    def test_success(self):
        cipher_list = "DEFAULT"
        result = connectivity_results.SslContextResult(True, cipher_list)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info._assert_called_once_with(
            connectivity_results.SslContextResult.SUCCESS_MESSAGE,
        )

    def test_failure_mismatched(self):
        """When SslContextResult is unsuccessful due to a mismatched key-cert pair,
        it should also output an extra clarifying message to logger.error
        """
        cipher_list = "DEFAULT"
        exception = ConnectivityMockException()
        exception.message = [('stuff', 'SSL_CTX_check_private_key', 'other stuff')]

        result = connectivity_results.SslContextResult(False, cipher_list, exception)
        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE_MISMATCHED),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )

    def test_failure_bad_ciphers(self):
        """When SslContextResult is unsuccessful due to a bad cipher list, it should
        also output an extra clarifying message to logger.error
        """
        cipher_list = "BAD:BAD:BAD"
        exception = ConnectivityMockException()
        exception.message = [('stuff', 'SSL_CTX_set_cipher_list', 'other stuff')]

        result = connectivity_results.SslContextResult(False, cipher_list, exception)
        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE_BAD_CIPHERS, cipher_list=cipher_list),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )

    def test_failure_other_error(self):
        """If SslContextResult fails but it's not due to bad ciphers or mismatched creds,
        there shouldn't be any extra output other than the exception
        """
        cipher_list = "DEFAULT"
        exception = ConnectivityMockException()
        exception.message = [('some other', 'apocryphal', 'openssl error')]

        result = connectivity_results.SslContextResult(False, cipher_list, exception)
        result.to_log_output(self.mock_logger)

        self.mock_logger.error._assert_called_once_with(
            connectivity_results.SslContextResult.FAILURE_MESSAGE,
        )
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )


class TestSslResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestSslResultOutput, self).setUp()
        self.good_key_path = "good/key/path"
        self.good_cert_path = "good/cert/path"

        self.good_key_result = connectivity_results.SslKeyFileResult(True, self.good_key_path)
        self.good_cert_result = connectivity_results.SslCertFileResult(True, self.good_cert_path)
        self.good_context_result = connectivity_results.SslContextResult(True, "DEFAULT")

        self.test_name = 'SSL context'
        self.prereq_str = 'SSL key and cert'
        self.skipped_context_result = connectivity_results.UnmetPrerequisiteSkippedTestResult(self.test_name, self.prereq_str)

    def test_success(self):
        """On all SslResult outputs, we should output only a success message.
        The result of the key check, the cert check, and the SSL context check are ignored.
        """
        good_ssl_result = connectivity_results.SslResult(
            self.good_key_result,
            self.good_cert_result,
            self.good_context_result,
        )

        good_ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.info._assert_called_once_with(connectivity_results.SslResult.SUCCESS_MESSAGE)

    def test_failure_bad_key(self):
        """If the cert checked out fine but the key didn't, we should log that
        the SSL context validation check skipped along with the key/cert results.
        """
        bad_key_path = "bad/key/path"
        exception = ConnectivityMockException()

        bad_key_result = connectivity_results.SslKeyFileResult(False, bad_key_path, exception)
        ssl_result = connectivity_results.SslResult(
            bad_key_result,
            self.good_cert_result,
            self.skipped_context_result,
        )

        ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.warn._assert_called_once_with(
            connectivity_results.UnmetPrerequisiteSkippedTestResult.OUTPUT_MESSAGE,
            test_name=self.test_name,
            prereq=self.prereq_str,
        )
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslKeyFileResult.FAILURE_MESSAGE, file_path=bad_key_path),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslKeyFileResult.EXCEPTION_MESSAGE,
            exception=str(exception),
        )
        self.mock_logger.info._assert_called_once_with(
            connectivity_results.SslCertFileResult.SUCCESS_MESSAGE,
            file_path=self.good_cert_path,
        )

    def test_failure_bad_cert(self):
        """If the key checked out fine but the cert didn't, we should log that
        the SSL context validation check skipped along with the key/cert results.
        """
        bad_cert_path = "bad/cert/path"
        exception = ConnectivityMockException()

        bad_cert_result = connectivity_results.SslCertFileResult(False, bad_cert_path, exception)
        ssl_result = connectivity_results.SslResult(
            self.good_key_result,
            bad_cert_result,
            self.skipped_context_result,
        )

        ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.warn._assert_called_once_with(
            connectivity_results.UnmetPrerequisiteSkippedTestResult.OUTPUT_MESSAGE,
            test_name=self.test_name,
            prereq=self.prereq_str,
        )
        self.mock_logger.info._assert_called_once_with(
            connectivity_results.SslKeyFileResult.SUCCESS_MESSAGE,
            file_path=self.good_key_path,
        )
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslCertFileResult.FAILURE_MESSAGE, file_path=bad_cert_path),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslCertFileResult.EXCEPTION_MESSAGE,
            exception=str(exception),
        )

    def test_failure_bad_context_validate(self):
        """If the SSL context validation check errors out but not due to bad ciphers
        or a mismatched key-cert pair, we should output just the standard 4 non-debug messages.
        """
        exception = ConnectivityMockException()
        exception.message = [('openssl style', 'whatever', 'error')]

        bad_context_result = connectivity_results.SslContextResult(
            False,
            "DEFAULT",
            exception=exception,
        )
        ssl_result = connectivity_results.SslResult(
            self.good_key_result,
            self.good_cert_result,
            bad_context_result,
        )

        ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_has_calls([
            mock.call(connectivity_results.SslKeyFileResult.SUCCESS_MESSAGE, file_path=self.good_key_path),
            mock.call(connectivity_results.SslCertFileResult.SUCCESS_MESSAGE, file_path=self.good_cert_path),
        ])
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception=str(exception),
        )

    def test_failure_bad_ciphers(self):
        """Even if the key and cert check out fine, bad ciphers can prevent SSL context creation
        and should log a "bad certs error" message specifically along with the 3 other messages.
        """
        exception = ConnectivityMockException()
        exception.message = [('openssl style', 'SSL_CTX_set_cipher_list', 'error')]

        bad_context_result = connectivity_results.SslContextResult(
            False,
            "DEFAULT",
            exception=exception,
        )
        ssl_result = connectivity_results.SslResult(
            self.good_key_result,
            self.good_cert_result,
            bad_context_result,
        )

        ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_has_calls([
            mock.call(connectivity_results.SslKeyFileResult.SUCCESS_MESSAGE, file_path=self.good_key_path),
            mock.call(connectivity_results.SslCertFileResult.SUCCESS_MESSAGE, file_path=self.good_cert_path),
        ])
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE_BAD_CIPHERS, cipher_list="DEFAULT"),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception=str(exception),
        )

    def test_failure_mismatched(self):
        """A mismatched key-cert pair can check out fine individually but will raise during
        SSL context creation, and should output an extra help message explaining this along
        with the standard 4 messages.
        """
        exception = ConnectivityMockException()
        exception.message = [('openssl style', 'SSL_CTX_check_private_key', 'error')]

        bad_context_result = connectivity_results.SslContextResult(
            False,
            "DEFAULT",
            exception=exception,
        )
        ssl_result = connectivity_results.SslResult(
            self.good_key_result,
            self.good_cert_result,
            bad_context_result,
        )

        ssl_result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_has_calls([
            mock.call(connectivity_results.SslKeyFileResult.SUCCESS_MESSAGE, file_path=self.good_key_path),
            mock.call(connectivity_results.SslCertFileResult.SUCCESS_MESSAGE, file_path=self.good_cert_path),
        ])
        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.SslResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE),
            mock.call(connectivity_results.SslContextResult.FAILURE_MESSAGE_MISMATCHED),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.SslContextResult.EXCEPTION_MESSAGE,
            exception=str(exception),
        )


class TestSkippedResultOutput(ResultOutputTest):
    def test_config_problem_skipped_output(self):
        test_name = 'some test'
        config_key = 'config key'
        result = connectivity_results.ConfigProblemSkippedTestResult(test_name, config_key)

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(
            connectivity_results.ConfigProblemSkippedTestResult.OUTPUT_MESSAGE,
            test_name=test_name,
            config_key=config_key)

    def test_unmet_prereq_skipped_output(self):
        test_name = 'some test'
        prereq = 'some prereq'
        result = connectivity_results.UnmetPrerequisiteSkippedTestResult(test_name, prereq)

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(
            connectivity_results.UnmetPrerequisiteSkippedTestResult.OUTPUT_MESSAGE,
            test_name=test_name,
            prereq=prereq)


class TestConfigProblemResultOutput(ResultOutputTest):
    def test_missing_key_result_output(self):
        key = 'some_key'
        result = connectivity_results.MissingConfigKeyProblem(key)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.MissingConfigKeyProblem.MESSAGE,
            config_key=key)

    def test_invalid_key_result_output(self):
        key = 'some_key'
        value = 'some value'
        result = connectivity_results.InvalidConfigKeyProblem(key, value)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.InvalidConfigKeyProblem.MESSAGE,
            config_key=key,
            value=value)

    def test_config_error_result_output(self):
        error = ConfigError("Some config error yo")
        result = connectivity_results.ConfigErrorProblem(error)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.ConfigErrorProblem.MESSAGE,
            error_string=str(error))


class TestPingResultOutput(ResultOutputTest):
    def test_success_result_output(self):
        good_api_host = 'good_api_host'
        latency = 100
        result = connectivity_results.DuoPingResult(True, good_api_host, latency=latency)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info._assert_called_once_with(
            connectivity_results.DuoPingResult.SUCCESS_MESSAGE,
            api_host=good_api_host,
            latency=latency)

    def test_failure_result_output(self):
        """A failure not due to bad main http_ca_certs (eg. a bad api-host)
        should let the admin know to check their api host/network.
        """
        bad_api_host = 'bad_api_host'
        exception = ConnectivityMockException()
        result = connectivity_results.DuoPingResult(False, bad_api_host, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE, api_host=bad_api_host),
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE_CHECK_NETWORK),
        ])
        self.mock_logger.debug.assert_called_with(
            connectivity_results.DuoPingResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))

    def test_failure_result_output_ca_certs(self):
        """A failure due to bad CA certs should let the admin know to check their main
        section's cert results.
        """
        good_api_host = 'good_api_host'
        exception = SSL.Error()
        result = connectivity_results.DuoPingResult(False, good_api_host, exception=exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE, api_host=good_api_host),
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE_CHECK_CERTS),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.DuoPingResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )

    def test_failure_result_socket_error(self):
        """socket.error is a subclass of IOError, but such an error indicates a problem with
        a bad api host rather than bad CA certs, as a type IOError exception would normally indicate-
        make sure our error output reflects this.
        """
        bad_api_host = 'bad_api_host'
        exception = socket.gaierror("can't resolve this host")
        result = connectivity_results.DuoPingResult(False, bad_api_host, exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_has_calls([
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE, api_host=bad_api_host),
            mock.call(connectivity_results.DuoPingResult.FAILURE_MESSAGE_CHECK_NETWORK),
        ])
        self.mock_logger.debug._assert_called_once_with(
            connectivity_results.DuoPingResult.EXCEPTION_MESSAGE,
            exception_string=str(exception)
        )


class TestTimeDriftResultOutput(ResultOutputTest):
    def test_successful_result_output(self):
        result = connectivity_results.TimeDriftResult(True, 5)

        result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(
            connectivity_results.TimeDriftResult.SUCCESS_MESSAGE)

    def test_high_drift_result_output(self):
        drift = 5000000000
        result = connectivity_results.TimeDriftResult(False, drift)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.TimeDriftResult.HIGH_DRIFT_MESSAGE,
            drift=drift)

    def test_failed_connect_result_output(self):
        exception = ConnectivityMockException()
        result = connectivity_results.TimeDriftResult(False, 0, exception)

        result.to_log_output(self.mock_logger)

        self.mock_logger.error.assert_called_with(
            connectivity_results.TimeDriftResult.FAILED_CONNECT_MESSAGE)
        self.mock_logger.debug.assert_called_with(
            connectivity_results.TimeDriftResult.EXCEPTION_MESSAGE,
            exception_string=str(exception))


class TestLdapHostResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestLdapHostResultOutput, self).setUp()

        self.mocks = [
            self.mock_tcp_connect_result,
            self.mock_ldap_connect_result,
            self.mock_bind_result,
            self.mock_search_result,
        ]

        self.hostname = 'hostname'
        self.port = 389
        self.transport = 'transport'
        self.ldap_host_result = connectivity_results.LdapHostResult(
            self.hostname,
            self.port,
            self.transport,
            *self.mocks
        )

    def test_good_result_output(self):
        self.ldap_host_result.to_log_output(self.mock_logger)

        self._assert_info_only_called(
            connectivity_results.LdapHostResult.SUCCESS_MESSAGE,
            host=self.hostname,
            port=self.port,
            transport=self.transport)
        self._assert_not_logged(self.mocks)

    def test_bad_result_output(self):
        self.mock_ldap_connect_result.is_successful.return_value = False

        self.ldap_host_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(
            connectivity_results.LdapHostResult.FAILURE_MESSAGE,
            host=self.hostname,
            port=self.port,
            transport=self.transport)
        self._assert_logged(self.mocks)


class TestAdClientResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestAdClientResultOutput, self).setUp()

        self.mock_config_result = mock.Mock(spec=connectivity_results.ConfigCheckResult)
        self.mock_host_result_1 = mock.Mock(spec=connectivity_results.LdapHostResult)
        self.mock_host_result_2 = mock.Mock(spec=connectivity_results.LdapHostResult)
        self.ad_client_result = connectivity_results.AdClientResult(
            self.mock_config_result,
            [
                self.mock_host_result_1,
                self.mock_host_result_2
            ]
        )

    def test_good_result_output(self):
        self.mock_config_result.is_successful.return_value = True
        self.mock_host_result_1.is_successful.return_value = True
        self.mock_host_result_2.is_successful.return_value = True

        self.ad_client_result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(connectivity_results.AdClientResult.SUCCESS_MESSAGE)
        self.mock_logger.warn.assert_not_called()
        self.mock_logger.error.assert_not_called()
        self.mock_logger.debug.assert_not_called()
        self.mock_config_result.to_log_output.assert_not_called()
        self.mock_host_result_1.to_log_output.assert_not_called()
        self.mock_host_result_2.to_log_output.assert_not_called()

    def test_bad_config_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        self.mock_host_result_1.is_successful.return_value = True
        self.mock_host_result_2.is_successful.return_value = True

        self.ad_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_one_bad_host_result_output(self):
        self.mock_config_result.is_successful.return_value = True
        self.mock_host_result_1.is_successful.return_value = False
        self.mock_host_result_2.is_successful.return_value = True

        self.ad_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_everything_bad_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        self.mock_host_result_1.is_successful.return_value = False
        self.mock_host_result_2.is_successful.return_value = False

        self.ad_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_no_host_failure_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        result = connectivity_results.AdClientResult(self.mock_config_result, [])

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()


class TestRadiusClientResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestRadiusClientResultOutput, self).setUp()

        self.mock_config_result = mock.Mock(spec=connectivity_results.ConfigCheckResult)
        self.mock_host_result_1 = mock.Mock(spec=connectivity_results.RadiusConnectResult)
        self.mock_host_result_2 = mock.Mock(spec=connectivity_results.RadiusConnectResult)
        self.radius_client_result = connectivity_results.RadiusClientResult(
            self.mock_config_result,
            [
                self.mock_host_result_1,
                self.mock_host_result_2
            ]
        )

    def test_good_result_output(self):
        self.mock_config_result.is_successful.return_value = True
        self.mock_host_result_1.is_successful.return_value = True
        self.mock_host_result_2.is_successful.return_value = True

        self.radius_client_result.to_log_output(self.mock_logger)

        self.mock_logger.info.assert_called_with(connectivity_results.RadiusClientResult.SUCCESS_MESSAGE)
        self.mock_logger.warn.assert_not_called()
        self.mock_logger.error.assert_not_called()
        self.mock_logger.debug.assert_not_called()
        self.mock_config_result.to_log_output.assert_not_called()
        self.mock_host_result_1.to_log_output.assert_not_called()
        self.mock_host_result_2.to_log_output.assert_not_called()

    def test_bad_config_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        self.mock_host_result_1.is_successful.return_value = True
        self.mock_host_result_2.is_successful.return_value = True

        self.radius_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.RadiusClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_one_bad_host_result_output(self):
        self.mock_config_result.is_successful.return_value = True
        self.mock_host_result_1.is_successful.return_value = False
        self.mock_host_result_2.is_successful.return_value = True

        self.radius_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.RadiusClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_everything_bad_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        self.mock_host_result_1.is_successful.return_value = False
        self.mock_host_result_2.is_successful.return_value = False

        self.radius_client_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.RadiusClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()
        self.mock_host_result_1.to_log_output.assert_called()
        self.mock_host_result_2.to_log_output.assert_called()

    def test_no_host_failure_result_output(self):
        self.mock_config_result.is_successful.return_value = False
        result = connectivity_results.RadiusClientResult(self.mock_config_result, [])

        result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.RadiusClientResult.FAILURE_MESSAGE)
        self.mock_config_result.to_log_output.assert_called()


class TestRadiusServerResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestRadiusServerResultOutput, self).setUp()

        self.mocks = [
            self.mock_config_result,
            self.mock_ping_result,
            self.mock_drift_result,
            self.mock_validate_result,
            self.mock_listen_result
        ]

        self.radius_server_result = connectivity_results.RadiusServerResult(*self.mocks)

    def test_good_result_output(self):
        self.radius_server_result.to_log_output(self.mock_logger)

        self._assert_info_only_called(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        self._assert_not_logged(self.mocks)

    def test_bad_result_output(self):
        for one_mock in self.mocks:
            one_mock.is_successful.return_value = False

        self.radius_server_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.RadiusServerResult.FAILURE_MESSAGE)
        self._assert_logged(self.mocks)


class TestLdapServerResultOutput(ResultOutputTest):
    def setUp(self):
        super(TestLdapServerResultOutput, self).setUp()

        self.mocks = [
            self.mock_config_result,
            self.mock_ping_result,
            self.mock_drift_result,
            self.mock_validate_result,
            self.mock_ssl_result,
            self.mock_listen_result,
        ]

        self.ldap_server_result = connectivity_results.LdapServerResult(*self.mocks)

    def test_good_result_output(self):

        self.ldap_server_result.to_log_output(self.mock_logger)

        self._assert_info_only_called(connectivity_results.LdapServerResult.SUCCESS_MESSAGE)
        self._assert_not_logged(self.mocks)

    def test_bad_result_output(self):
        self.mock_ping_result.is_successful.return_value = False

        self.ldap_server_result.to_log_output(self.mock_logger)

        self.mock_logger.warn.assert_called_with(connectivity_results.LdapServerResult.FAILURE_MESSAGE)
        self._assert_logged(self.mocks)


if __name__ == '__main__':
    import unittest
    unittest.main()
