#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

import textwrap
from collections import OrderedDict

from twisted.internet import defer
from twisted.logger import Logger
import mock

from duoauthproxy.lib.ldap.client import AD_TRANSPORT_CLEAR
from duoauthproxy.lib.config_provider import ConfigProvider

from duoauthproxy.lib.connectivity import connectivity_tester, connectivity_results
from duoauthproxy.lib.connectivity.check import ad_client
from base import (
    ModuleConnectivityTest,
    MockConnectivityTestToolbox,
    MockTesterResolver,
    MockProtocol,
)


class TestConnectivityTester(ModuleConnectivityTest):

    def setUp(self):
        super(TestConnectivityTester, self).setUp()
        self.maxDiff = None

        self.mock_logger = mock.Mock(spec=Logger)

        self.BIG_CONFIG = ConfigProvider(textwrap.dedent("""
            [main]
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt

            [cloud]
            ikey=someikey
            skey=someskey
            api_host=some.api.host

            [http_proxy]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

            [ad_client1]
            host=ad.domain.controller.one
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [ad_client2]
            host=ad.domain.controller.two
            host_2=ad.domain.controller.two.by.two
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [radius_client]
            host=11.22.33.44
            port=1813
            secret=radiussecretone
            host_2=44.33.22.11
            port_2=1814
            secret_2=radiussecrettwo

            [duo_only_client]

            [radius_server_auto]
            ikey=ikey
            skey=skey
            api_host=api.host
            port=1818
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_2=radiussecrettwo

            [radius_server_iframe]
            ikey=ikey
            skey=skey
            api_host=api.host
            port=1819
            radius_ip_1=111.222.111.222
            radius_secret_protected_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_protected_2=radiussecrettwo

            [radius_server_challenge]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [radius_server_concat]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [radius_server_duo_only]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [radius_server_eap]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [ldap_server_auto]
            ikey=ikey
            skey=skey
            api_host=api.host
        """))

        self.RADIUS_SERVER_RESULT = connectivity_results.RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.GOOD_LISTEN_UDP_RESULT
        )

        self.CONFIG_RESULT = OrderedDict([
            ('main', connectivity_results.MainSectionResult(
                self.GOOD_CERTS_RESULT,
                connectivity_results.NOT_APPLICABLE_TEST_RESULT)),
            ('cloud', connectivity_results.SKIPPED_SECTION_RESULT),
            ('http_proxy', connectivity_results.HttpProxyResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                self.GOOD_LISTEN_TCP_RESULT,
                self.GOOD_PING_RESULT,
                self.GOOD_TIME_DRIFT_RESULT)),
            ('ad_client1', connectivity_results.AdClientResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                [
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.one',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.one', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    )
                ]
            )),
            ('ad_client2', connectivity_results.AdClientResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                [
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.two',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.two', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    ),
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.two.by.two',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.two.by.two', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    )
                ]
            )),
            ('radius_client', connectivity_results.RadiusClientResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                [self.GOOD_CONNECT_RADIUS_RESULT, self.GOOD_CONNECT_RADIUS_RESULT]
            )),
            ('duo_only_client', connectivity_results.SKIPPED_SECTION_RESULT),
            ('radius_server_auto', self.RADIUS_SERVER_RESULT),
            ('radius_server_iframe', self.RADIUS_SERVER_RESULT),
            ('radius_server_challenge', self.RADIUS_SERVER_RESULT),
            ('radius_server_concat', self.RADIUS_SERVER_RESULT),
            ('radius_server_duo_only', self.RADIUS_SERVER_RESULT),
            ('radius_server_eap', self.RADIUS_SERVER_RESULT),
            ('ldap_server_auto', connectivity_results.LdapServerResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                self.GOOD_PING_RESULT,
                self.GOOD_TIME_DRIFT_RESULT,
                self.GOOD_VALIDATE_RESULT,
                connectivity_results.NOT_APPLICABLE_TEST_RESULT,
                self.GOOD_LISTEN_TCP_RESULT
            )),
        ])

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()
        self.RESOLVER = MockTesterResolver()

    @defer.inlineCallbacks
    def test_connectivity_tester_with_all_possible_sections(self):
        expected_result = {
            'main': self.RESOLVER.MOCK_RESULT,
            'cloud': self.RESOLVER.MOCK_RESULT,
            'http_proxy': self.RESOLVER.MOCK_RESULT,
            'ad_client1': self.RESOLVER.MOCK_RESULT,
            'ad_client2': self.RESOLVER.MOCK_RESULT,
            'radius_client': self.RESOLVER.MOCK_RESULT,
            'duo_only_client': self.RESOLVER.MOCK_RESULT,
            'radius_server_auto': self.RESOLVER.MOCK_RESULT,
            'radius_server_iframe': self.RESOLVER.MOCK_RESULT,
            'radius_server_challenge': self.RESOLVER.MOCK_RESULT,
            'radius_server_concat': self.RESOLVER.MOCK_RESULT,
            'radius_server_duo_only': self.RESOLVER.MOCK_RESULT,
            'radius_server_eap': self.RESOLVER.MOCK_RESULT,
            'ldap_server_auto': self.RESOLVER.MOCK_RESULT,
        }

        actual_result = yield connectivity_tester.test_config(self.BIG_CONFIG,
                                                              self.mock_logger,
                                                              toolbox=self.SUCCESS_TOOLBOX,
                                                              section_test_resolver=self.RESOLVER)
        self.mock_logger.info.assert_called()
        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_real_resolver(self):
        """An integration test of the connectivity tester + real resolver"""
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)
        expected_result = self.CONFIG_RESULT

        actual_result = yield connectivity_tester.test_config(self.BIG_CONFIG,
                                                              self.mock_logger,
                                                              toolbox=self.SUCCESS_TOOLBOX)

        # Main Section
        self.mock_logger.info.assert_any_call(connectivity_results.MainSectionResult.SUCCESS_MESSAGE)
        # http proxy
        self.mock_logger.info.assert_any_call(connectivity_results.HttpProxyResult.SUCCESS_MESSAGE)
        # ad client 1
        self.mock_logger.info.assert_any_call(connectivity_results.AdClientResult.SUCCESS_MESSAGE)
        # ad client 2
        self.mock_logger.info.assert_any_call(connectivity_results.AdClientResult.SUCCESS_MESSAGE)
        # radius client
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusClientResult.SUCCESS_MESSAGE)
        # radius auto
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius iframe
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius challenge
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius concat
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius eap
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # ldap auto
        self.mock_logger.info.assert_any_call(connectivity_results.LdapServerResult.SUCCESS_MESSAGE)

        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_real_resolver_and_failed_ad_client(self):
        """An integration test where we mock ad client to fail the bind"""
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        skipped_search_result = connectivity_results.UnmetPrerequisiteSkippedTestResult('search', 'bind')
        for res in self.CONFIG_RESULT['ad_client1'].ldap_host_results:
            res.bind_result = self.BAD_BIND_SERVICE_ACCOUNT
            res.search_result = skipped_search_result

        for res in self.CONFIG_RESULT['ad_client2'].ldap_host_results:
            res.bind_result = self.BAD_BIND_SERVICE_ACCOUNT
            res.search_result = skipped_search_result
        expected_result = self.CONFIG_RESULT

        toolbox = MockConnectivityTestToolbox(fail_bind_service_account=True)
        actual_result = yield connectivity_tester.test_config(self.BIG_CONFIG,
                                                              self.mock_logger,
                                                              toolbox=toolbox)

        # Main Section
        self.mock_logger.info.assert_any_call(connectivity_results.MainSectionResult.SUCCESS_MESSAGE)
        # http proxy
        self.mock_logger.info.assert_any_call(connectivity_results.HttpProxyResult.SUCCESS_MESSAGE)
        # ad client 1
        self.mock_logger.warn.assert_any_call(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_logger.warn.assert_any_call(
                connectivity_results.LdapHostResult.FAILURE_MESSAGE,
                host='ad.domain.controller.one',
                port=389,
                transport=AD_TRANSPORT_CLEAR
        )
        self.mock_logger.error.assert_any_call(
                connectivity_results.LdapBindResult.FAILURE_MESSAGE,
                bind_username='bad_username'
        )
        # ad client 2
        self.mock_logger.warn.assert_any_call(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        self.mock_logger.warn.assert_any_call(
                connectivity_results.LdapHostResult.FAILURE_MESSAGE,
                host='ad.domain.controller.two',
                port=389,
                transport=AD_TRANSPORT_CLEAR
        )
        self.mock_logger.warn.assert_any_call(
                connectivity_results.LdapHostResult.FAILURE_MESSAGE,
                host='ad.domain.controller.two.by.two',
                port=389,
                transport=AD_TRANSPORT_CLEAR
        )
        # radius client
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusClientResult.SUCCESS_MESSAGE)
        # radius auto
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius iframe
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius challenge
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius concat
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius eap
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # ldap auto
        self.mock_logger.info.assert_any_call(connectivity_results.LdapServerResult.SUCCESS_MESSAGE)

        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_incomplete_config(self):
        incomplete_config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=
            service_account_username=
            service_account_password=
            search_dn=

            [radius_server_auto]
            ikey=
            skey=
            api_host=
            radius_ip_1=
            radius_secret_1=
            failmode=safe
            client=ad_client
            port=1812
        """))

        yield connectivity_tester.test_config(incomplete_config,
                                              self.mock_logger,
                                              toolbox=self.SUCCESS_TOOLBOX)

        # AD Client failure
        self.mock_logger.warn.assert_any_call(connectivity_results.AdClientResult.FAILURE_MESSAGE)
        # Radius server failure
        self.mock_logger.warn.assert_any_call(connectivity_results.RadiusServerResult.FAILURE_MESSAGE)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_blank_config(self):
        blank_config = ConfigProvider('')

        yield connectivity_tester.test_config(blank_config,
                                              self.mock_logger,
                                              toolbox=self.SUCCESS_TOOLBOX)

        self.mock_logger.warn.assert_any_call(connectivity_tester.NO_SECTIONS_MESSAGE)
        self.mock_logger.info.assert_not_called()

    @defer.inlineCallbacks
    def test_connectivity_tester_catches_high_level_exceptions(self):
        """
        An integration test where we catch any exceptions at a high
        level to prevent the tool from completely crashing
        """

        with mock.patch('duoauthproxy.lib.connectivity.connectivity_tester._test_all_sections', mock.Mock(side_effect=Exception)):
            yield connectivity_tester.test_config(self.BIG_CONFIG,
                                                  self.mock_logger,
                                                  toolbox=self.SUCCESS_TOOLBOX,
                                                  section_test_resolver=self.RESOLVER)

            self.mock_logger.error.assert_called()

    @defer.inlineCallbacks
    def test_connectivity_tester_with_skipped_sections(self):
        skipping_config = ConfigProvider(textwrap.dedent("""
            [duo_only_client]

            [misspelled_section_name]
            ikey=DIAAA
            skey=AAA
            api_host=api-host.com
        """))

        yield connectivity_tester.test_config(skipping_config,
                                              self.mock_logger,
                                              toolbox=self.SUCCESS_TOOLBOX)

        # Duo Only Client skip
        self.mock_logger.info.assert_any_call(connectivity_results.SkippedSectionResult.MESSAGE)
        # Bad section name
        self.mock_logger.error.assert_any_call(connectivity_results.InvalidSectionResult.MESSAGE)


if __name__ == '__main__':
    import unittest
    unittest.main()
