#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import BaseHTTPServer
import functools
import io
import json

from twisted.internet import defer
from twisted.web.client import HTTPClientFactory

import mox

from duoauthproxy.lib import drpc
from duoauthproxy.lib import duo_async
from duoauthproxy.lib import http

from base import StringTransport, TestCase

HOST = 'example.test'


class MockHTTP(object):
    def request(self, method, url, body, headers, timeout=None):
        if not timeout:
            self.request = (method, url, body, headers)
            return defer.succeed(self.response)
        else:
            # Forcing a timeout response
            self.request = (method, url, body, headers)
            return defer.fail(defer.TimeoutError("Response timed out"))


class TestDuoAsync_parse_response(TestCase):
    def setUp(self):
        # When testing _parse_response(), don't use mox because of the
        # chance of false positives due to unexpected method calls.
        self.mhttp = MockHTTP()
        self.client = duo_async.AsyncDuoClient(
            host=HOST,
            ikey='DIQ8V458EE6BEWKGIAS0',
            skey='6HPo5AwYcNVSUGqvcrOgByvVrhcyd32w8kIFT4PJ',
            port=443,
            requestor=self.mhttp.request,
        )

    @defer.inlineCallbacks
    def test_duo_error(self):
        payload = {
            'code': 40001,
            'stat': 'FAIL',
            'message': 'Missing required request parameter'
        }
        self.mhttp.response = (400, json.dumps(payload), {})
        with self.assertRaises(duo_async.DuoAPIFailClosedError):
            yield self.client.call('POST', '/rest/v1/auth', {})

    @defer.inlineCallbacks
    def test_fail_response(self):
        payload = {
            'code': 20202,
            'stat': 'FAIL',
            'message': 'Missing required request parameter'
        }
        self.mhttp.response = (200, json.dumps(payload), {})
        with self.assertRaises(duo_async.DuoAPIFailClosedError):
            yield self.client.call('POST', '/rest/v1/status', {})

    @defer.inlineCallbacks
    def test_response_not_json(self):
        self.mhttp.response = (200, 'wtf', {})
        with self.assertRaises(duo_async.DuoAPIFailClosedError):
            yield self.client.call('POST', '/rest/v1/ping', {})

    @defer.inlineCallbacks
    def test_auth_timeout(self):
        self.client.timeout_default = 3
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'allow',
                'status': 'hi',
            }
        }
        self.mhttp.response = (200, json.dumps(payload), {})
        with self.assertRaises(duo_async.DuoAPIFailClosedError):
            yield self.client.auth('some_user_1', '123456', '1.2.3.4')


class TestDuoAsync(TestCase):
    PORT = 443
    IKEY = 'DIQ8V458EE6BEWKGIAS0'
    SKEY = '6HPo5AwYcNVSUGqvcrOgByvVrhcyd32w8kIFT4PJ'

    def setUp(self):
        self.mhttp = mox.MockObject(http.HTTPClient)

        def requestor(*args, **kwargs):
            return defer.maybeDeferred(self.mhttp.request,
                                       *args, **kwargs)
        self.client = duo_async.AsyncDuoClient(
            HOST, self.IKEY, self.SKEY,
            self.PORT, requestor)

    def expect_request(self, method, url, body, response):
        def assert_headers(headers):
            if body:
                self.assertEqual('application/x-www-form-urlencoded',
                                 headers['Content-Type'])
            else:
                self.assertNotIn('Content-Type', headers)
            self.assertTrue(
                headers['Authorization'].startswith(
                    'Basic RElROFY0NThFRTZCRVdLR0lBUzA'
                )
            )
            return True
        self.mhttp.request(
            method,
            ('https://{0}:{1:d}{2:s}'.format(HOST, self.PORT, url)),
            body,
            mox.Func(assert_headers),
            timeout=0,
        ).AndReturn(response)

    @defer.inlineCallbacks
    def test_signature_GET(self):
        payload = {
            'stat': 'OK',
            'response': {'Hello': 'World'}
        }
        expected_url = '/om/nom/nom?bar=bob&baz=mallory&foo=alice'

        def assert_headers(headers):
            self.assertNotIn('Content-Type', headers)
            self.assertEqual(headers['Authorization'],
                             'Basic RElROFY0NThFRTZCRVdLR0lBUzA6MDZiNWY3YWIzYzUyNjU1M2MzNWQ1ODc0MWYwMTVjMzMxNDcyOTBhNQ==')
            return True
        self.mhttp.request(
            'GET',
            ('https://{0}:{1:d}{2:s}'.format(HOST, self.PORT, expected_url)),
            None,
            mox.Func(assert_headers),
            timeout=0,
        ).AndReturn((200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        rsp = yield self.client.call('GET', '/om/nom/nom', {
            'foo': 'alice',
            'bar': 'bob',
            'baz': 'mallory',
        })
        mox.Verify(self.mhttp)
        self.assertEqual(payload['response'], rsp)

    @defer.inlineCallbacks
    def test_signature_POST(self):
        payload = {
            'stat': 'OK',
            'response': {'Hello': 'World'}
        }
        url = '/om/nom/nom'

        def assert_headers(headers):
            self.assertEqual('application/x-www-form-urlencoded',
                             headers['Content-Type'])
            self.assertEqual(headers['Authorization'],
                             'Basic RElROFY0NThFRTZCRVdLR0lBUzA6N2UwZmY5YWU4NDZiNjU3ODNkZTBhYWU0NzE4N2I2NTYwNDRlZDE3NQ==')
            return True
        self.mhttp.request(
            'POST',
            ('https://{0}:{1:d}{2:s}'.format(HOST, self.PORT, url)),
            'bar=bob&baz=mallory&foo=alice',
            mox.Func(assert_headers),
            timeout=0,
        ).AndReturn((200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        rsp = yield self.client.call('POST', '/om/nom/nom', {
            'foo': 'alice',
            'bar': 'bob',
            'baz': 'mallory',
        })
        mox.Verify(self.mhttp)
        self.assertEqual(payload['response'], rsp)

    @defer.inlineCallbacks
    def test_preauth_well_formed_auth_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'auth',
                'factors': {'1': 'phone1',
                            'default': 'phone1'},
                'prompt': 'yeah, whatever'
            }
        }
        self.expect_request('POST', '/rest/v1/preauth', 'user=some_user_1',
                            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        response = yield self.client.preauth('some_user_1')
        mox.Verify(self.mhttp)
        self.assertEqual(response, payload['response'])

    @defer.inlineCallbacks
    def test_preauth_well_formed_deny_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'deny',
                'status': 'shoo. shoo, i tell you'
            }
        }
        self.expect_request('POST', '/rest/v1/preauth', 'user=some_user_1',
                            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        response = yield self.client.preauth('some_user_1')
        mox.Verify(self.mhttp)
        self.assertEqual(response, payload['response'])

    @defer.inlineCallbacks
    def test_preauth_malformed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'foo',
                'status': 'bar'
            }
        }
        self.expect_request('POST', '/rest/v1/preauth', 'user=some_user_1',
                            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        with self.assertRaises(duo_async.DuoAPIFailOpenError):
            yield self.client.preauth('some_user_1')
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_auth_well_formed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'allow',
                'status': 'everything looks good sir',
            }
        }
        self.expect_request(
            'POST', '/rest/v1/auth',
            'auto=123456&factor=auto&ipaddr=1.2.3.4&user=some_user_1',
            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        response = yield self.client.auth('some_user_1', '123456', '1.2.3.4')
        mox.Verify(self.mhttp)
        self.assertEqual(response, payload['response'])

    @defer.inlineCallbacks
    def test_auth_malformed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'result': 'allow',
            }
        }
        self.expect_request(
            'POST', '/rest/v1/auth',
            'auto=123456&factor=auto&ipaddr=1.2.3.4&user=some_user_1',
            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        with self.assertRaises(duo_async.DuoAPIFailOpenError):
            yield self.client.auth('some_user_1', '123456', '1.2.3.4')
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_proxy_init_well_formed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'proxy_txid': '123456'
            }
        }
        self.expect_request(
            'POST', '/rest/v1/tx/proxy_init',
            'user=some_user_1',
            (200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        response = yield self.client.proxy_init('some_user_1')
        mox.Verify(self.mhttp)
        self.assertEqual(response, payload['response'])

    @defer.inlineCallbacks
    def test_proxy_init_malformed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'bzsdfzxfv': 'sdfglkjxbv',
            }
        }
        self.expect_request(
            'POST', '/rest/v1/tx/proxy_init',
            'user=some_user_1',
            (200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        with self.assertRaises(duo_async.DuoAPIFailOpenError):
            yield self.client.proxy_init('some_user_1')
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_proxy_finish_well_formed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'valid_cookie': False
            }
        }
        self.expect_request(
            'POST', '/rest/v1/tx/proxy_finish',
            'auth_cookie=BAADBEEFDEADBEEF',
            (200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        response = yield self.client.proxy_finish('BAADBEEFDEADBEEF')
        mox.Verify(self.mhttp)
        self.assertEqual(response, payload['response'])

    @defer.inlineCallbacks
    def test_proxy_finish_malformed_response(self):
        # valid_cookie True but user missing.
        payload = {
            'stat': 'OK',
            'response': {
                'valid_cookie': True
            }
        }
        self.expect_request(
            'POST', '/rest/v1/tx/proxy_finish',
            'auth_cookie=BAADBEEFDEADBEEF',
            (200, json.dumps(payload), {}))

        mox.Replay(self.mhttp)
        with self.assertRaises(duo_async.DuoAPIFailOpenError):
            yield self.client.proxy_finish('BAADBEEFDEADBEEF')
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_ping_well_formed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'time': 26,
            }
        }
        self.expect_request('GET', '/auth/v2/ping', '',
                            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        yield self.client.ping()
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_ping_malformed_response(self):
        payload = {
            'stat': 'OK',
            'response': {
                'wat': 'duck',
            }
        }
        self.expect_request('GET', '/auth/v2/ping', '',
                            (200, json.dumps(payload), {}))
        mox.Replay(self.mhttp)
        with self.assertRaises(duo_async.DuoAPIFailOpenError):
            yield self.client.ping()
        mox.Verify(self.mhttp)

    @defer.inlineCallbacks
    def test_proxy_join(self):
        class ServerFactory(drpc.ServerFactory):
            pass

        payload = {
            'stat': 'OK',
            'response': {
                'time': 24,
            }
        }
        api_server_sentinel = object()
        self.expect_request('GET', '/auth/v2/ping', '',
                            (200, json.dumps(payload), {}))

        def assert_protocol_class(protocol_class):
            self.assertEqual(
                protocol_class.keywords['drpc_server_factory_class'],
                ServerFactory)
            return True
        self.mhttp.request(
            'POST',
            ('https://{0}:{1:d}/'.format(HOST, self.PORT)),
            '',
            {},
            protocol_class=mox.Func(assert_protocol_class),
            timeout=0,
            disconnect=False,
        ).AndReturn((200, api_server_sentinel, {}))

        mox.Replay(self.mhttp)
        response = yield self.client.proxy_join(ServerFactory)
        mox.Verify(self.mhttp)
        self.assertIs(api_server_sentinel, response)


class HTTPRequest(BaseHTTPServer.BaseHTTPRequestHandler):
    def __init__(self, request_str):
        self.rfile = io.BytesIO(request_str)
        self.raw_requestline = self.rfile.readline()
        self.parse_request()
        self.body = self.rfile.read()  # whatever's left (ignore Content-Length)

    def send_error(self, code, message):
        raise Exception('HTTP Request Invalid: {0:d} {1}'.format(
            code, message))


class DRPCServerProtocol(drpc.ServerProtocol):
    jsonify_sort_keys = True


class DRPCServerFactory(drpc.ServerFactory):
    protocol = DRPCServerProtocol

    def do_test(self):
        return {
            'ikey': self.ikey,
            'success': True,
        }


class TestHTTPClientDRPC(TestCase):
    ikey = 'foo'
    skey = 'bar'
    user_agent = 'User Agent (Not Like Mozilla) (No Slashes)'
    server_time = 24

    def connect(self):
        duo_client = duo_async.AsyncDuoClient(
            host=HOST,
            ikey=self.ikey,
            skey=self.skey,
            port=443,
            requestor=None,
        )
        factory = HTTPClientFactory(
            url=('http://' + HOST + ':334/'),
            agent=self.user_agent,
        )
        factory.protocol = functools.partial(
            duo_async.HTTPClientDRPC,
            duo_client=duo_client,
            drpc_server_factory_class=DRPCServerFactory,
        )
        protocol = factory.buildProtocol('/Test/Foo')
        trans = StringTransport()
        trans.protocol = protocol
        protocol.makeConnection(trans)
        return protocol

    def test_connectionMade(self):
        """
        connectionMade() must send a proxy_join API request over HTTP.
        """
        protocol = self.connect()
        request = HTTPRequest(protocol.transport.value())
        self.assertEqual('POST', request.command)
        self.assertEqual('/auth/v2/proxy_join', request.path)
        self.assertIn('date', request.headers)
        self.assertIn('authorization', request.headers)
        self.assertEqual(HOST, request.headers['host'])
        self.assertEqual(self.user_agent, request.headers['user-agent'])
        self.assertEqual('0', request.headers['content-length'])
        self.assertEqual('application/x-www-form-urlencoded',
                         request.headers['content-type'])
        self.assertEqual('upgrade', request.headers['connection'])
        self.assertEqual('drpc', request.headers['upgrade'])
        self.assertEqual('', request.body)

    @defer.inlineCallbacks
    def test_handleResponse_wrong_HTTP_status(self):
        protocol = self.connect()
        protocol.factory.status = 400
        protocol.handleResponse('')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_not_JSON(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse('')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_not_JSON_object(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse('1')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_failed(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse('{"stat": "FAIL"}')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_no_response(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse('{"stat": "OK"}')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_response_not_JSON_object(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse('{"stat": "OK", "response": null}')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_HTTP_response_wrong_version(self):
        protocol = self.connect()
        protocol.factory.status = 101
        protocol.handleResponse(
            '{"stat": "OK", "response": {"call_message_type": 1}}')
        with self.assertRaises(duo_async.DuoAPIProxyJoinError):
            yield protocol.factory.deferred
        self.assertTrue(protocol.transport.aborted)

    @defer.inlineCallbacks
    def connect_and_upgrade(self, extra_data='',
                            extra_proxy_join_response=None,
                            wrapping=None):
        """
        * extra_data: Injected after the JSON body of the Upgrade
          response.
        * wrapping: Simulate a transport that was wrapped by either
          'connectSSL' or 'startTLS'. None specifies no wrapping.
        * extra_proxy_join_response: Dict added to the proxy join
          response.
        """
        protocol = self.connect()

        # When testing wrapping, need to test whether the wrapped or
        # raw protocol was replaced.
        raw_protocol = protocol.transport.protocol
        if wrapping == 'connectSSL':
            protocol.transport.wrappedProtocol = 'wrapper not replaced'
        if wrapping == 'startTLS':
            protocol.transport.TLS = True

        proxy_join_response = {
            'call_message_type': 0,
            'foo': 'bar',
            'time': self.server_time,
        }
        if extra_proxy_join_response:
            proxy_join_response.update(extra_proxy_join_response)

        protocol.factory.status = 101
        res = json.dumps({
            'stat': 'OK',
            'response': proxy_join_response,
        })
        protocol.lineReceived('HTTP/1.0 101 Upgrade')
        protocol.lineReceived('Upgrade: drpc')
        protocol.lineReceived('Connection: upgrade')
        protocol.lineReceived('Content-Length: {0:d}'.format(len(res)))
        protocol.lineReceived('')
        protocol.rawDataReceived(res + extra_data)

        rpc_server = yield protocol.factory.deferred
        self.assertIs(protocol.rpc_server, rpc_server)

        # FIXME: these test duo_async's assumptions about how Twisted
        # works. Should use actual TLS transports.
        if wrapping is None:
            self.assertIs(protocol.transport.protocol, rpc_server)
        elif wrapping == 'connectSSL':
            self.assertIs(protocol.transport.protocol, raw_protocol)
            self.assertIs(protocol.transport.wrappedProtocol, rpc_server)
        elif wrapping == 'startTLS':
            self.assertIs(protocol.transport.protocol, raw_protocol)
            self.assertFalse(hasattr(protocol.transport, 'wrappedProtocol'))
            self.assertIs(protocol.transport.protocol.wrappedProtocol,
                          rpc_server)

        defer.returnValue((protocol, rpc_server))

    @defer.inlineCallbacks
    def test_handleResponse_success(self):
        """
        Test upgrading and then receiving a call message.
        """
        protocol, rpc_server = yield self.connect_and_upgrade()
        rpc_server.dataReceived(
            '\x00\x00\x00\x19'
            + '\x00{"call":"test","id":"z"}'
            + '\x00\x00\x00T'
            + '{"sig":"036456073b35e0267bd3309c6e7e2e443a611cb4edceba7f5c203bbbd31ff67d","time":25}'
        )
        request = HTTPRequest(protocol.transport.value())
        self.assertEqual(
            '\x00\x00\x00O\x00{"error":null,"error_args":{},"id":"z","result":{"ikey":"foo","success":true}}\x00\x00\x00T{"sig":"4e9663ca16a73135de3b49ee8f7684f4c885692b26331902cd1dd6dd800c1dde","time":24}',
            # Not really the body; just data sent after the request.
            request.body,
        )
        self.assertFalse(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_success_after_connectSSL(self):
        """
        Simulate upgrading an HTTPS connection.
        """
        protocol, rpc_server = yield self.connect_and_upgrade(
            wrapping='connectSSL',
        )

    @defer.inlineCallbacks
    def test_handleResponse_success_after_startTLS(self):
        """
        Simulate upgrading an HTTPS connection through an HTTPS proxy.
        """
        protocol, rpc_server = yield self.connect_and_upgrade(
            wrapping='startTLS',
        )

    @defer.inlineCallbacks
    def test_handleResponse_success_and_rpc_pass_through(self):
        """
        If the server sends an initial call message quickly enough it
        will be received before the HTTP response that triggers the
        upgrade is handled. Test that the HTTP protocol passes it
        through after instantiating the RPC protocol.
        """
        test_call = (
            '\x00\x00\x00\x19'
            + '\x00{"call":"test","id":"z"}'
            + '\x00\x00\x00T'
            + '{"sig":"036456073b35e0267bd3309c6e7e2e443a611cb4edceba7f5c203bbbd31ff67d","time":25}'
        )
        protocol, rpc_server = yield self.connect_and_upgrade(test_call)
        request = HTTPRequest(protocol.transport.value())
        self.assertEqual(
            '\x00\x00\x00O\x00{"error":null,"error_args":{},"id":"z","result":{"ikey":"foo","success":true}}\x00\x00\x00T{"sig":"4e9663ca16a73135de3b49ee8f7684f4c885692b26331902cd1dd6dd800c1dde","time":24}',
            # Not really the body; just data sent after the request.
            request.body,
        )
        self.assertFalse(protocol.transport.aborted)

    @defer.inlineCallbacks
    def test_handleResponse_zlib_disabled(self):
        """
        If the service doesn't say it supports auto-zlib, disable it.
        """
        protocol, rpc_server = yield self.connect_and_upgrade()
        self.assertIs(rpc_server.auto_zlib_min_length, None)
        self.assertTrue(rpc_server.enable_unzlib)

    @defer.inlineCallbacks
    def test_handleResponse_zlib_enabled(self):
        """
        If the service wants auto-zlib, use the requested threshold.
        """
        protocol, rpc_server = yield self.connect_and_upgrade(
            extra_proxy_join_response={
                'auto_zlib_min_length': 55,
            },
        )
        self.assertEqual(55, rpc_server.auto_zlib_min_length)
        self.assertTrue(rpc_server.enable_unzlib)

    @defer.inlineCallbacks
    def test_handleResponse_idle_rpc_disabled(self):
        """
        Verify response can disable idle rpc timeout.
        """
        protocol, rpc_server = yield self.connect_and_upgrade(
            extra_proxy_join_response={
                'idle_rpc_timeout': None,
            },
        )
        self.assertIs(rpc_server.factory.idle_rpc_timeout, None)

    @defer.inlineCallbacks
    def test_handleResponse_idle_rpc_configured(self):
        """
        Verify response can modify idle rpc timeout.
        """
        protocol, rpc_server = yield self.connect_and_upgrade(
            extra_proxy_join_response={
                'idle_rpc_timeout': 12,
            },
        )
        self.assertEqual(12, rpc_server.factory.idle_rpc_timeout)


if __name__ == '__main__':
    import unittest
    unittest.main()
