#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import socket
import time

from twisted.internet import reactor, defer, protocol, ssl
from twisted.internet.error import (
    AlreadyCalled, ConnectionAborted, ConnectionRefusedError)
from twisted.python import log
from twisted.protocols.basic import LineReceiver
from twisted.trial.unittest import FailTest
from twisted.web import server, resource
import twisted.web.client
import twisted.web.http
from OpenSSL import crypto, SSL

from duoauthproxy.lib import http

from base import TestCase

# note: intentionally avoiding duo.dabus, duo.web,
# to minimize other variables in this...

ca_cert_str = """-----BEGIN CERTIFICATE-----
MIIEyjCCA7KgAwIBAgIJALeTLuN1QCWfMA0GCSqGSIb3DQEBBQUAMIGdMQswCQYD
VQQGEwJVUzELMAkGA1UECAwCTUkxEjAQBgNVBAcMCUFubiBBcmJvcjEVMBMGA1UE
CgwMRHVvIFNlY3VyaXR5MRgwFgYDVQQLDA9IVFRQIFVOSVQgVEVTVFMxFTATBgNV
BAMMDFRlc3QgUm9vdCBDQTElMCMGCSqGSIb3DQEJARYWYWtnb29kQGR1b3NlY3Vy
aXR5LmNvbTAeFw0xNDAxMjIyMjI1MDNaFw0yNDAxMjAyMjI1MDNaMIGdMQswCQYD
VQQGEwJVUzELMAkGA1UECAwCTUkxEjAQBgNVBAcMCUFubiBBcmJvcjEVMBMGA1UE
CgwMRHVvIFNlY3VyaXR5MRgwFgYDVQQLDA9IVFRQIFVOSVQgVEVTVFMxFTATBgNV
BAMMDFRlc3QgUm9vdCBDQTElMCMGCSqGSIb3DQEJARYWYWtnb29kQGR1b3NlY3Vy
aXR5LmNvbTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALuEMEqC0Nae
cOWuzCwjBH45yEeYhZu2M1XZFePG5rvk8nNFgwRk1+QpMTYaptVr0M25pYrpDDUu
Qfc65BEelo/GVhc0o0UR18wCKPUZugGtZEojsP/Qs5BOatfPQxfVf5KQNLq4SrFg
rn/6kTSe5IweABdonig3ytrHnFgOKVLh4X2CM8Z54gO359EPvuyGaffPwcddcDZl
vrXPfvl/fVP7U/CD2Ez5DeOTAYCER81TIq6vfswN13XZpY79jry8P7qflEWZ7yGB
vGiSc7W0kNAlM9RIowLhayxGgvS8cr0fyjev1CP8uOLi7mXoe5+marKCsuTbrxr6
d4nZUtsRBu0CAwEAAaOCAQkwggEFMB0GA1UdDgQWBBRRm1uHAgloU/xIgc1l1foi
R+/0bjCB0gYDVR0jBIHKMIHHgBRRm1uHAgloU/xIgc1l1foiR+/0bqGBo6SBoDCB
nTELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3Ix
FTATBgNVBAoMDER1byBTZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RT
MRUwEwYDVQQDDAxUZXN0IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBk
dW9zZWN1cml0eS5jb22CCQC3ky7jdUAlnzAPBgNVHRMBAf8EBTADAQH/MA0GCSqG
SIb3DQEBBQUAA4IBAQAMDKnVRUvj/UjYU5xwsue0UIueII/aVWudoUhWRZ0pwZEK
3oelPM3jAct5CCd2IZzwy5hlGdHqkkor+G50ELvNqeg6GZkRQMrBqt9SE1bUibpU
x1BmxxguQNOn7JItN3JvZTjKBAuq0pB9gqazJgtdHpXeDpY1Rx2mt+TpSiQklIn9
JlhFK7hpE2lkT+nXV5BRZyVlWI+Xn+W7bRgykZHhux3cFxZF6zp8/yNqBAPVufRs
ThNX/r/SqXuYPadNTRa/GvrsN46w/UhmaW4cD3pdlnEnkr44OLbq9rG8YppNHJyR
7nKcdKlkovtFqbVNR9aRU3ls8EPqBAkccs5vYf6C
-----END CERTIFICATE-----
"""
ca_cert = crypto.load_certificate(crypto.FILETYPE_PEM, ca_cert_str)

wrong_ca_cert_str = """-----BEGIN CERTIFICATE-----
MIIDLjCCApegAwIBAgIJAO8B6jNt3y7rMA0GCSqGSIb3DQEBBQUAMG4xCzAJBgNV
BAYTAlVTMREwDwYDVQQIEwhNaWNoaWdhbjEVMBMGA1UEChMMRHVvIFNlY3VyaXR5
MQ8wDQYDVQQDEwZEZXYgQ0ExJDAiBgkqhkiG9w0BCQEWFWFkbWluQGR1b3NlY3Vy
aXR5LmNvbTAeFw0xMTAzMjkxNTUxMTZaFw0xNDAzMjgxNTUxMTZaMG4xCzAJBgNV
BAYTAlVTMREwDwYDVQQIEwhNaWNoaWdhbjEVMBMGA1UEChMMRHVvIFNlY3VyaXR5
MQ8wDQYDVQQDEwZEZXYgQ0ExJDAiBgkqhkiG9w0BCQEWFWFkbWluQGR1b3NlY3Vy
aXR5LmNvbTCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAp8CaEtMfQ1iFGHfo
B+8KHp3SF5/PzrgB795WjUVfLD7r80uxJJ7TLyF1OfAGPTbeMu+0mxpEITxgoOGU
m5URR1xjstGgmmw9haAYYTomuFZt8kmvezzjNYwJrmiKmehtFNNpfMCQ8nZsrU76
MxDW3RmqYRSGobND/N1TKCJDi2cCAwEAAaOB0zCB0DAdBgNVHQ4EFgQUmEdcLJIe
QG5dWsSgJYs5S9G5MGMwgaAGA1UdIwSBmDCBlYAUmEdcLJIeQG5dWsSgJYs5S9G5
MGOhcqRwMG4xCzAJBgNVBAYTAlVTMREwDwYDVQQIEwhNaWNoaWdhbjEVMBMGA1UE
ChMMRHVvIFNlY3VyaXR5MQ8wDQYDVQQDEwZEZXYgQ0ExJDAiBgkqhkiG9w0BCQEW
FWFkbWluQGR1b3NlY3VyaXR5LmNvbYIJAO8B6jNt3y7rMAwGA1UdEwQFMAMBAf8w
DQYJKoZIhvcNAQEFBQADgYEAc+M/dNk7i5sDIB4TJ6JXJGcRgi+zhl6M647Pj8+j
eh2t8JPd0IZ2egbodlElM7kBCJw5KO/oMSAP6zN+ImdfXQJkADOVQteoBkgG5Gok
q6Ylh8O1z8+u1x5xbeBQUt/Ck9iZtPpgN+IT9fZU8X0D9TpzjTGqr7aUqwmbmD3A
PBo=
-----END CERTIFICATE-----
"""
wrong_ca_cert = crypto.load_certificate(crypto.FILETYPE_PEM, wrong_ca_cert_str)

# Subject: C=US, ST=MI, O=Duo Security, OU=HTTP UNIT TESTS, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# SANs: (none)
http_cert_str = """-----BEGIN CERTIFICATE-----
MIIE3DCCA8SgAwIBAgIBATANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIyMjIyNTAzWhcNMjQwMTIwMjIyNTAzWjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAu+QLLybUhhRfYTYT4HkLQDR2TpeUp3sTi1UvIfGio8oGuAWA/LPq
gpN2oBr/x9LjCG8kRAjfrVerGi43N0V3Fl6SmGy3RqdiN/iXSFUn9af+o9HZEX2i
zOcWoLs3yjvHyrzDN6xSgXi1HJtW6FsDQNW58pfLwdA3yb4a7W1sloj1yEUsmjzI
uhKgMf5ZIQ/vd5cYnmzdEmyPGwyoY01GFM9iHWHsPyR+QKBoNJy9LLpJZGZ6UypV
0YwLkXOfL6R2SEDHm4ft5TB1UmRZCGfoBq5nuBwZKFUhdrTLh9Rb5BDCuzUlv8kS
VLawaoTyuH9VWw3obi0AVh7gtx7XQehqoQIDAQABo4IBOjCCATYwCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUI8l5gqaG8yjz6QctWMbojB4Ve0QwgdIGA1UdIwSByjCBx4AU
UZtbhwIJaFP8SIHNZdX6Ikfv9G6hgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkAt5Mu
43VAJZ8wNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMA0GCSqGSIb3DQEBBQUAA4IBAQABisV1k/UeatAAUtNSfUwx
KDf2tx7JO3IMr06dyGcCjwm9hUTaY/JcZi7xztgjrbp/AtrZYwdtJX4GN0lfGQ8m
8oaR7pMSfvRMEMaDtrQcR05FmuxH8lIIZZft8sgjOoZjtBbVrDfMTjMtsORZdt8q
EuJj0S1W8umkjr5mzCgwILdlX4832NbIgmPT+hWVnEooMgiOthMVnSoyy6IEjwLj
8WyDBQn2xpuasVSQPw5GzwaMShzjrsdAAVUD1z6K+xZAJ/ZrTZ2/Y5ckOQdwEEoh
982ziP6tHtzfUKn8CkuOoF3JvInjMb35grLt36ycT7ljk7hpfEK7lmENlGaOJ8ED
-----END CERTIFICATE-----
"""
http_cert = crypto.load_certificate(crypto.FILETYPE_PEM, http_cert_str)

http_key_str = """-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAu+QLLybUhhRfYTYT4HkLQDR2TpeUp3sTi1UvIfGio8oGuAWA
/LPqgpN2oBr/x9LjCG8kRAjfrVerGi43N0V3Fl6SmGy3RqdiN/iXSFUn9af+o9HZ
EX2izOcWoLs3yjvHyrzDN6xSgXi1HJtW6FsDQNW58pfLwdA3yb4a7W1sloj1yEUs
mjzIuhKgMf5ZIQ/vd5cYnmzdEmyPGwyoY01GFM9iHWHsPyR+QKBoNJy9LLpJZGZ6
UypV0YwLkXOfL6R2SEDHm4ft5TB1UmRZCGfoBq5nuBwZKFUhdrTLh9Rb5BDCuzUl
v8kSVLawaoTyuH9VWw3obi0AVh7gtx7XQehqoQIDAQABAoIBAC6P39mqhMZJbC2a
Cij5tjkOBzeIxvUOQhSoYorF5+kxGqMl8gYty/LZ+m7I4hQyBcNET7dwf4XkkOln
BJX06AMqIqgVYQHoXyXBvDPXrpRpMmiEi4eTG/ItXmUsoDHEyMlMKBC26qddM3zI
Xiwsycyaan9c8MiyAfFD5EZsdJmFIKVzqdyUOiLDkc/TDezUq5rPBsSevj8BtkiD
JDHeclex5Rpy5RiUjkydGJP3fiDhTyvw9dlhZk6C+bQaC7Wl1EBCD2h8Le9yzfve
6xt+comhMEMiDyiEwBUUpU0nIQZth9LlTkPZ6oOTFBvwpz9ldLY5VJB1nvYjr31F
k11CMkUCgYEA43xxIUVZU+XNhbivkI6tXJ1n6TKN3kK0W1hSnQ+uQZpbUH2zzsCG
oQvne2if6X/xpLYFPRp0pnW1jlzgq2GRVhW4wZyAQI6Tu/+c5LkyHmRHP8s+C3pe
gQm0ppFoe99rczw6JHRx622nVg1Blty9mySp9Elf5kRmU3VSplaGZocCgYEA03ES
BIzD85OBk0ms49fHPgrFOFk7TZjBj4LVlgmC3LkPAJxKjNmGYTcAb9B6lEWgZo6R
ojDzWmZc4J5kZIcqiOgkkRutVcyBLgzjr3nCpDqEIRhR20L//vTk3n3EkXrvZAMd
ujvwAvzZq67irxmM++6DLsSheusTnjfdvZZZx5cCgYEAtO9fk9TyzbY0F63jhhjV
bpqL8YESMbvgc972JpI4SVmaU5UHlOSwsBslzJRUzIXFDvrvg8AtE9D/zPG2iNmb
nVq9jtEJeU9t1dPjWFCsuluMCAOBn+N+kyfDFdEFx7DWZ9t5a1Fq4pReyiTVDEca
T/GDdBeAPUgeEzd3u3GKiLUCgYAxntNvpelGfFNoUV31imoDsVtfQfGc4EBsgXPG
Zi0Dx5oxtnv4WgQjRFkh5Hcf3McrykF7Iqm5iVQ6knFU9jORNYl/VFN1UyzttbCp
OZXoHKNtcNuk7TBOkQI1BKbgzUzo16JeVjqZOOWMWNpx5Tkk1nCKxTJwOkK8Wrf7
ChG/uQKBgHFTrVcTVgbcRSq+znqoJCi9L1bhlraHPSPjx6C3jI919nWj2+vz7XY6
KoCGK34ZTA+StCxasI2eT+T8DPuH4bpv9QhhcvdhfvOIYQCc958D8ga7liwgiYaK
8IxzyNvXrhJn7xweGwnlwRiKGLnlua7Vq1VxouM/tPRx9Sh+K2wo
-----END RSA PRIVATE KEY-----
"""
http_key = crypto.load_privatekey(crypto.FILETYPE_PEM, http_key_str)

# subject: C=US, ST=Michigan, L=Ann Arbor, O=Duo Security, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# SANs: DNS:localhost, DNS:127.0.0.1
http_san_cert_str = """-----BEGIN CERTIFICATE-----
MIIFADCCA+igAwIBAgIBAzANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIyMjIyNTAzWhcNMjQwMTIwMjIyNTAzWjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAx7ns8aPV+oC0bP9y5rmKrzOeySKtZNP4ZXh7oCLLmNPVdzwEFIJE
pdBdxC23Rypy8O1KLmdErMuYW14WUBBW/AmUgfHDVHg/gdkfKaKxhCDENbKK7vcA
PGpGJkeWYRJAcf1LYfxV1XnLJ0bWGMebnhOSd1f61K4tXCymUJ2wLEWgrQ3LFV7x
WTw/UbX2dAuEaRhOXYihQjfK7AZ0mtB1FZsM+bCPk4JgyHA/vxPF7UZ/X0wsx5Ir
bDx6tcMOCUD5UxDlsRuPNs+xvwNwVjxE+PwEK9IHheretnAbyFFg6TB8KHo06EOZ
8EPVePNn13nFmekgAXKBWdW8gtr7wmJxmwIDAQABo4IBXjCCAVowCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUgjiAYWXC0t1JvdcJTBffLOa08VowgdIGA1UdIwSByjCBx4AU
UZtbhwIJaFP8SIHNZdX6Ikfv9G6hgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkAt5Mu
43VAJZ8wNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMCIGA1UdEQEB/wQYMBaCCWxvY2FsaG9zdIIJMTI3LjAuMC4x
MA0GCSqGSIb3DQEBBQUAA4IBAQCkdoje4gsr/y0mjkr2VewFYlBWdxwZ64MF0mTI
GRBWoZQywKUm7RQDuW8r8/dP4V3/IkqH5DwkJE8cnBQdseJdHepIpTZtX2kirfwW
WNiReLHAbm1MIjAXqmGV3QL8d9/K7loozV8JxgmZOCLNE5WJ/Wm+xgFyPC/TqLqc
g1kxTHcjpvBQWJNBLhCCrw4czIOVmI/VC71FEkn2avEpC5RLv0VhqWxNhaTrbwvj
xiqgOZUbxQHNxltZzbWEnpOSSFRfXckCoZvjgB9OHu1u2+re6+TAWjUJxzueS9nT
CkOU4NzanBShJDIuhOYHhDVjAti+5At/zRbAfrR/h2I4mT7A
-----END CERTIFICATE-----
"""
http_san_cert = crypto.load_certificate(crypto.FILETYPE_PEM, http_san_cert_str)

http_san_key_str = """-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAx7ns8aPV+oC0bP9y5rmKrzOeySKtZNP4ZXh7oCLLmNPVdzwE
FIJEpdBdxC23Rypy8O1KLmdErMuYW14WUBBW/AmUgfHDVHg/gdkfKaKxhCDENbKK
7vcAPGpGJkeWYRJAcf1LYfxV1XnLJ0bWGMebnhOSd1f61K4tXCymUJ2wLEWgrQ3L
FV7xWTw/UbX2dAuEaRhOXYihQjfK7AZ0mtB1FZsM+bCPk4JgyHA/vxPF7UZ/X0ws
x5IrbDx6tcMOCUD5UxDlsRuPNs+xvwNwVjxE+PwEK9IHheretnAbyFFg6TB8KHo0
6EOZ8EPVePNn13nFmekgAXKBWdW8gtr7wmJxmwIDAQABAoIBADLH9BWwOVya0sUP
4hAFxtwq/MlZFssXFgX+rzgVAL0XxyOsz8DDKl8abmzNzta6VUUqn/zSZuSkU1WP
w46RyPwGF4s+hBJ7DHKgWMsGQao9hoW8opSFmrrBrppXA0YlFNumsj70oIJJmDyh
Rxg/0DJN+aR8axseOgxAkfzXSw7GJbhSgAedlORA4tOQMlNgpZ6NozkdAz5X5rEs
SWEsGTq/iFMi7zfxsMPHoBU/yKtgfMTD90wjwS7xytXRH8IJ2LZqdXL9pKFlM2kP
MLw1P/LZX+FLsN3YZKGGONpd3ZbjRaKY19XvnvBldEmeH4YgllM8Uv4iMBlQIwPo
qkucOyECgYEA+drqojOXDnX7DgqmQwE/gq+dbA4dnKQHfV9U7UXO/rglggce4w9H
LMvcpLGvqQ+EfsG+wQH2gCgjRAcn++aMTzt6vb2aHBlf+aFrZFGF6o8BWNuTH+Xv
aAH/j8MxfUptU25qphALrmTHccTh9XuCOXXjcCTLgY1vjLIzFF2rm4sCgYEAzKNm
B/TaTJcd/mcMK0oRqhLIjKHqb7nXLe48Y8YuLdJREaii79c5w21iRqXYC8uI0VK0
xACOe3W1DDlCyLYjh8gb8UQZvMBGhIGJ8k8xt/2loxVw+LdUpAsfQOY/nX141qB9
H3qWRL6QJtE0x8KrYlrj5j3C8Q6oZdzb+4vChDECgYAE7uufFEGxSWcRmeg6fnXQ
kMMc9RJkm2CCezVkEfLKWLx02ssjVJPlMIlEvzVmsyW8kvq/ONizIxe/Kh0B/kVa
8yOzR2EZmO6qMSO7sVhhjz1JkdxPzuxxM0jmf8YkYSnTceStLrfsU5UUi/yiKjQi
Ck/ZUQCrqC1Bq+sSRd8zjQKBgAOY27JIG449GcnfBde89ZKHwf622egc3CtyhJ08
XpNvnv0weay9m8ZTpeaYA7QxX7ZOPd80I00IRKZyWzor+XPpkJQtHkazGoo5vLMF
sVoKm17Cgbndfx9jEqmRYoziJ72eFJT5EqsirmYAR207fse70UW60DJsj4k62KMT
UdgBAoGBALTRlxgT8/Kyd5wiRUbADNG/yLXiAAG1dvhLp5BGHDMmb1MmNqHYCAAF
ndHP719l4zQJm+mOuZBSn5C7qkbrjAaPRhBOZ2Ul/x22ViZKAaQdaDMxzbHbwYmi
MY2aHHMNjP3FiD17Wplu6Ef2GKUAMF2gEb2husgH+yYeCMTGPBGB
-----END RSA PRIVATE KEY-----
"""
http_san_key = crypto.load_privatekey(crypto.FILETYPE_PEM, http_san_key_str)


class TestHandler(resource.Resource):
    isLeaf = True

    def render_GET(self, request):
        return "Hello, World!"


class HTTPTests(TestCase):
    def setUp(self):
        self.site = server.Site(TestHandler())
        self.tcp_listener = reactor.listenTCP(0, self.site,
                                              interface='127.0.0.1')

        ssl_ctx_factory = ssl.CertificateOptions(http_key, http_cert)
        self.ssl_listener = reactor.listenSSL(0, self.site,
                                              ssl_ctx_factory,
                                              interface="127.0.0.1")

    def tearDown(self):
        self.tcp_listener.stopListening()
        self.ssl_listener.stopListening()

    @defer.inlineCallbacks
    def test_http(self):
        client = http.HTTPClient()

        # try plain http request
        port = self.tcp_listener.getHost().port
        status, response, headers = yield client.get("http://127.0.0.1:%d/" % port)
        self.assertEqual(status, 200)
        self.assertEqual(response, "Hello, World!")

        # try https request
        port = self.ssl_listener.getHost().port
        client.ca_certs = [ca_cert]
        status, response, headers = yield client.get("https://127.0.0.1:%d/" % port)
        self.assertEqual(status, 200)
        self.assertEqual(response, "Hello, World!")

        # try wrong hostname (shoudl fail)
        try:
            status, response, headers = yield client.get("https://localhost:%d/" % port)
            self.fail("https client accepted incorrect hostname")
        except SSL.Error:
            pass

        # try wrong hostname with forced override (should succeed)
        status, response, headers = yield client.get("https://localhost:%d/" % port,
                                                     ssl_hostname_override='127.0.0.1')
        self.assertEqual(status, 200)
        self.assertEqual(response, "Hello, World!")

        # try wrong hostname with list of valid hostnames
        status, response, headers = yield client.get("https://localhost:%d/" % port,
                                                     ssl_hostname_override=['127.0.0.1', 'localhost'])
        self.assertEqual(status, 200)
        self.assertEqual(response, "Hello, World!")

        # try wrong ca cert (should fail)
        client.ca_certs = [wrong_ca_cert]
        try:
            status, response, headers = yield client.get("https://127.0.0.1:%d/" % port)
            self.fail("https client accepted bogus certificate!")
        except SSL.Error:
            pass

    @defer.inlineCallbacks
    def test_alternate_protocol_class(self):
        """
        Test http.request()'s protocol_class option.
        """
        class HTTPClientAlternate(twisted.web.client.HTTPPageGetter, object):
            tests = []

            def handleResponse(self, response):
                sentinel = object()
                self.tests.append((response, sentinel))
                self.factory.page(sentinel)

        port = self.ssl_listener.getHost().port
        status, response, headers = yield http.get(
            url=('https://127.0.0.1:{0:d}/'.format(port)),
            protocol_class=HTTPClientAlternate,
        )

        # If HTTPClientAlternate was used instead of a regular
        # HTTPPageGetter, server_response will be the actual response
        # body from TestHandler.render_GET. protocol_response and will
        # the sentinel the protocol told the factory was used for the
        # response.
        self.assertEqual(1, len(HTTPClientAlternate.tests))
        server_response, protocol_response = HTTPClientAlternate.tests[0]
        self.assertIs(protocol_response,
                      response)
        self.assertEqual(200, status)
        self.assertEqual([str(len(server_response))],
                         headers['content-length'])

    @defer.inlineCallbacks
    def test_disconnect(self):
        """
        Test http.request()'s disconnect option.
        """
        factory = _LineReceiverTestFactory()
        listener = reactor.listenTCP(0, factory, interface='127.0.0.1')

        expected_response = 'this is the expected response'

        class NonDisconnectingHTTPClient(protocol.Protocol, object):
            made = []
            lost = []

            def connectionMade(self):
                self.made.append(self)
                self.factory.gotStatus('HTTP/-1.0', 200, 'foo')
                self.factory.page(expected_response)
                self.factory._disconnectedDeferred.callback(None)

            def connectionLost(self, reason):
                self.lost.append(reason)

        try:
            port = listener.getHost().port
            status, response, headers = yield http.get(
                url=('http://127.0.0.1:{0:d}/'.format(port)),
                protocol_class=NonDisconnectingHTTPClient,
                disconnect=False,
            )
            self.assertEqual(1,
                             len(NonDisconnectingHTTPClient.made))
            self.assertEqual(expected_response, response)
            # http.request() must not have called disconnect.
            self.assertEqual(0,
                             len(NonDisconnectingHTTPClient.lost))
        finally:
            listener.stopListening()

        # Clean up client sockets.
        NonDisconnectingHTTPClient.made[0].transport.abortConnection()


class HTTPLogTests(TestCase):
    def setUp(self):
        self.site = server.Site(TestHandler())
        self.tcp_listener = reactor.listenTCP(0, self.site,
                                              interface='127.0.0.1')
        self.logs = []
        log.addObserver(self.logs.append)

    def tearDown(self):
        log.removeObserver(self.logs.append)
        self.tcp_listener.stopListening()

    @defer.inlineCallbacks
    def test_log_postbody(self):
        client = http.HTTPClient(logger=log)
        port = self.tcp_listener.getHost().port
        secret_canary = 'this_is_some_sensitive_data'
        status, response, headers = yield client.post(
            'http://127.0.0.1:%d' % port, {'secret': secret_canary})

        # Handle both old and new style Twisted logging events
        def format_event(e):
            return e["message"][0] if e.get("message") else e["format"] % e

        self.assertFalse(any(secret_canary in format_event(l)
                             for l in self.logs))

        client.debug = True
        status, response, headers = yield client.post(
            'http://127.0.0.1:%d' % port, {'secret': secret_canary})

        self.assertTrue(any(secret_canary in format_event(l)
                            for l in self.logs))


class HTTPSSanCertTests(TestCase):
    def setUp(self):
        self.site = server.Site(TestHandler())
        ssl_ctx_factory = ssl.CertificateOptions(http_san_key, http_san_cert)
        self.ssl_listener = reactor.listenSSL(0, self.site,
                                              ssl_ctx_factory,
                                              interface="127.0.0.1")

    def tearDown(self):
        self.ssl_listener.stopListening()

    @defer.inlineCallbacks
    def test_http_san_cert(self):
        """SAN cert should work with both 127.0.0.1 and localhost!"""
        client = http.HTTPClient()
        client.ca_certs = [ca_cert]

        port = self.ssl_listener.getHost().port
        status, response, headers = yield client.get("https://127.0.0.1:%d/" % port)
        self.assertEqual(status, 200)

        status, response, headers = yield client.get("https://localhost:%d/" % port)
        self.assertEqual(status, 200)


class TimeoutTestHandler(resource.Resource):
    def __init__(self, timeout):
        self.timeout = timeout
        self.requests = 0
        self.calls = []
        resource.Resource.__init__(self)

    isLeaf = True

    def render_GET(self, request):
        self.requests += 1

        def finish():
            self.write("Hello, World")
            self.finish()
        self.calls.append(reactor.callLater(self.timeout, finish))
        return server.NOT_DONE_YET

    def cleanup_calls(self):
        for call in self.calls:
            try:
                call.cancel()
            except AlreadyCalled:
                pass


class HTTPTimeoutTests(TestCase):
    def setUp(self):
        self.handler = TimeoutTestHandler(5)
        self.site = server.Site(self.handler)
        ssl_ctx_factory = ssl.CertificateOptions(http_key, http_cert)
        self.ssl_listener = reactor.listenSSL(0, self.site,
                                              ssl_ctx_factory,
                                              interface="127.0.0.1")

        self.listen_sock = socket.socket()
        self.listen_sock.bind(('', 0))
        self.listen_sock.listen(0)

    def tearDown(self):
        self.ssl_listener.stopListening()
        self.listen_sock.close()
        self.handler.cleanup_calls()

    @defer.inlineCallbacks
    def test_timeouts(self):
        client = http.HTTPClient([ca_cert])

        # test connect timeout (by attempting to connect to a socket that will
        # never accept())
        now = time.time()
        port = self.listen_sock.getsockname()[1]
        try:
            yield client.get("http://127.0.0.1:%d/" % port, timeout=1)
            self.fail('timeout did not occur!')
        except defer.TimeoutError:
            # make sure the connect timed out
            self.assertTrue(time.time() - now < 2)

        # test http factory timeout
        now = time.time()
        port = self.ssl_listener.getHost().port
        try:
            yield client.get("https://127.0.0.1:%d/" % port, timeout=1)
            self.fail('timeout did not occur!')
        except defer.TimeoutError:
            # make sure it connected
            self.assertEqual(self.handler.requests, 1)
            # but that it timed out after
            self.assertTrue(time.time() - now < 2)


class _LineReceiverTestProtocol(LineReceiver):
    def __init__(self):
        self.lines = defer.DeferredQueue()

    def connectionMade(self):
        self.factory.deferred.callback(self)

    def connectionLost(self, reason):
        for d in self.lines.waiting:
            d.errback(reason)

    def lineReceived(self, line):
        self.lines.put(line)


class _LineReceiverTestFactory(protocol.ServerFactory):
    protocol = _LineReceiverTestProtocol

    def __init__(self):
        self.deferred = defer.Deferred()


class TestProxyProtocol(TestCase):
    def setUp(self):
        self.factory = _LineReceiverTestFactory()
        self.listener = reactor.listenTCP(0, self.factory, interface='127.0.0.1')
        self.srv_port = self.listener.getHost().port

        # no ca certs == no validation, so we can spoof whatever
        # hostname we want
        self.client = http.HTTPClient(
            [], http_proxy_host='127.0.0.1', http_proxy_port=self.srv_port,
            user_agent='DuoUnitTestClient')
        self.ssl_ctx_factory = ssl.CertificateOptions(http_key, http_cert)

    def tearDown(self):
        self.listener.stopListening()

    def delay(self, secs):
        d = defer.Deferred()
        reactor.callLater(secs, d.callback, None)
        return d

    @defer.inlineCallbacks
    def test_basic_proxy(self):
        d = self.client.get('https://foo.invalid/hello')

        proto = yield self.factory.deferred
        line = yield proto.lines.get()
        self.assertEqual(line, 'CONNECT foo.invalid:443 HTTP/1.0')
        line = yield proto.lines.get()
        self.assertEqual(line, 'User-Agent: DuoUnitTestClient')
        line = yield proto.lines.get()
        self.assertEqual(line, '')
        proto.sendLine('HTTP/1.0 200 Connection Established')
        proto.sendLine('Proxy-Agent: DuoUnitTestServer')
        proto.sendLine('')
        proto.transport.startTLS(self.ssl_ctx_factory)

        line = yield proto.lines.get()
        self.assertEqual(line, 'GET /hello HTTP/1.0')
        while (line != ''):
            line = yield proto.lines.get()

        payload = 'Hello, World!'
        proto.sendLine('HTTP/1.0 200 OK')
        proto.sendLine('Content-Length: %d' % len(payload))
        proto.sendLine('')
        proto.transport.write(payload)

        status, body, headers = yield d

        self.assertEqual(status, 200)
        self.assertEqual(body, payload)

    @defer.inlineCallbacks
    def test_proxy_timeout(self):
        d = self.client.get('https://foo.invalid/hello', timeout=1)

        proto = yield self.factory.deferred
        line = yield proto.lines.get()
        self.assertEqual(line, 'CONNECT foo.invalid:443 HTTP/1.0')
        line = yield proto.lines.get()
        self.assertEqual(line, 'User-Agent: DuoUnitTestClient')
        line = yield proto.lines.get()
        self.assertEqual(line, '')
        proto.sendLine('HTTP/1.0 200 Connection Established')

        try:
            status, body, headers = yield d
            self.fail('Did not get a timeout exception!')
        except ConnectionAborted:
            pass

    @defer.inlineCallbacks
    def test_cumulative_timeout(self):
        d = self.client.get('https://foo.invalid/hello', timeout=2.5)

        proto = yield self.factory.deferred
        line = yield proto.lines.get()
        self.assertEqual(line, 'CONNECT foo.invalid:443 HTTP/1.0')
        line = yield proto.lines.get()
        self.assertEqual(line, 'User-Agent: DuoUnitTestClient')
        line = yield proto.lines.get()
        self.assertEqual(line, '')
        proto.sendLine('HTTP/1.0 200 Connection Established')
        yield self.delay(1)
        proto.sendLine('Proxy-Agent: DuoUnitTestServer')
        proto.sendLine('')
        proto.transport.startTLS(self.ssl_ctx_factory)

        line = yield proto.lines.get()
        self.assertEqual(line, 'GET /hello HTTP/1.0')
        while (line != ''):
            line = yield proto.lines.get()

        yield self.delay(2)
        try:
            payload = 'Hello, World!'
            proto.sendLine('HTTP/1.0 200 OK')
            proto.sendLine('Content-Length: %d' % len(payload))
            proto.sendLine('')
            proto.transport.write(payload)
        except Exception:
            pass

        try:
            status, body, headers = yield d
            self.fail('Did not get a timeout exception!')
        except defer.TimeoutError:
            pass


class TestProxyProtocolConnectFail(TestCase):
    def setUp(self):
        # get an ephemeral port number that will refuse our connection attempt
        # by opening and then immediately closing a listen socket
        listen_sock = socket.socket()
        listen_sock.bind(('', 0))
        listen_sock.listen(0)
        self.srv_port = listen_sock.getsockname()[1]
        listen_sock.close()

        self.client = http.HTTPClient(
            [], http_proxy_host='127.0.0.1', http_proxy_port=self.srv_port,
            user_agent='DuoUnitTestClient')
        self.ssl_ctx_factory = ssl.CertificateOptions(http_key, http_cert)

    @defer.inlineCallbacks
    def test_proxy_connect_fail(self):
        timeout_dc = None
        try:
            d = self.client.get("https://foo.invalid/hello")
            timeout_dc = reactor.callLater(
                2, d.errback, FailTest('did not propagate connect error'))
            yield d
        except ConnectionRefusedError:
            pass
        finally:
            if timeout_dc:
                try:
                    timeout_dc.cancel()
                except AlreadyCalled:
                    pass


class SSL2ContextFactory(ssl.ContextFactory):
    def __init__(self, key, cert):
        self.key = key
        self.cert = cert

    def getContext(self):
        ctx = SSL.Context(SSL.SSLv2_METHOD)
        ctx.use_certificate(self.cert)
        ctx.use_privatekey(self.key)
        return ctx


class TestSSL2Protocol(TestCase):
    def setUp(self):
        self.site = server.Site(TestHandler())
        ssl_ctx_factory = SSL2ContextFactory(http_key, http_cert)
        try:
            self.ssl_listener = reactor.listenSSL(
                0, self.site, ssl_ctx_factory, interface='127.0.0.1')
        except ValueError:
            self.ssl_listener = None

    def tearDown(self):
        if self.ssl_listener:
            self.ssl_listener.stopListening()

    @defer.inlineCallbacks
    def test_ssl2(self):
        """ Ensure that connections to an SSLv2 server will fail! """
        if self.ssl_listener:
            client = http.HTTPClient()
            port = self.ssl_listener.getHost().port
            client.ca_certs = [ca_cert]
            try:
                status, response, headers = yield client.get('https://127.0.0.1:%d/' % port)
                self.fail('https client accepted SSLv2 handshake!')
            except SSL.Error as e:
                expected_args = ([('SSL routines', 'SSL23_READ', 'ssl handshake failure')],)
                self.assertEqual(e.args, expected_args)


if __name__ == '__main__':
    import unittest
    unittest.main()
