from twisted.trial.unittest import TestCase

from duoauthproxy.lib import log


class MockLog(object):
    def __init__(self):
        self.results = []

    def base(self, *args, **kwargs):
        args = [
            str(arg)
            for arg in args
        ]
        kwargs = dict(
            (str(k), str(v))
            for k, v in kwargs.items()
        )

        args_string = " ".join(args)
        kwargs_string = " ".join([
            "{0}={1}".format(k, v)
            for k, v in kwargs.items()
        ])

        return " ".join([
            s for s in args_string, kwargs_string if s
        ])

    def msg(self, *args, **kwargs):
        self.results.append(self.base("MSG", *args, **kwargs))

    def err(self, *args, **kwargs):
        self.results.append(self.base("ERR", *args, **kwargs))


class TestLog(TestCase):

    def setUp(self):
        self.mock_logger = MockLog()
        self.logger = log.BufferingLogger(logger=self.mock_logger)

    def test_functions_exist(self):
        function_names = [
            'ready',
            'msg',
            'err',
            'config',
            'auth',
            'auth_standard',
        ]

        for function_name in function_names:
            self.assertTrue(hasattr(log, function_name))

    def test_log_msg_basic(self):
        self.logger.msg("BASIC")
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG BASIC",
        ]

        self.assertEquals(results, expected)

    def test_log_err_basic(self):
        self.logger.err("BASIC")
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "ERR BASIC",
        ]

        self.assertEquals(results, expected)

    def test_log_mixed_basic(self):
        self.logger.msg("1")
        self.logger.err("2")
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG 1",
            "ERR 2",
        ]

        self.assertEquals(results, expected)

    def test_log_after_ready(self):
        self.logger.ready()
        self.logger.msg("1")
        self.logger.msg("2")

        results = self.mock_logger.results
        expected = [
            "MSG 1",
            "MSG 2",
        ]

        self.assertEquals(results, expected)

    def test_log_before_ready(self):
        self.logger.msg("1")
        self.logger.msg("2")
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG 1",
            "MSG 2",
        ]

        self.assertEquals(results, expected)

    def test_log_kwargs(self):
        self.logger.msg(key="value")
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG key=value",
        ]

        self.assertEquals(results, expected)

    def test_log_config_basic(self):
        self.logger.config({
            "key": "value",
        })
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG {'key': 'value'}",
        ]

        self.assertEquals(results, expected)

    def test_log_config_scrub(self):
        self.logger.config(
            {
                "key": "value",
            },
            scrub=lambda x: x == "key"
        )
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG {'key': '*****'}",
        ]

        self.assertEquals(results, expected)

    def test_log_config_scrub_skey(self):
        skey = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMN"
        self.logger.config(
            {
                "skey": skey,
            },
            scrub=lambda x: True
        )
        self.logger.ready()

        results = self.mock_logger.results
        expected = [
            "MSG {'skey': '*****[40]'}",
        ]

        self.assertEquals(results, expected)

    def test_log_auth_before_ready(self):
        expected = [
            "'msg': 'BASIC'",
            "'timestamp':",
            "msg_type=AUTH",
        ]

        self.logger.auth(msg="BASIC")
        self.logger.ready()

        results = self.mock_logger.results
        for entry in expected:
            self.assertIn(entry, results[0])

    def test_log_auth_after_ready(self):
        expected = [
            "'msg': 'BASIC'",
            "'timestamp':",
            "msg_type=AUTH",
        ]

        self.logger.ready()
        self.logger.auth(msg="BASIC")

        results = self.mock_logger.results
        for entry in expected:
            self.assertIn(entry, results[0])

    def test_log_auth_standard(self):
        expected = [
            "'msg': 'message'",
            "'username': 'testuser'",
            "'auth_stage': '{0}'".format(log.AUTH_SECONDARY),
            "'status': '{0}'".format(log.AUTH_REJECT),
            "'client_ip': '186.10.56.12'",
            "'server_section': 'radius_server_auto'",
            "'timestamp':",
        ]

        self.logger.ready()
        self.logger.auth_standard(
            "message",
            "testuser",
            log.AUTH_SECONDARY,
            log.AUTH_REJECT,
            "186.10.56.12",
            "radius_server_auto"
        )

        results = self.mock_logger.results
        for entry in expected:
            self.assertIn(entry, results[0])


if __name__ == '__main__':
    import unittest
    unittest.main()
