#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
import os
import tempfile

from twisted.internet import defer, reactor
from twisted.python import log

from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import ldaperrors

from duoauthproxy.modules import ad_client
from duoauthproxy.lib.config_provider import ConfigDict, ConfigError
from duoauthproxy.lib.connectivity.connect import ldap_operations
from duoauthproxy.lib import protect
from duoauthproxy.lib import ldap
from duoauthproxy.lib import util

try:
    import sspi
    import sspicon
except ImportError:
    sspi = None
    sspicon = None

from base import ReopenableTemporaryFile
from base import TestCase, ClientProtectedConfigParsingTestCase
from mock_ldap_server import (
    ca_cert_pem,
    make_server_factory, MoreADLikeLDAPServer, LDAPServerFactory)


class TestModuleADClientBase(object):
    server_factory = LDAPServerFactory
    listen_ldaps = False

    @defer.inlineCallbacks
    def setUp(self):
        self.instance = None

        self.server_port = 10389
        (self.server_factory,
         self.server_listener) = yield make_server_factory(
            self.server_port, self.server_factory, ssl=self.listen_ldaps)

        # Real AD is smart enough to translate (objectCategory=user).
        # Here, must use the user with the right objectCategory in
        # LDIF_DATA.
        self.username = 'jane'
        self.password = 'password'
        self.client_ip = '12.34.56.78'
        ad_client._ADAuthClientFactory.cleanup_connections = True

    @defer.inlineCallbacks
    def tearDown(self):
        if self.instance:
            self.instance.stopService()
        yield self.server_listener.stopListening()

    def make_instance(self, config_items):
        self.config_items = {
            'host': 'localhost',
            'port': str(self.server_port),
            'service_account_username': 'adminer',
            'service_account_password': 'password!',
            'search_dn': 'DC=bongos,DC=test',
            # MoreADLikeLDAPServer isn't AD-like enough for <ROOT>.
            'bind_dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
            # Nor NTLM.
            'auth_type': 'plain',
        }
        if config_items:
            self.config_items.update(config_items)
        config = ConfigDict(self.config_items)
        self.instance = ad_client.Module(config)


class TestModuleADClientConfigFlags(TestModuleADClientBase, TestCase):

    def setUp(self):
        super(TestModuleADClientConfigFlags, self).setUp()
        self.is_password_logged = False
        self.original_log_msg = log.msg

    @defer.inlineCallbacks
    def tearDown(self):
        yield super(TestModuleADClientConfigFlags, self).tearDown()
        log.msg = self.original_log_msg

    @defer.inlineCallbacks
    def test_plaintext_password_logged(self):
        log.msg = self.find_password_in_log_msg
        self.enable_is_logging_insecure('True')
        yield self.instance.authenticate(
            self.username,
            self.password,
            self.client_ip
        )
        self.assertTrue(self.is_password_logged)

    @defer.inlineCallbacks
    def test_plaintext_password_not_logged(self):
        log.msg = self.find_password_in_log_msg
        self.enable_is_logging_insecure('False')
        yield self.instance.authenticate(
            self.username,
            self.password,
            self.client_ip
        )
        self.assertFalse(self.is_password_logged)

    def find_password_in_log_msg(self, msg):
        if self.password in msg:
            self.is_password_logged = True

    def enable_is_logging_insecure(self, is_enabled):
        config_items = {
            'debug': 'True',
            'is_logging_insecure': is_enabled,
        }
        self.make_instance(config_items)


class TestModuleADClient(TestModuleADClientBase, TestCase):
    @defer.inlineCallbacks
    def test_ldap_proxy_fallback_hosts(self):
        self.make_instance({
            'host': 'badhostname', 'host_2': 'localhost'})
        client = yield self.instance.ldap_proxy()
        # Clean up our connection to the Mock ADServer
        client.transport.abortConnection()
        self.assertEqual(client.factory.bind_dn, 'CN=adminer adminer,CN=Users,DC=bongos,DC=test')

    @defer.inlineCallbacks
    def test_ldap_proxy_fallback_hosts_fail(self):
        self.make_instance({
            'host': 'badhostname', 'host_2': 'badhostnametoo'})
        with self.assertRaises(ldap.client.ADClientError):
            yield self.instance.ldap_proxy()

    @defer.inlineCallbacks
    def test_authenticate_right_password(self):
        self.make_instance({})
        res = yield self.instance.authenticate(self.username,
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_authenticate_openldap_success(self):
        self.make_instance({
            'username_attribute': 'cn'
        })
        res = yield self.instance.authenticate('OpenLDAPer',
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success, "Login should succeed, got {0}".format(res))

    @defer.inlineCallbacks
    def test_authenticate_username_split_for_sAMAccountName(self):
        self.make_instance({})
        res = yield self.instance.authenticate('domain\\' + self.username,
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_authenticate_username_not_split_if_not_sAMAccountName(self):
        self.make_instance({
            'username_attribute': 'givenName',
        })
        res = yield self.instance.authenticate('domain\\' + 'Jane',
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_custom_username_attribute_one_match_suceeds(self):
        self.make_instance({
            'username_attribute': 'cn',
        })
        res = yield self.instance.authenticate('Jane Blackwood',
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_custom_username_attribute_no_matches_fails(self):
        self.make_instance({
            'username_attribute': 'cn',
        })
        res = yield self.instance.authenticate('No Such CN',
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_custom_username_attribute_multiple_matches_fails(self):
        self.make_instance({
            'username_attribute': 'objectClass',
        })
        res = yield self.instance.authenticate('person',
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_authenticate_wrong_password(self):
        self.make_instance({})
        res = yield self.instance.authenticate(self.username,
                                               self.password + 'nope',
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_ldap_filter(self):
        self.make_instance({
            'ldap_filter': '(&(username=*foo*)(primaryGroupID=513))',
        })
        res = yield self.instance.authenticate(self.username,
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

        self.make_instance({
            'ldap_filter': '(primaryGroupID=513)',
        })
        res = yield self.instance.authenticate(self.username,
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def assert_security_group_auth_result(self, member, username):
        """Assert that if Jane is a member of the security group she can auth
        with the given username"""
        if member:
            self.make_instance({
                'security_group_dn': 'CN=Users,DC=bongos,DC=test',
            })
        else:
            self.make_instance({
                'security_group_dn': 'CN=No Janes Allowed,DC=bongos,DC=test',
            })

        res = yield self.instance.authenticate(username,
                                               self.password,
                                               self.client_ip)

        self.assertEqual(res.success, member)

    @defer.inlineCallbacks
    def test_security_group_dn_permitted(self):
        yield self.assert_security_group_auth_result(member=True, username=self.username)

    @defer.inlineCallbacks
    def test_security_group_dn_unpermitted(self):
        yield self.assert_security_group_auth_result(member=False, username=self.username)

    @defer.inlineCallbacks
    def test_security_filtering_permitted_upn(self):
        yield self.assert_security_group_auth_result(member=True, username='jane@bongos.test')

    @defer.inlineCallbacks
    def test_security_filtering_permitted_netbios(self):
        yield self.assert_security_group_auth_result(member=True, username='domain\\' + 'Jane')

    @defer.inlineCallbacks
    def test_security_filtering_unpermitted_upn(self):
        yield self.assert_security_group_auth_result(member=False, username='jane@bongos.test')

    @defer.inlineCallbacks
    def test_security_filtering_unpermitted_netbios(self):
        yield self.assert_security_group_auth_result(member=False, username='domain\\' + 'Jane')

    @defer.inlineCallbacks
    def test_backing_response_timeout(self):
        self.server_factory.response_delay = 5
        self.make_instance({'timeout': '3'})
        res = yield self.instance.authenticate(self.username,
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)
        # Wait for cleanup of artificially delayed requests.
        yield self.server_factory.delayed

    @defer.inlineCallbacks
    def test_ntlm_domain_split_from_username(self):
        self.make_instance({})
        res = yield self.instance.authenticate('subdomain\\' + 'dupe',
                                               'password2',
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_domain_discovery_with_good_principal_name(self):
        self.make_instance({'domain_discovery': 'True'})
        res = yield self.instance.authenticate('jane',
                                               'password',
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_domain_discovery_with_bad_principal_name(self):
        self.make_instance({'domain_discovery': 'True'})
        res = yield self.instance.authenticate('pskinner',
                                               'password',
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_ntlm_domain_split_from_username_fail(self):
        self.make_instance({})
        res = yield self.instance.authenticate('dupe',
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_upn_attribute_success(self):
        self.make_instance({})
        res = yield self.instance.authenticate('dupe@bongos.test',
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_upn_attribute_incorrect_attribute(self):
        self.make_instance({
            'at_attribute': 'not_a_good_attribute_name',
        })
        res = yield self.instance.authenticate('jane@bongos.test',
                                               self.password,
                                               self.client_ip)
        self.assertFalse(res.success)

    @defer.inlineCallbacks
    def test_empty_password(self):
        self.make_instance({})
        res = yield self.instance.authenticate('jane@bongos.test',
                                               '',
                                               self.client_ip)
        self.assertEqual(
            0, self.server_factory.bind_requests_received,
            'Empty password should be rejected, not sent to DC.')
        self.assertFalse(res.success)


class TestModuleADClientSSPI(TestModuleADClientBase, TestCase):
    @defer.inlineCallbacks
    def test_authenticate_sspi(self):
        """Have Windows negotiate with itself using the current
        process creds, until it succeeds..."""

        # Of course, this can only work on Windows...
        if not sspi:
            return

        self.make_instance({
            'service_account_username': '',
            'service_account_password': '',
            'auth_type': 'sspi'
        })

        # these return codes mean we need to continue the handshake
        sspi_continue = set([
            sspicon.SEC_I_CONTINUE_NEEDED,
            sspicon.SEC_I_COMPLETE_AND_CONTINUE
        ])
        # any return code not in this set should be considered an error
        sspi_ok = sspi_continue.union([
            sspicon.SEC_E_OK,
            sspicon.SEC_I_COMPLETE_NEEDED
        ])

        sa = sspi.ServerAuth('Negotiate')

        def handle_bind_2(protocol, request, controls, reply):
            self.assertTrue(request.sasl)
            self.assertIs(type(request.auth), tuple)
            protocol.boundUser = "don't look at me"

            # Ignore all authentication on the second bind and
            # immediately return a success (ugly, but it should
            # work)
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=str(request.dn)
            )
            return response

        def handle_bind_1(protocol, request, controls, reply):
            self.assertTrue(request.sasl)
            self.assertIs(type(request.auth), tuple)
            protocol.boundUser = "don't look at me"

            err, out_buf = sa.authorize(request.auth[1])
            if err in sspi_continue:
                ssc = pureldap.LDAPBindResponse_serverSaslCreds(
                    value=out_buf[0].Buffer
                )
                response = pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPSaslBindInProgress.resultCode,
                    matchedDN=str(request.dn),
                    serverSaslCreds=ssc
                )
            elif err in sspi_ok:
                response = pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    matchedDN=str(request.dn)
                )
                # AcceptSecurityContext requires your credentials to actually
                # exist as a real account on the windows system, so there's
                # no way we can call it legitimately for the 2nd bind.
                # Instead, we'll use the routine above that ignores credentials
                # and considers every bind a success.
                self.server_factory.handle_LDAPBindRequest = handle_bind_2
            else:
                response = pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode,
                    matchedDN=str(request.dn),
                    errorMessage='SSPI ERROR: %d' % err
                )
            return response

        self.server_factory.handle_LDAPBindRequest = handle_bind_1

        res = yield self.instance.authenticate(self.username,
                                               self.password,
                                               self.client_ip)
        self.assertTrue(res.success)


class DisconnectingLDAPServer(MoreADLikeLDAPServer):
    """LDAP server protocol that disconnects immediately upon receiving
    a search request"""

    def _cbSearchGotBase(self, base, dn, request, reply):
        self.transport.abortConnection()


class DisconnectingLDAPServerFactory(LDAPServerFactory):
    protocol = DisconnectingLDAPServer


class TestModuleADClientDisconnect(TestModuleADClientBase, TestCase):
    server_factory = DisconnectingLDAPServerFactory

    @defer.inlineCallbacks
    def test_authenticate_disconnect(self):
        """Test for T4408 - make sure we propagate connectionLost
        errors correctly"""
        self.make_instance({})
        d = self.instance.authenticate(self.username,
                                       self.password,
                                       self.client_ip)
        d2 = defer.Deferred()

        def timeout():
            if not d2.called:
                d2.callback(True)
        dc = reactor.callLater(5, timeout)

        def auth_callback(ignored):
            if not d2.called:
                d2.callback(False)
            if dc.active:
                dc.cancel()
        d.addBoth(auth_callback)

        timed_out = yield d2
        self.assertFalse(timed_out)


class TestModuleADClientLDAPS(TestModuleADClientBase, TestCase):
    """
    Test that ad_client can connect to an LDAPS server.

    No need to re-test everything TestModuleADClient does. Just test
    one case for each set of client connection parameters.
    """
    listen_ldaps = True

    @defer.inlineCallbacks
    def test_ssl_verify_hostname_enabled(self):
        fh = ReopenableTemporaryFile(ca_cert_pem)
        try:
            self.make_instance({
                'transport': 'ldaps',
                'ssl_ca_certs_file': fh.name,
                'ssl_verify_hostname': 'True',
            })
            res = yield self.instance.authenticate(self.username,
                                                   self.password,
                                                   self.client_ip)
            self.assertTrue(res.success)
        finally:
            if fh.close_called and os.path.isfile(fh.name):
                os.remove(fh.name)

    @defer.inlineCallbacks
    def test_ssl_verify_hostname_disabled(self):
        fh = ReopenableTemporaryFile(ca_cert_pem)
        try:
            self.make_instance({
                'transport': 'ldaps',
                'ssl_ca_certs_file': fh.name,
                'ssl_verify_hostname': 'False',
            })
            res = yield self.instance.authenticate(self.username,
                                                   self.password,
                                                   self.client_ip)
            self.assertTrue(res.success)
        finally:
            if fh.close_called and os.path.isfile(fh.name):
                os.remove(fh.name)


class TestModuleADClientAddressConfigParsing(TestCase):
    def _make_instance(self, more_config_items):
        config_items = {
            'service_account_username': 'adminer',
            'service_account_password': 'password!',
            'search_dn': 'CN=Users,DC=bongos,DC=test',
            # MoreADLikeLDAPServer isn't AD-like enough for <ROOT>.
            'bind_dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
            # Nor NTLM.
            'auth_type': 'plain',
        }
        config_items.update(more_config_items)
        config = ConfigDict(config_items)
        return ad_client.Module(config)

    def test_host(self):
        instance = self._make_instance({
            'host': '1.2.3.5',
        })
        self.assertEqual(instance.hosts, [
            '1.2.3.5',
        ])

    def test_host_1(self):
        instance = self._make_instance({
            'host_1': '1.2.3.5',
        })
        self.assertEqual(instance.hosts, [
            '1.2.3.5',
        ])

    def test_host_and_host_2(self):
        instance = self._make_instance({
            'host': '1.2.3.4',
            'host_2': '1.2.3.5',
        })
        self.assertEqual(instance.hosts, [
            '1.2.3.4',
            '1.2.3.5',
        ])

    def test_host_1_and_host_2(self):
        instance = self._make_instance({
            'host_1': '1.2.3.4',
            'host_2': '1.2.3.5',
        })
        self.assertEqual(instance.hosts, [
            '1.2.3.4',
            '1.2.3.5',
        ])

    def test_host_2(self):
        with self.assertRaisesRegexp(ConfigError, "Missing.*host'"):
            self._make_instance({
                'host_2': '1.2.3.5',
            })

    def test_ssl_ca_certs_file_missing(self):
        """
        Error when the specified file does not exist.
        """
        with self.assertRaisesRegexp(ConfigError,
                                     'Could not load ssl ca certificates from'):
            self._make_instance({
                'host': '1.2.3.4',
                'ssl_ca_certs_file': 'foo.bar',
            })

    def test_ssl_ca_certs_not_loadable(self):
        """
        Attempt to load a file with valid BEGIN and END CERTIFICATE
        but no valid certificate between them.
        """
        path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            '..', '..', 'duoauthproxy', 'lib', 'ssl_verify.py')
        with self.assertRaisesRegexp(ConfigError,
                                     'Could not load ssl ca certificates from'):
            self._make_instance({
                'host': '1.2.3.4',
                'ssl_ca_certs_file': path,
            })

    def test_ssl_ca_certs_loadable(self):
        """
        Test successful CA certificates load.
        """
        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmp.write("""
-----BEGIN CERTIFICATE-----
MIIE2DCCBEGgAwIBAgIEN0rSQzANBgkqhkiG9w0BAQUFADCBwzELMAkGA1UE
BhMCVVMxFDASBgNVBAoTC0VudHJ1c3QubmV0MTswOQYDVQQLEzJ3d3cuZW50
cnVzdC5uZXQvQ1BTIGluY29ycC4gYnkgcmVmLiAobGltaXRzIGxpYWIuKTEl
MCMGA1UECxMcKGMpIDE5OTkgRW50cnVzdC5uZXQgTGltaXRlZDE6MDgGA1UE
AxMxRW50cnVzdC5uZXQgU2VjdXJlIFNlcnZlciBDZXJ0aWZpY2F0aW9uIEF1
dGhvcml0eTAeFw05OTA1MjUxNjA5NDBaFw0xOTA1MjUxNjM5NDBaMIHDMQsw
CQYDVQQGEwJVUzEUMBIGA1UEChMLRW50cnVzdC5uZXQxOzA5BgNVBAsTMnd3
dy5lbnRydXN0Lm5ldC9DUFMgaW5jb3JwLiBieSByZWYuIChsaW1pdHMgbGlh
Yi4pMSUwIwYDVQQLExwoYykgMTk5OSBFbnRydXN0Lm5ldCBMaW1pdGVkMTow
OAYDVQQDEzFFbnRydXN0Lm5ldCBTZWN1cmUgU2VydmVyIENlcnRpZmljYXRp
b24gQXV0aG9yaXR5MIGdMA0GCSqGSIb3DQEBAQUAA4GLADCBhwKBgQDNKIM0
VBuJ8w+vN5Ex/68xYMmo6LIQaO2f55M28Qpku0f1BBc/I0dNxScZgSYMVHIN
iC3ZH5oSn7yzcdOAGT9HZnuMNSjSuQrfJNqc1lB5gXpa0zf3wkrYKZImZNHk
mGw6AIr1NJtl+O3jEP/9uElY3KDegjlrgbEWGWG5VLbmQwIBA6OCAdcwggHT
MBEGCWCGSAGG+EIBAQQEAwIABzCCARkGA1UdHwSCARAwggEMMIHeoIHboIHY
pIHVMIHSMQswCQYDVQQGEwJVUzEUMBIGA1UEChMLRW50cnVzdC5uZXQxOzA5
BgNVBAsTMnd3dy5lbnRydXN0Lm5ldC9DUFMgaW5jb3JwLiBieSByZWYuIChs
aW1pdHMgbGlhYi4pMSUwIwYDVQQLExwoYykgMTk5OSBFbnRydXN0Lm5ldCBM
aW1pdGVkMTowOAYDVQQDEzFFbnRydXN0Lm5ldCBTZWN1cmUgU2VydmVyIENl
cnRpZmljYXRpb24gQXV0aG9yaXR5MQ0wCwYDVQQDEwRDUkwxMCmgJ6AlhiNo
dHRwOi8vd3d3LmVudHJ1c3QubmV0L0NSTC9uZXQxLmNybDArBgNVHRAEJDAi
gA8xOTk5MDUyNTE2MDk0MFqBDzIwMTkwNTI1MTYwOTQwWjALBgNVHQ8EBAMC
AQYwHwYDVR0jBBgwFoAU8BdiE1U9s/8KAGv7UISX8+1i0BowHQYDVR0OBBYE
FPAXYhNVPbP/CgBr+1CEl/PtYtAaMAwGA1UdEwQFMAMBAf8wGQYJKoZIhvZ9
B0EABAwwChsEVjQuMAMCBJAwDQYJKoZIhvcNAQEFBQADgYEAkNwwAvpkdMKn
CqV8IY00F6j7Rw7/JXyNEwr75Ji174z4xRAN95K+8cPV1ZVqBLssziY2Zcgx
xufuP+NXdYR6Ee9GTxj005i7qIcyunL2POI9n9cd2cNgQ4xYDiKWL2KjLB+6
rQXvqzJ4h6BUcxm1XAX5Uj5tLUUL9wqT6u0G+bI=
-----END CERTIFICATE-----
""")
            tmp.flush()
            tmp.close()  # hack for Windows and NamedTemporaryFile.
            self._make_instance({
                'host': '1.2.3.4',
                'ssl_ca_certs_file': tmp.name,
            })
        finally:
            os.remove(tmp.name)


class TestAdClientProtectedConfigParsing(ClientProtectedConfigParsingTestCase):

    def test_plain_keys_use_protected_api(self):
        instance = self._make_instance({
            'service_account_password': 'password!',
        }, ad_client)
        self.assertEqual(instance.factory_kwargs['service_account_password'], 'password!')

    def test_protected_keys(self):
        if protect.PROTECT_ENABLED:
            protected = protect.protect('password!')
            instance = self._make_instance({
                'service_account_password_protected': protected,
            }, ad_client)
            self.assertEqual(instance.factory_kwargs['service_account_password'], 'password!')


class TestLdapOperations(TestModuleADClientBase, TestCase):

    @defer.inlineCallbacks
    def test_ldap_search_has_results(self):
        self.make_instance({})
        client = yield self.instance.ldap_proxy()

        search_dn = client.factory.search_dn
        filter_object = util.parse_ldap_filter("(objectclass=organizationalPerson)")

        search_results = yield ldap_operations.ldap_search_has_results(client, search_dn, filter_object)
        client.transport.abortConnection()

        self.assertTrue(search_results.is_successful(), "Search should return results")

    @defer.inlineCallbacks
    def test_ldap_search_no_results(self):
        self.make_instance({})
        client = yield self.instance.ldap_proxy()

        search_dn = client.factory.search_dn
        filter_object = util.parse_ldap_filter("(objectclass=aaron)")

        search_results = yield ldap_operations.ldap_search_has_results(client, search_dn, filter_object)
        client.transport.abortConnection()

        self.assertFalse(search_results.is_successful(), "Search should not return results")

    @defer.inlineCallbacks
    def test_bind_service_user_good(self):
        self.make_instance({})
        client = yield self.instance.ldap_proxy()

        could_bind_service_user = yield ldap_operations.can_bind_service_account(client)
        client.transport.abortConnection()

        self.assertTrue(could_bind_service_user.is_successful())

    @defer.inlineCallbacks
    def test_bind_service_user_bad_password(self):
        self.make_instance(
                {
                    'service_account_password': 'wrong_password'
                })
        client = yield self.instance.ldap_proxy()

        could_bind_service_user = yield ldap_operations.can_bind_service_account(client)
        client.transport.abortConnection()

        self.assertFalse(could_bind_service_user.is_successful())


if __name__ == '__main__':
    import unittest
    unittest.main()
