#
# Copyright (c) 2015 Duo Security
# All Rights Reserved
#

import functools
import netaddr

import twisted
from twisted.internet import reactor, defer, protocol
from twisted.protocols.basic import LineReceiver
from twisted.web import server, resource, http

import connect_proxy

from base import TestCase
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.modules import http_proxy


class TestHandler(resource.Resource):
    isLeaf = True

    def render_GET(self, request):
        return "Hello, World!"


class BaseHttpProxyTestCases(TestCase):

    client_ips = []

    @defer.inlineCallbacks
    def setUp(self):
        # create a server to proxy through to
        self.site = server.Site(TestHandler())
        self.tcp_listener = reactor.listenTCP(0, self.site,
                                              interface='127.0.0.1')

        # create the HTTP connect proxy
        self.factory = http.HTTPFactory()
        self.factory.protocol = functools.partial(connect_proxy.ConnectProxy, host='127.0.0.1', client_ips=self.client_ips)
        self.proxy_listener = reactor.listenTCP(0, self.factory, interface='127.0.0.1')

        # create the client to connect to the HTTP connect proxy
        self.client_factory = _LineReceiverTestFactory()
        self.client_connector = reactor.connectTCP("127.0.0.1", self.proxy_listener.getHost().port, self.client_factory)
        self.proto = yield self.client_factory.deferred

    @defer.inlineCallbacks
    def tearDown(self):
        self.client_connector.disconnect()
        yield self.proxy_listener.stopListening()
        yield self.tcp_listener.stopListening()


class HttpProxyTestCases(BaseHttpProxyTestCases):

    @defer.inlineCallbacks
    def test_http_proxy_reject_get_without_connect(self):
        # send GET request through proxy without first sending CONNECT
        self.proto.sendLine('GET http://127.0.0.1:%d/ HTTP/1.1' % self.tcp_listener.getHost().port)
        self.proto.sendLine('')

        # expect 403 response
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.1 405 Bad Proxy Request')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Transfer-Encoding: chunked')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Type: text/html')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')
        data = yield self.proto.data.get()
        self.assertEqual(data, '35\r\nAttempted connection creation without initial CONNECT\r\n0\r\n\r\n')

    @defer.inlineCallbacks
    def test_http_proxy_reject_connect_to_unapproved_host(self):
        # send CONNECT request from unapproved host
        self.proto.sendLine('CONNECT example.org:%d HTTP/1.0' % self.tcp_listener.getHost().port)
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 403 response
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 403 Bad Proxy Request')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Type: text/html')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')
        data = yield self.proto.data.get()
        self.assertEqual(data, 'Attempted connection creation to non-approved host: example.org')

    @defer.inlineCallbacks
    def test_http_proxy_connect_and_get(self):
        # send CONNECT request
        self.proto.sendLine('CONNECT 127.0.0.1:%d HTTP/1.0' % self.tcp_listener.getHost().port)
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 200 OK
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 200 CONNECT OK')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'X-Connected-Ip: 127.0.0.1')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Length: 0')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')

        # send GET request through proxy
        self.proto.sendLine('GET http://127.0.0.1:%d/ HTTP/1.1' % self.tcp_listener.getHost().port)
        self.proto.sendLine('')

        # expect 200 OK and proper payload
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.1 200 OK')
        line = yield self.proto.lines.get()
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Length: 13')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Type: text/html')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Server: TwistedWeb/{0}'.format(twisted.__version__))
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')
        data = yield self.proto.data.get()
        self.assertEqual(data, 'Hello, World!')

    @defer.inlineCallbacks
    def test_http_proxy_reject_http_request(self):
        # send CONNECT request
        self.proto.sendLine('CONNECT http://127.0.0.1:%d HTTP/1.0' % self.tcp_listener.getHost().port)
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 403 response
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 403 Bad Proxy Request')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Type: text/html')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')
        data = yield self.proto.data.get()
        self.assertEqual(data, 'Unable to accept non-HTTPS requests')

    @defer.inlineCallbacks
    def test_http_proxy_reject_bad_gateway(self):
        # send GET request through proxy without first sending CONNECT
        self.proto.sendLine('CONNECT 127.0.0.1:0 HTTP/1.0')
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 200 OK
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 502 Gateway Error')


class HttpProxyCidrRestrictedClientSuccessTests(BaseHttpProxyTestCases):

    @defer.inlineCallbacks
    def setUp(self):
        self.client_ips = [netaddr.IPNetwork('127.0.0.0/24')]
        yield super(HttpProxyCidrRestrictedClientSuccessTests, self).setUp()

    @defer.inlineCallbacks
    def test_http_proxy_accept_connect_from_approved_client(self):
        self.proto.sendLine('CONNECT 127.0.0.1:%d HTTP/1.0' % self.tcp_listener.getHost().port)
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 200 OK
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 200 CONNECT OK')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'X-Connected-Ip: 127.0.0.1')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Length: 0')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')


class HttpProxyRestrictedClientFailureTestCases(BaseHttpProxyTestCases):

    @defer.inlineCallbacks
    def setUp(self):
        self.client_ips = [netaddr.IPNetwork('9.9.9.9')]
        yield super(HttpProxyRestrictedClientFailureTestCases, self).setUp()

    @defer.inlineCallbacks
    def test_http_proxy_reject_connect_from_unapproved_client(self):
        # send CONNECT request from unapproved host
        self.proto.sendLine('CONNECT 127.0.0.1:%d HTTP/1.0' % self.tcp_listener.getHost().port)
        self.proto.sendLine('User-Agent: DuoUnitTestClient')
        self.proto.sendLine('')

        # expect 403 response
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'HTTP/1.0 403 Bad Proxy Request')
        line = yield self.proto.lines.get()
        self.assertEqual(line, 'Content-Type: text/html')
        line = yield self.proto.lines.get()
        self.assertEqual(line, '')
        data = yield self.proto.data.get()
        self.assertEqual(data, 'Attempted connection creation from non-approved client IP: 127.0.0.1')


class TestHttpProxyAllowedIps(TestCase):
    def test_single_ip(self):
        expected_result = [netaddr.IPNetwork('1.2.3.4/32')]

        config = ConfigDict({'client_ip': '1.2.3.4'})

        actual_result = http_proxy.get_allowed_ip_networks(config)

        self.assertItemsEqual(expected_result, actual_result)

    def test_ip_range(self):
        expected_result = [netaddr.IPNetwork('1.2.3.0/24')]

        config = ConfigDict({'client_ip': '1.2.3.0-1.2.3.255'})

        actual_result = http_proxy.get_allowed_ip_networks(config)

        self.assertItemsEqual(expected_result, actual_result)

    def test_ip_cidr(self):
        expected_result = [netaddr.IPNetwork('1.2.0.0/16')]

        config = ConfigDict({'client_ip': '1.2.0.0/16'})

        actual_result = http_proxy.get_allowed_ip_networks(config)

        self.assertItemsEqual(expected_result, actual_result)

    def test_mix(self):
        expected_result = [
            netaddr.IPNetwork('1.2.3.4/32'),
            netaddr.IPNetwork('2.3.4.0/24'),
            netaddr.IPNetwork('3.4.0.0/16')
        ]

        config = ConfigDict({'client_ip': '1.2.3.4,2.3.4.0-2.3.4.255,3.4.0.0/16'})

        actual_result = http_proxy.get_allowed_ip_networks(config)

        self.assertItemsEqual(expected_result, actual_result)


class _LineReceiverTestProtocol(LineReceiver):
    def __init__(self):
        self.lines = defer.DeferredQueue()
        self.data = defer.DeferredQueue()
        self.length = -1

    def connectionMade(self):
        self.factory.deferred.callback(self)

    def connectionLost(self, reason):
        for d in self.lines.waiting:
            d.errback(reason)

    def lineReceived(self, line):
        self.lines.put(line)

        # set length of body data
        if 'Content-Length: ' in line:
            self.length = int(line[16:])
        # switch to raw mode to receive body
        if line == '' and (self.length != 0):
            self.setRawMode()

    def rawDataReceived(self, data):
        # only decrement the length for non-chunked encoding
        if self.length != -1:
            data, rest = data[:self.length], data[self.length:]
            self.length = self.length - len(data)
        self.data.put(data)
        # switch back to line mode for future requests
        if self.length == 0:
            self.setLineMode(rest)


class _LineReceiverTestFactory(protocol.ClientFactory):
    protocol = _LineReceiverTestProtocol

    def __init__(self):
        self.deferred = defer.Deferred()


if __name__ == '__main__':
    import unittest
    unittest.main()
