#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
import socket

from pyrad import packet

from twisted.internet import defer

import duoauthproxy.modules.radius_client
from duoauthproxy.lib.config_provider import ConfigDict, ConfigError
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.lib import protect

from base import TestCase, ClientProtectedConfigParsingTestCase
import mock_objects


class TestModuleRadiusClient(TestCase):
    def setUp(self):
        self.username = 'someuser'
        self.password = 'somePassword!'
        self.client_ip = '12.34.56.78'
        self.config_items = {
            'host': '1.2.3.4',
            'port': '1812',
            'host_2': '1.2.3.5',
            'port_2': '11812',
            'secret': 'jawokenmahtubercon',
            'nas_ip': '5.6.7.8',
        }
        config = ConfigDict(self.config_items)
        self.instance = duoauthproxy.modules.radius_client.Module(config)
        self.instance.protocol.transport = mock_objects.DeferredMockTransport()

    def tearDown(self):
        self.instance.stopService()

    @defer.inlineCallbacks
    def test_basic_response(self):
        d = self.instance.authenticate(self.username,
                                       self.password,
                                       self.client_ip)

        queue = self.instance.protocol.transport.queue
        (req_data, dst) = yield queue.get()
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.config_items['secret']
        self.assertEqual(request.code, packet.AccessRequest)
        self.assertEqual(request[1], [self.username])
        self.assertEqual(request.PwDecrypt(request[2][0]), self.password)
        self.assertEqual(request[4],
                         [socket.inet_aton(self.config_items['nas_ip'])])
        self.assertEqual(request[31], [self.client_ip])

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        self.instance.protocol.datagramReceived(response_packet.ReplyPacket(),
                                                (self.config_items['host'],
                                                 self.config_items['port']))
        res = yield d
        self.assertFalse(res.msg)
        self.assertTrue(res.success)

    @defer.inlineCallbacks
    def test_pass_through(self):
        attr_name = 'Filter-Id'
        attr_value = 'moo'
        self.instance.pass_through_attr_names = [attr_name]

        d = self.instance.authenticate(self.username,
                                       self.password,
                                       self.client_ip)

        queue = self.instance.protocol.transport.queue
        (req_data, dst) = yield queue.get()
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.config_items['secret']
        self.assertEqual(request.code, packet.AccessRequest)
        self.assertEqual(request[1], [self.username])
        self.assertEqual(request.PwDecrypt(request[2][0]), self.password)
        self.assertEqual(request[4],
                         [socket.inet_aton(self.config_items['nas_ip'])])
        self.assertEqual(request[31], [self.client_ip])

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet[attr_name] = attr_value
        self.instance.protocol.datagramReceived(response_packet.ReplyPacket(),
                                                (self.config_items['host'],
                                                 self.config_items['port']))
        res = yield d
        self.assertFalse(res.msg)
        self.assertTrue(res.success)
        self.assertEqual(res.radius_attrs[attr_name], [attr_value])

    @defer.inlineCallbacks
    def test_pass_through_all(self):
        attr_name = 'Filter-Id'
        attr_num = 11
        attr_value = 'moo'
        self.instance.pass_through_all = True

        d = self.instance.authenticate(self.username,
                                       self.password,
                                       self.client_ip)

        queue = self.instance.protocol.transport.queue
        (req_data, dst) = yield queue.get()
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.config_items['secret']

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet[attr_name] = attr_value
        self.instance.protocol.datagramReceived(response_packet.ReplyPacket(),
                                                (self.config_items['host'],
                                                 self.config_items['port']))
        res = yield d

        self.assertFalse(res.msg)
        self.assertTrue(res.success)
        self.assertEqual(res.radius_attrs[attr_num], [attr_value])

    def test_parse_radius_addr(self):
        self.assertEqual(self.instance.protocol.addrs, [
            ('1.2.3.4', 1812),
            ('1.2.3.5', 11812),
        ])


class TestModuleRadiusClientAddressConfigParsing(TestCase):
    def _make_instance(self, more_config_items):
        config_items = {
            'secret': 'jawokenmahtubercon',
        }
        config_items.update(more_config_items)
        config = ConfigDict(config_items)
        instance = duoauthproxy.modules.radius_client.Module(config)
        return instance

    def test_host(self):
        instance = self._make_instance({
            'host': '1.2.3.5',
            'port': '11812',
        })
        self.assertEqual(instance.protocol.addrs, [
            ('1.2.3.5', 11812),
        ])

    def test_host_1(self):
        instance = self._make_instance({
            'host_1': '1.2.3.5',
            'port_1': '11812',
        })
        self.assertEqual(instance.protocol.addrs, [
            ('1.2.3.5', 11812),
        ])

    def test_host_and_host_2(self):
        instance = self._make_instance({
            'host': '1.2.3.4',
            'port': '1812',
            'host_2': '1.2.3.5',
            'port_2': '11812',
        })
        self.assertEqual(instance.protocol.addrs, [
            ('1.2.3.4', 1812),
            ('1.2.3.5', 11812),
        ])

    def test_host_1_and_host_2(self):
        instance = self._make_instance({
            'host_1': '1.2.3.4',
            'port_1': '1812',
            'host_2': '1.2.3.5',
            'port_2': '11812',
        })
        self.assertEqual(instance.protocol.addrs, [
            ('1.2.3.4', 1812),
            ('1.2.3.5', 11812),
        ])

    def test_host_2(self):
        with self.assertRaisesRegexp(ConfigError, "Missing.*host'"):
            self._make_instance({
                'host_2': '1.2.3.5',
                'port_2': '11812',
            })


class TestRadiusClientProtectedConfigParsing(ClientProtectedConfigParsingTestCase):

    def test_plain_keys_use_protected_api(self):
        instance = self._make_instance({
            'secret': 'secret!',
        }, duoauthproxy.modules.radius_client)
        self.assertEqual(instance.protocol.secret, 'secret!')

    def test_protected_keys(self):
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance({
                'secret_protected': protected,
            }, duoauthproxy.modules.radius_client)
            self.assertEqual(instance.protocol.secret, 'secret!')


if __name__ == '__main__':
    import unittest
    unittest.main()
