#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#

from pyrad import packet

from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.radius.base import radius_dictionary

from twisted.internet import defer
from twisted.trial.unittest import TestCase

import duoauthproxy.modules.radius_server_challenge
from duoauthproxy.lib import duo_async, base, protect, log

from base import (
    ServerModuleTest,
    ServerModuleTestPassThrough,
    ServerModuleTestPassThroughMultipleValues,
    ServerProtectedConfigParsingTestCase,
    AuthLoggingTest
)


class RadiusServerChallengeTestRunner(ServerModuleTest, TestCase):
    """Tester class inheritable by all other RadiusServerChallenge test classes.

    Any test class that wants to inherit from ServerModuleTest, TestCase, or both
    of them can inherit from this class instead.

    This class was created due to multiple inheritance super() incompatibilities
    with TestRadiusServerChallengeAuthLogging, though it's since been co-opted
    for use with other RadiusServerChallenge test classes as well to cut down on
    extra setUp() redefinitions.
    """

    def setUp(self):
        super(RadiusServerChallengeTestRunner, self).setUp(
            {},
            duoauthproxy.modules.radius_server_challenge,
            'radius_server_challenge'
        )


class RadiusServerChallengeTestCases(object):
    """Test suite covering various DuoTextChallengeRadiusServer auth flows.

    Since the user effectively interacts with DuoTextChallengeRadiusServer twice
    (once to pass primary creds on, then again to respond to the text challenge
    we provide), there are tests that deal with the initial request as well as
    others that manipulate challenge_state in order to emulate the second stage.

    MS-CHAP2 requests contain different RADIUS attributes and handles enrollment
    challenges somewhat differently, so we keep MS-CHAP2 tests separate from the
    others here.

    Note:
        Since there's no setUp() or tearDown() defined in this class, it needs
        to make use of the setUp() defined in RadiusServerChallengeTestRunner.
    """

    # coverage for response to initial request:

    @defer.inlineCallbacks
    def test_missing_username(self):
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

    @defer.inlineCallbacks
    def test_wrong_primary_auth_password(self):
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='foo')

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_enroll_enroll_challenge_false(self):
        server = self.instance.protocol
        server.enroll_challenge = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_enroll_enroll_challenge_true(self):
        server = self.instance.protocol
        server.enroll_challenge = True
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_auth(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password)

    # coverage for response to enroll challenge:

    @defer.inlineCallbacks
    def test_enroll_challenge_always_rejects(self):
        # If we send an enrollment challenge to the user, DuoTextChallengeRadiusServer
        # will automatically send an Access-Reject in response to the user's response
        # to the text challenge
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_response='anything',
                              challenge_state={
                                  'challenge_handler': self.instance.protocol.handle_enroll_challenge,
                                  'is_chap2': False})

    # coverage for response to auth challenge:

    @defer.inlineCallbacks
    def test_auth_challenge_auth_allow(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_without_user_password(self):
        # CHAP or some other request with known State but no User-Password.
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              challenge_state=challenge_state,
                              challenge_send_user_password=False,
                              challenge_response='')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_deny(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_fail_open_failmode_secure(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              failmode=base.FAILMODE_SECURE,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_fail_closed_failmode_secure(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              failmode=base.FAILMODE_SECURE,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_fail_open_failmode_safe(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              failmode=base.FAILMODE_SAFE,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_auth_fail_closed_failmode_safe(self):
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              failmode=base.FAILMODE_SAFE,
                              challenge_state=challenge_state,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_auth_challenge_sms_refresh(self):
        self.enqueue_duo_auth('not checked')
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {
                '2': 'sms0',
            },
            'prompt': '?',
        }
        yield self._test_auth(
            packet.AccessChallenge,
            username=self.username,
            challenge_response='2',
            challenge_state=challenge_state,
        )

    # coverage for MS-CHAP2 requests

    def test_chap2_multiple_subattrs(self):
        # Verify that pyrad correctly parses a RADIUS Vendor-Specific attribute.
        pckt = '\x01\x03\x00\xaa\xd9Ca\xe7\x82\xa7D\x17\xba\xaf<\x1e6"A\x06 \x12FGT50B3G11620726\x01\x08ktest1\x1aL\x00\x00\x017\x194\x82\x00\xa3\x85B\xc9\x949\xa4A\x95\x18^T\x049\xb5}\x00\x00\x00\x00\x00\x00\x00\x00[\x87\x8f3\xb9}%/\xd7\x12\x9e\xc49\xdd\n\xa1\xa2j[p\xec\x1b\xe3\xc8\x0b\x12\xdb\xb6GT\xb5\xc7\x1f/\x97`\xb8\xa1\xed\x87\xd8\xdf\x1f\x11192.168.240.210,\n01000003M\tvpn-ssl\x1a\x0c\x00\x000D\x03\x06root'
        p = packet.AuthPacket(packet=pckt, dict=radius_dictionary())
        for k, _ in self.chap2_request_attrs.items():
            self.assertTrue(k in p)

    @defer.inlineCallbacks
    def test_chap2_missing_username(self):
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_ps_expired_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_exempt_username_expired(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_ps_expired_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_enroll(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        server = self.instance.protocol
        server.enroll_challenge = False
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_enroll_challenge(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        server = self.instance.protocol
        server.enroll_challenge = True
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              reject_attrs=self.chap2_ps_expired_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_enroll(self):
        server = self.instance.protocol
        server.enroll_challenge = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_enroll_challenge(self):
        server = self.instance.protocol
        server.enroll_challenge = True
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_exempt_username_enroll_challenge(self):
        # An exempt user should be accepted without having to enroll or respond
        # to challenges
        self.instance.protocol.exempt_usernames = [self.username]
        server = self.instance.protocol
        server.enroll_challenge = True
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_allow(self):
        server = self.instance.protocol
        server.enroll_challenge = True
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_allow_fail(self):
        server = self.instance.protocol
        server.enroll_challenge = True
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              request_attrs=self.chap2_bad_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_auth(self):
        server = self.instance.protocol
        server.enroll_challenge = True
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_allow(self):
        # Respond to the challenge; keep the original cpw request in state
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_allow_fail(self):
        # Respond to the challenge; keep the original cpw request in state
        # By verifying a failure, this test case ensures that the
        # (invalid) credentials are actually forwarded to the mock client.
        request = self.create_cpw_req(bad=True)
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_deny(self):
        # Respond to the challenge; keep the original cpw request in state
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_fail_open_safe_fail(self):
        # Respond to the challenge; Bad primary auth in CPW request should
        # result in AccessReject, verifying that the CPW is correctly
        # forwarded to the (mock) RADIUS server
        request = self.create_cpw_req(bad=True)
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_fail_open_safe(self):
        # Respond to the challenge; keep the original cpw request in state
        # Password wasn't changed, but we should fail open in this case?
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_fail_open_secure(self):
        # Respond to the challenge; keep the original cpw request in state
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_fail_closed_safe(self):
        # Respond to the challenge; keep the original cpw request in state
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_chap2_cpw_challenge_response_fail_closed_secure(self):
        # Respond to the challenge; keep the original cpw request in state
        request = self.create_cpw_req()
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_cpw_auth_challenge,
            'cpw': request,
            'is_chap2': True,
            'factors': {},
            'prompt': '?'
        }
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              challenge_state=challenge_state,
                              reject_attrs=self.chap2_error_attr,
                              challenge_response='push')


class TestRadiusServerChallenge(RadiusServerChallengeTestCases,
                                RadiusServerChallengeTestRunner):
    """Runs all tests in RadiusServerChallengeTestCases.

    Note:
        test_exempt_username_config_parsing is kept here and not in
        RadiusServerChallengeTestCases since we don't need to run it more than once.
    """

    def test_exempt_username_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
        }
        self.instance = duoauthproxy.modules.radius_server_challenge.Module(
            ConfigDict(server_config),
            None,
            'radius_server_challenge'
        )

        server = self.instance.protocol
        self.assertEqual(server.exempt_usernames, ['testuser1', 'testuser2'])


class TestRadiusServerChallengePassThrough(RadiusServerChallengeTestCases,
                                           ServerModuleTestPassThrough,
                                           RadiusServerChallengeTestRunner):
    """Runs all tests in RadiusServerChallengeTestCases with an additional
    pass-through attribute to be sent and received by the server."""
    pass


class TestRadiusServerChallengePassThroughMultipleValues(RadiusServerChallengeTestCases,
                                                         ServerModuleTestPassThroughMultipleValues,
                                                         RadiusServerChallengeTestRunner):
    """Runs all tests in RadiusServerChallengeTestCases with several additional
    pass-through attributes to be sent and received by the server."""
    pass


class TestRadiusChallengeProtectedConfigParsing(ServerProtectedConfigParsingTestCase):
    """Tests that an instance of DuoTextChallengeRadiusServer with a protected
    config properly accesses and/or decrypts certain protected keys defined in
    ProtectedValidatingConfigDict."""

    def test_plain_keys_use_protected_api(self):
        # Both of these new keys should be accessed with get_protected_str()
        # though neither are encrypted
        instance = self._make_instance(
            {
                'radius_secret_1': 'secret!',
                'skey': 'skey!'
            },
            duoauthproxy.modules.radius_server_challenge,
            'radius_server_challenge'
        )
        self.assertEqual(instance.protocol.client.skey, 'skey!')
        self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')

    def test_protected_keys(self):
        # radius_secret_protected_1 should decrypt first with get_protected_str()
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance(
                {
                    'radius_secret_protected_1': protected,
                    'skey': 'skey!'
                },
                duoauthproxy.modules.radius_server_challenge,
                'radius_server_challenge'
            )
            self.assertEqual(instance.protocol.client.skey, 'skey!')
            self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')


class TestRadiusServerChallengeAuthLogging(AuthLoggingTest,
                                           RadiusServerChallengeTestRunner):
    """Tests SIEM logging from log.auth_standard() calls during various
    authentication flows in DuoTextChallengeRadiusServer."""

    # coverage for calls to get_initial_response():

    @defer.inlineCallbacks
    def test_initial_response_no_username(self):
        expected = [{
            'msg': 'No username provided',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ERROR,
        }]
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_initial_response_no_password(self):
        expected = [{
            'msg': 'No password or CHAP2 attributes provided',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ERROR,
        }]
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=None)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_initial_response_chap2_cpw_preauth_deny_during_reenroll(self):
        # If a user tries to auth in with an expired MS-CHAP2 password,
        # the server will preauth the user against the Duo API to make sure they
        # exist, have proper permissions, etc. In this case, Duo preauth fails.
        expected = [{
            'msg': 'Primary credentials rejected - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_REJECT,
        }, {
            'msg': 'Duo preauthentication failed - deny',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              request_attrs=self.chap2_expired_request_attrs)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_initial_response_chap2_cpw_reenroll(self):
        # A user with an expired MS-CHAP2 password that also successfully
        # re-preauths with Duo during the call to get_initial_response() will be
        # prompted by DuoTextChallengeRadiusServer to re-enroll.
        expected = [{
            'msg': 'Primary credentials rejected - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_REJECT,
        }, {
            'msg': log.AUTH_ENROLL_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        server = self.instance.protocol
        server.enroll_challenge = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)
        self.assert_logged(expected)

    # coverage for calls to preauth():

    @defer.inlineCallbacks
    def test_preauth_exempt_username(self):
        expected = [{
            'msg': 'Primary authentication successful - Hello',
            'auth_stage': 'Primary authentication',
            'status': log.AUTH_ALLOW,
        }, {
            'msg': 'Duo preauthentication successful - User exempted from 2FA',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
        }]
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_preauth_api_fail_open(self):
        expected = [{
            'msg': 'Primary authentication successful - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'msg': 'Duo preauthentication successful - {0}'.format(base.FAILMODE_SAFE_MSG),
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
        }]
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_preauth_api_fail_closed(self):
        expected = [{
            'msg': 'Primary authentication successful - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'msg': 'Duo preauthentication failed - API call failed',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ERROR,
        }]
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    # coverage for calls to _preauth_response():

    @defer.inlineCallbacks
    def test_preauth_response_duo_preauth_reject(self):
        expected = [{
            'msg': 'Primary authentication successful - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'msg': 'Duo preauthentication failed - deny',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_preauth_response_duo_enroll_reject(self):
        expected = [{
            'msg': 'Primary authentication successful - Hello',
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'msg': log.AUTH_ENROLL_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        server = self.instance.protocol
        server.enroll_challenge = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    # coverage for calls to handle_auth_challenge():

    @defer.inlineCallbacks
    def test_handle_auth_challenge_api_fail_open(self):
        expected = [{
            'msg': base.FAILMODE_SAFE_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
        }]
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              failmode=base.FAILMODE_SAFE,
                              challenge_state=challenge_state,
                              challenge_response='push')
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_handle_auth_challenge_api_fail_closed(self):
        expected = [{
            'msg': base.FAILMODE_SECURE_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              failmode=base.FAILMODE_SECURE,
                              challenge_state=challenge_state,
                              challenge_response='push')
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_handle_auth_challenge_duo_auth_reject(self):
        expected = [{
            'msg': 'deny',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': True,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password,
                              challenge_state=challenge_state,
                              challenge_response='push')
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_handle_auth_challenge_duo_auth_allow(self):
        expected = [{
            'msg': 'allow',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
        }]
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_auth_challenge,
            'is_chap2': True,
            'factors': {},
            'prompt': '?',
        }
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password,
                              challenge_state=challenge_state,
                              challenge_response='push')
        self.assert_logged(expected)

    # coverage for calls to handle_enroll_challenge():

    @defer.inlineCallbacks
    def test_handle_enroll_challenge(self):
        expected = [{
            'msg': 'Please complete the enrollment process and login again',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        challenge_state = {
            'challenge_handler': self.instance.protocol.handle_enroll_challenge,
            'is_chap2': False,
            'factors': {},
            'prompt': '?',
        }
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password,
                              challenge_state=challenge_state,
                              challenge_response='push')
        self.assert_logged(expected)

    # coverage for calls to _preauth_cpw_response():

    @defer.inlineCallbacks
    def test_preauth_cpw_response_duo_preauth_reject(self):
        expected = [{
            'msg': 'Duo preauthentication failed - deny',
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_preauth_cpw_response_duo_enroll_reject(self):
        expected = [{
            'msg': log.AUTH_ENROLL_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]
        server = self.instance.protocol
        server.enroll_challenge = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)
        self.assert_logged(expected)


if __name__ == '__main__':
    import unittest
    unittest.main()
