#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#

from pyrad import packet

from twisted.internet import defer
from twisted.trial.unittest import TestCase

import duoauthproxy.modules.radius_server_concat
from duoauthproxy.lib import duo_async, base, protect, log
from duoauthproxy.lib.config_provider import ConfigDict

from base import (
    ServerModuleTest,
    ServerModuleTestPassThrough,
    ServerModuleTestPassThroughMultipleValues,
    ServerProtectedConfigParsingTestCase,
    AuthLoggingTest
)


class RadiusServerConcatTestRunner(ServerModuleTest, TestCase):
    """Tester class inheritable by all other RadiusServerConcat test classes.

    Any test class that wants to inherit from ServerModuleTest, TestCase, or both
    of them can inherit from this class instead.

    This class was created due to multiple inheritance super() incompatibilities
    with TestRadiusServerConcatAuthLogging, though it's since been co-opted
    for use with other RadiusServerConcat test classes as well to cut down on
    extra setUp() redefinitions.
    """
    def setUp(self):
        self.factor = 'phone9'
        super(RadiusServerConcatTestRunner, self).setUp(
            {},
            duoauthproxy.modules.radius_server_concat,
            'radius_server_concat'
        )


class RadiusServerConcatTestCases(object):
    """Test suite covering various DuoConcatRadiusServer auth flows.

    Be sure to pass in the password for _test_auth() as a concatenation of the
    actual password, the delimiter, and the chosen factor, like how a real user
    would authenticate.

    Note:
        Since there's no setUp() or tearDown() defined in this class, it needs
        to make use of the setUp() defined in RadiusServerConcatTestRunner.
    """

    # coverage for improperly-formatted primary credentials:

    @defer.inlineCallbacks
    def test_missing_username(self):
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

    @defer.inlineCallbacks
    def test_missing_concat(self):
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=('not ' + self.password))

    @defer.inlineCallbacks
    def test_wrong_primary_auth_password(self):
        server = self.instance.protocol
        password = server.delim.join((('not ' + self.password),
                                      self.factor))
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    # coverage for preauthentication:

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_enroll(self):
        # treated as deny by DuoSimpleRadiusServer
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_secure(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_secure(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_safe(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_safe(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=password)

    # coverage for secondary authentication:

    @defer.inlineCallbacks
    def test_preauth_auth_auth_allow(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.factor)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_secure(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_secure(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_safe(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_safe(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_deny(self):
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY,
                              factor=self.factor)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=password)

    # coverage for non-normal delimiters:

    @defer.inlineCallbacks
    def test_wrong_delim(self):
        # Other delimiters don't work. Also assert password format
        # errors cause a reject before any Duo API calls.
        server = self.instance.protocol
        wrong_delim = '/'
        if server.delim == wrong_delim:
            # Ensure default delim doesn't work when non-default
            # configured.
            wrong_delim = ','
        password = wrong_delim.join((self.password, self.factor))
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_exempt_username_wrong_delim(self):
        # Other delimiters don't work. Also assert password format
        # errors cause a reject before any Duo API calls.
        server = self.instance.protocol
        wrong_delim = '/'
        if server.delim == wrong_delim:
            # Ensure default delim doesn't work when non-default
            # configured.
            wrong_delim = ','
        password = wrong_delim.join((self.password, self.factor))
        server.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)

    @defer.inlineCallbacks
    def test_multi_char_delim(self):
        delim = ';;'
        self.instance.protocol.delim = delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=packet.AccessAccept,
            username=self.username,
            password=(self.password + delim + self.factor),
        )

    # coverage for parsing radius_server_concat's server configuration:

    def test_delim_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
        }

        # Both spellings accepted
        for spelling in ('delimeter', 'delimiter'):
            delim = 'foo'
            a_server_config = dict(server_config)
            a_server_config[spelling] = delim
            self.instance = duoauthproxy.modules.radius_server_concat.Module(
                ConfigDict(a_server_config),
                None,
                'radius_server_concat',
            )
            self.assertEqual(self.instance.protocol.delim, delim)

        # "delimiter" takes precedence over "delimeter"
        a_server_config = dict(server_config)
        a_server_config['delimiter'] = delim
        a_server_config['delimeter'] = 'not ' + delim
        self.instance = duoauthproxy.modules.radius_server_concat.Module(
            ConfigDict(a_server_config),
            None,
            'radius_server_concat',
        )
        self.assertEqual(self.instance.protocol.delim, delim)


class TestRadiusServerConcat(RadiusServerConcatTestCases,
                             RadiusServerConcatTestRunner):
    """Runs all tests in RadiusServerConcatTestCases.

    Note:
        test_exempt_username_config_parsing is kept here and not in
        RadiusServerConcatTestCases since we don't need to run it more than once.
    """

    def test_exempt_username_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
        }
        self.instance = duoauthproxy.modules.radius_server_concat.Module(
            ConfigDict(server_config),
            None,
            'radius_server_concat'
        )
        server = self.instance.protocol
        self.assertEqual(server.exempt_usernames, ['testuser1', 'testuser2'])


class TestRadiusServerConcatNonDefaultDelim(RadiusServerConcatTestCases,
                                            RadiusServerConcatTestRunner):
    """Runs the tests in RadiusServerConcatTestCases with '/' as the delimiter
    instead of the default ','. """

    def setUp(self, *args, **kwargs):
        super(TestRadiusServerConcatNonDefaultDelim,
              self).setUp(*args, **kwargs)
        server = self.instance.protocol
        server.delim = '/'
    pass


class TestRadiusServerConcatPassThrough(RadiusServerConcatTestCases,
                                        ServerModuleTestPassThrough,
                                        RadiusServerConcatTestRunner):
    """Runs all tests in RadiusServerConcatTestCases with an additional
    pass-through attribute to be sent and received by the server."""
    pass


class TestRadiusServerConcatPassThroughMultipleValues(
        RadiusServerConcatTestCases,
        ServerModuleTestPassThroughMultipleValues,
        RadiusServerConcatTestRunner):
    """Runs all tests in RadiusServerConcatTestCases with multiple additional
    pass-through attributes to be sent and received by the server."""
    pass


class TestRadiusConcatProtectedConfigParsing(ServerProtectedConfigParsingTestCase):
    """Tests that an instance of DuoConcatRadiusServer with a protected
    config properly accesses and/or decrypts certain protected keys defined in
    ProtectedValidatingConfigDict."""

    def test_plain_keys_use_protected_api(self):
        # Both of these new keys should be accessed with get_protected_str()
        # though neither are encrypted
        instance = self._make_instance({
            'radius_secret_1': 'secret!',
            'skey': 'skey!'
        }, duoauthproxy.modules.radius_server_concat, 'radius_server_concat')
        self.assertEqual(instance.protocol.client.skey, 'skey!')
        self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')

    def test_protected_keys(self):
        # radius_secret_protected_1 should decrypt first with get_protected_str()
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance(
                {
                    'radius_secret_protected_1': protected,
                    'skey': 'skey!'
                },
                duoauthproxy.modules.radius_server_concat,
                'radius_server_concat'
            )
            self.assertEqual(instance.protocol.client.skey, 'skey!')
            self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')


class TestRadiusServerConcatAuthLogging(AuthLoggingTest, RadiusServerConcatTestRunner):
    """Tests SIEM logging from log.auth_standard() calls during various
    authentication flows in DuoConcatRadiusServer."""

    @defer.inlineCallbacks
    def test_missing_username(self):
        expected = [{
            "msg": "No username provided",
            "auth_stage": log.AUTH_UNKNOWN,
            "status": log.AUTH_ERROR,
        }]

        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_missing_password(self):
        expected = [{
            "msg": "Missing or improperly-formatted password - No password provided",
            "auth_stage": log.AUTH_PRIMARY,
            "status": log.AUTH_ERROR,
        }]

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=None)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_missing_password_delim(self):
        expected = [{
            "msg": "Missing or improperly-formatted password - Invalid Password",
            "auth_stage": log.AUTH_PRIMARY,
            "status": log.AUTH_ERROR,
        }]

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_wrong_password_primary_auth_deny(self):
        expected = [{
            "msg": "Primary credentials rejected - Hello",
            "auth_stage": log.AUTH_PRIMARY,
            "status": log.AUTH_REJECT,
        }]

        server = self.instance.protocol
        password = server.delim.join((('not ' + self.password),
                                     self.factor))
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_auth_deny(self):
        expected = [{
            "msg": "Primary authentication successful - Hello",
            "auth_stage": log.AUTH_PRIMARY,
            "status": log.AUTH_ALLOW,
        }, {
            'auth_stage': log.AUTH_SECONDARY,
            'msg': 'deny',
            'status': log.AUTH_REJECT,
        }]

        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY,
                              factor=self.factor)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=password)
        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_auth_success(self):
        expected = [{
            "msg": "Primary authentication successful - Hello",
            "auth_stage": log.AUTH_PRIMARY,
            "status": log.AUTH_ALLOW,
        }, {
            'auth_stage': log.AUTH_SECONDARY,
            'msg': 'allow',
            'status': log.AUTH_ALLOW,
        }]

        server = self.instance.protocol
        password = server.delim.join((self.password, self.factor))
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.factor)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=password)
        self.assert_logged(expected)


if __name__ == '__main__':
    import unittest
    unittest.main()
