#
# Copyright (c) 2013-2015 Duo Security
# All Rights Reserved
#

from pyrad import packet

from twisted.internet import defer
from twisted.trial.unittest import TestCase

import duoauthproxy.modules.radius_server_concat
from duoauthproxy.lib import duo_async, base, protect, log
from duoauthproxy.lib.config_provider import ConfigDict

from base import (
    ServerModuleTest,
    ServerProtectedConfigParsingTestCase,
    AuthLoggingTest
)


class RadiusServerDuoOnlyTestRunner(ServerModuleTest, TestCase):
    """Tester class inheritable by all other RadiusServerDuoOnly test classes.

    Any test class that wants to inherit from ServerModuleTest, TestCase, or both
    of them can inherit from this class instead.

    This class was created due to multiple inheritance super() incompatibilities
    with TestRadiusServerDuoOnlyAuthLogging, though it's since been co-opted
    for use with other RadiusServerDuoOnly test classes as well to cut down on
    extra setUp() redefinitions.
    """
    def setUp(self):
            super(RadiusServerDuoOnlyTestRunner, self).setUp(
                {},
                duoauthproxy.modules.radius_server_duo_only,
                'radius_server_duo_only'
            )


class RadiusServerDuoOnlyTestCases(object):
    """Test suite covering various DuoOnlyRadiusServer auth flows.

    Note:
        Since there's no setUp() or tearDown() defined in this class, it needs
        to make use of the setUp() defined in RadiusServerDuoOnlyTestRunner.
    """

    @defer.inlineCallbacks
    def test_missing_username(self):
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

    # coverage for preauthentication:

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_allow_with_groups(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW, groups=[u'Spinal Tap', u'The Folksmen'])
        self.pass_through_radius_attrs['Class'] = ['OU=Spinal Tap;']
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_duo_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_duo_preauth_enroll(self):
        # treated as deny by DuoSimpleRadiusServer
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password='push')

    # coverage for secondary authentication:

    @defer.inlineCallbacks
    def test_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SAFE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=base.FAILMODE_SECURE,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_preauth_auth_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

    @defer.inlineCallbacks
    def test_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')


class TestRadiusServerDuoOnly(RadiusServerDuoOnlyTestCases,
                              RadiusServerDuoOnlyTestRunner):
    """Runs all tests in RadiusServerDuoOnlyTestCases.

    Note:
        test_exempt_username_config_parsing is kept here and not in
        RadiusServerDuoOnlyTestCases since we don't need to run it more than once.
    """

    def test_exempt_username_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
        }
        self.instance = duoauthproxy.modules.radius_server_duo_only.Module(
            ConfigDict(server_config),
            None,
            "radius_server_duo_only",
        )
        server = self.instance.protocol
        self.assertEqual(server.exempt_usernames, ['testuser1', 'testuser2'])


class TestRadiusServerDuoOnlyAuthLogging(AuthLoggingTest,
                                         RadiusServerDuoOnlyTestRunner):
    """Tests SIEM logging from log.auth_standard() calls during various
    authentication flows in DuoOnlyRadiusServer."""

    @defer.inlineCallbacks
    def test_missing_username(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ERROR,
            'msg': "No username provided",
        }]

        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_missing_password(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ERROR,
            'msg': 'Missing or improperly-formatted password'
        }]

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=None)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_auth_success(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW
        }]

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_auth_fail(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT
        }]

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
        }]
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_preauth_deny(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]

        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_preauth_enroll(self):
        expected = [{
            'msg': log.AUTH_ENROLL_MSG,
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
        }]

        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='push')

        self.assert_logged(expected)


class TestRadiusDuoOnlyProtectedConfigParsing(ServerProtectedConfigParsingTestCase):
    """Tests that an instance of DuoOnlyRadiusServer with a protected
    config properly accesses and/or decrypts certain protected keys defined in
    ProtectedValidatingConfigDict."""

    def test_plain_keys_use_protected_api(self):
        # Both of these new keys should be accessed with get_protected_str()
        # though neither are encrypted
        instance = self._make_instance(
            {
                'radius_secret_1': 'secret!',
                'skey': 'skey!'
            },
            duoauthproxy.modules.radius_server_duo_only,
            'radius_server_duo_only')
        self.assertEqual(instance.protocol.client.skey, 'skey!')
        self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')

    def test_protected_keys(self):
        # radius_secret_protected_1 should decrypt first with get_protected_str()
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance(
                {
                    'radius_secret_protected_1': protected,
                    'skey': 'skey!'
                },
                duoauthproxy.modules.radius_server_duo_only,
                'radius_server_duo_only'
            )
            self.assertEqual(instance.protocol.client.skey, 'skey!')
            self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')


if __name__ == '__main__':
    import unittest
    unittest.main()
